// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Michael Lehto.
 * Copyright (C) 1999 Quality First Software.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib;

// {{{ imports

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.DirectoryScanner;

// }}}

// {{{ doc

/**
 * QFS pre-parser, removes/adds conditional statements within source files.
 *
 * Source files (.jpp by default) are scanned for conditions and flags that
 * are prepended with the '#' character, such as: "#if JDK == 12" and
 * generates this code to the output file (*.java by default).
 *
 * Conditional code is grouped into two categories: conditions and flags.
 *
 * Flags are simple boolean statements in the form of:
 *   #ifdef <FLAG>
 *      <some java code here>
 *   #elifdef <FLAG>
 *      <more code>
 *   #else
 *      <more code>
 *   #endif
 * A flag is expressed in the negative as "#ifndef <FLAG>"
 *
 * Conditions are statements which are evaluated according to the value
 * assigned. They are in the form of:
 *   #if <CONDITION> == <VALUE>
 *      <code>
 *   #elif <CONDITION> == <VALUE>
 *      <code>
 *   #else
 *      <code>
 *   #endif
 * The negative of a condition is expressed as "#if <CONDITION> != <VALUE>"
 *
 * Nested conditions and flags are suppored.
 *
 * Note that this class is NOT a complete pre-processor and cannot replace
 * more sophisticated tools such as CPP, in the case where more complex
 * macros are added to the source files.
 *
 * Usage:
 *   Qfpp extends the Ant 'Task' class and thus can only be used within
 *   this framework. Within an Ant build script, you will first need to
 *   define the task for Ant within the <project> tag as follows:
 *
 *     <taskdef  name="qfpp"  classname="de.qfs.lib.qfpp"/>
 *
 *   When Ant executes the build script, it will immediately attempt to
 *   evaluate the class reference. Thus, the class must be in Ant's
 *   classpath (see the 'Install' notes below).
 *   The qfpp task can now be executed as any other Ant task. Here's a
 *   simple call to qfpp with a few of the optional properties set:
 *
 *     <qfpp jdk="12" junit="false">
 *       <fileset dir="." includes="**_/*.jpp"/>
 *     </qfpp>
 *
 *   (NOTE: the '_' character in "**_/*.jpp" should not be in your build
 *   script, it was added here just so that the Java comments are not
 *   disturbed)
 *   The nested <fileset> element allows you to specify the contents of
 *   directories and subdirectories to be processed. Our call here
 *   specifies that all files with the extension .jpp under the current
 *   directory will be passed to qfpp. For a more complete description
 *   of file sets, see the Ant documentation.
 *   The complete list of properties which can be used within the qfpp
 *   task call are as follows. Note that all properties are optional.
 *
 *    jdk="value"        Set the implementation JDK, either "11" for
 *                       version 1.1 or "12" for version 1.2 and higher.
 *                       (default = "12")
 *    optcondition="condition:value"
 *                       Set an optional condition. This property is
 *                       intended to provide flexibility for future source
 *                       code modifications. The ':' character represents
 *                       '=' (Ant does not allow an additional '=' in the
 *                       task call). Input is case-sensitive. For example:
 *                         optcondition="VERSION:1.5"
 *                       Would cause the following conditional statment to
 *                       be evaluated as true:
 *                         #if VERSION == 1.5
 *    junit="value"      Include unit-testing code if set to "true" or
 *                       "yes" (default = "false")
 *    optflag="value"    Set an optional flag to be scanned for during
 *                       preprocessing. This property is intended to provide
 *                       flexibility for future source code modifications.
 *                       Input is case-sensitive. For example:
 *                           optflag="QFLIB"
 *                       Would cause the following statement to be evaluated
 *                       as true:
 *                         #ifdef QFLIB
 *    extin="value"      Set the input file extension (default="jpp")
 *    extout="value"     Set the output file extension (default="java")
 *    filetype="value"   Set the generated output file type, either "unix"
 *                       "windows" - specify "default" to match the
 *                       current OS. (default = "default")
 *    fileset            See Ant documentation.
 *    verbose="value"    Verbose output if "true" (default = "false")
 *    debug="value"      Unsupported option, just displays some debug info
 *                       during runtime.
 *
 *
 * Build:
 *   Besides a few standard libaries, qfpp requires several classes out
 *   of $ANT_HOME/lib/ant.jar in order to extend the Task implementation.
 *   For example:
 *      javac -d '.' -classpath '$ANT_HOME/lib/ant.jar' qfpp.java
 *   Afterwards, you can create the jar file for this Ant task:
 *      jar cvf qfpp.jar './de/qfs/lib/qfpp.class'
 *
 * Install:
 *   In order to run Ant with the qfpp task, qfpp.jar needs to be in the
 *   classpath for Ant. The easiest solution is to place qfpp.jar in the
 *   $ANT_HOME/lib directory, to which Ant already has its classpath set.
 *
 * @author      M.Lehto
 */

// }}}
public class qfpp extends Task
{
    // {{{ variables

    /**
     *******************************************************************
     *******************************************************************
     *	Note: The default settings are attributes that are accessible
     *  to Ant via the public 'set' methods.
     *******************************************************************
     *******************************************************************
     */

    /**
     * Default JDK setting.
     */
    private static String    jdk        = "12";

    /**
     * Default JUnit (unit-testing) flag setting.
     */
    private static boolean   junit      = false;

    /**
     * Default generated file extension setting.
     */
    private static String    extout     = "java";

    /**
     * Default input file extension setting.
     */
    private static String    extin      = "jpp";

    /**
     * Default verbose output flag setting.
     */
    private static boolean   verbose    = false;

    /**
     * Default file-processing type setting.
     */
    private static String    filetype   = "default";

    /**
     * Default program debug setting
     */
    private static boolean   debug      = false;

    /**
     * The Vector used to store flags from pre-processed file.
     */
    private static Vector    flags      = new Vector();

    /**
     * The Hashtable used to store conditions from pre-processed file.
     */
    private static Hashtable conditions = new Hashtable();

    /**
     * The OS-specific EOF setting.
     */
    private static String    eof        = "\n";

    /**
     * The FileSet used as list of files to pre-process.
     */
    private static FileSet   fileset;


    /**
     * The following are constant-defined hash conditions and flags
     */
    public final static String _IF      = "if";
    public final static String _ELSE    = "else";
    public final static String _ELSEIF  = "else if";
    public final static String _ELIF    = "elif";
    public final static String _IFDEF   = "ifdef";
    public final static String _IFNDEF  = "ifndef";
    public final static String _ELIFDEF = "elifdef";
    public final static String _ELSEIFDEF="else ifdef";
    public final static String _ENDIF   = "endif";
    public final static String _WINDOWS = "windows";
    public final static String _WIN     = "win";
    public final static String _UNIX    = "unix";
    public final static String _DEFAULT = "default";
    public final static String _JDK     = "JDK";
    public final static String _JUNIT   = "JUNIT";

    /**
     * The following are constant-defined index types for marking conditions and flags.
     */
    public final static int    T_NONE   = -1;
    public final static int    T_IFDEF  = 10;
    public final static int    T_IFDEF_IF        = T_IFDEF + 1;
    public final static int    T_IFDEF_ELIFDEF   = T_IFDEF + 2;
    public final static int    T_IFDEF_ELSEIFDEF = T_IFDEF + 3;
    public final static int    T_IFCOND = 20;
    public final static int    T_IFCOND_IF       = T_IFCOND + 1;
    public final static int    T_IFCOND_ELIF     = T_IFCOND + 2;
    public final static int    T_IFCOND_ELSEIF   = T_IFCOND + 3;
    public final static int    T_IFNDEF = 30;
    public final static int    T_ELSE   = 40;
    public final static int    T_END    = 50;

    // }}}

    // {{{ constructor

    /**
     * No constructor.
     */

    // }}}

    //----------------------------------------------------------------------
    // Public interface for Ant.
    //----------------------------------------------------------------------
    // {{{ createFileSet

    /**
     * Interface method for Ant.
     * Instantiates the object for nested Ant "fileset" support,
     * which simply implements org.apache.tools.ant.types.FileSet.
     *
     * @return   FileSet     the instantiated object.
     */
    public FileSet createFileSet()
    {
	fileset = new FileSet();
	return fileset;
    }

    // }}}
    // {{{ setJdk

    /**
     * Interface method for Ant.
     * Set the JDK version used to run the pre-processor.
     *
     * @param    val    JDK version, in the form of "11" or "12."
     *                  Note that only JDK 1.2 is implemented, even
     *                  if a string greater than "12" is given
     */
    public void setJdk(String val)
    {
	if (val.equals("11")) {
	    jdk = val;
	}
	else {
	    jdk = "12";
	}
    }

    // }}}
    // {{{ setJunit

    /**
     * Interface method for Ant.
     * Turn on/off conditional JUnit code.
     *
     * @param    val    True to include JUnit conditional code.
     *                  Note that the implementation within the Ant
     *                  allows for a string of either "true" or "yes."
     */
    public void setJunit(boolean val)
    {
	junit = val;
    }

    // }}}
    // {{{ setOptflag

    /**
     * Interface method for Ant.
     * Set an additional pre-processing flag (see notes above).
     *
     * @param    val    Name of the flag to define.
     */
    public void setOptflag(String val)
    {
	flags.addElement(val);
    }

    // }}}
    // {{{ setOptcondition

    /**
     * Interface method for Ant.
     * Set an additional pre-processing condition (see notes above).
     *
     * @param    val    The condition in the form of: "cond:value"
     *                  Where the ':' represents essentially '='
     *                  (However, an additional '=' character is
     *                  not allowed within the Ant call.)
     *                  For example setting "VER:1.5" would allow
     *                  the following conditional statment to
     *                  evaluate to true:
     *                         #if VER == 1.5
     * @throws   IllegalArgumentException if "val" is improperly formed.
     */
    public void setOptcondition(String val) throws IllegalArgumentException
    {
	String key, value;
	int i = val.indexOf(":");
	if ( (i > 0) && (i < val.length() - 1) ) {
	    key = val.substring(0,i).trim();
	    value = val.substring(i+1).trim();
	    addCondition(key, value);
	}
	else {
	    throw new IllegalArgumentException
		("Optional condition syntax: " + val + " not in the form of "
		 + "\"condition:value\"");
	}
    }

    // }}}
    // {{{ setExtin

    /**
     * Interface method for Ant.
     * Set the input file extension type.
     *
     * @param    val    The file extension ('.' not required)
     */
    public void setExtin(String val)
    {
	if (val.indexOf(".") == 0) {
	    val = val.substring(1);
	}
	extin = val;
    }

    // }}}
    // {{{ setExtout

    /**
     * Interface method for Ant.
     * Set the output file extension type.
     *
     * @param    val         The file extension ('.' not required)
     */
    public void setExtout(String val)
    {
	if (val.indexOf(".") == 0) {
	    val = val.substring(1);
	}
	extout = val;
    }

    // }}}
    // {{{ setFiletype

    /**
     * Interface method for Ant.
     * Specify the output file format type.
     * Note that this task by default sets the file-type to the OS of
     * your system. However, for some users (such as Cygwin users on
     * Windows), a different file type is preferred.
     *
     * @param    val         The file type, valid values are:
     *                       "windows"  (0x0D,0x0A used as EOF)
     *                       "unix"     (0x0A used as EOF)
     *                       "default"  use the default for current OS
     * @throws   IllegalArgumentException if "val" is improperly formed.
     */
    public void setFiletype(String val) throws IllegalArgumentException
    {
	if (val.equalsIgnoreCase(_UNIX)) {
	    filetype = _UNIX;
	}
	else if ((val.equalsIgnoreCase(_WINDOWS)) ||
		 (val.equalsIgnoreCase(_WIN)) ) {
	    filetype = _WINDOWS;
	}
	else if (val.equalsIgnoreCase(_DEFAULT)) {
	    filetype = _DEFAULT;
	}
	else {
	    throw new IllegalArgumentException (val + " is not one of " +
			_UNIX + ", " + _WINDOWS + ", or " + _DEFAULT);
	}
    }

    // }}}
    // {{{ setVerbose

    /**
     * Interface method for Ant.
     * Turn on/off verbose output.
     *
     * @param    val         True to turn verbose output on.
     */
    public void setVerbose(boolean val)
    {
	verbose = val;
    }

    // }}}
    // {{{ setDebug

    /**
     * Interface method for Ant.
     * Turn on/off debug output. This is an *unsupported* option.
     *
     * @param    val         True to turn debug output on.
     */
    public void setDebug(boolean val)
    {
	debug = val;
    }

    // }}}

    //----------------------------------------------------------------------
    // Internal implementation
    //----------------------------------------------------------------------
    // {{{ processFile

    /**
     * Pre-process the indicated file.
     *
     * @param    inname      Input file name
     * @param    outname     Output file name
     * @throws   IOException If file problems arise
     */
    private void processFile(String inname, String outname)
	throws IOException
    {
	FileInputStream fin = new FileInputStream(inname);
	BufferedReader in
	    = new BufferedReader(new InputStreamReader(fin));
	FileOutputStream fout
	    = new FileOutputStream(outname);
	OutputStreamWriter out = new OutputStreamWriter(fout,"UTF-8");

	int hashlevel = 0;  // for nested conditional statements
	// note that the hashlevel is limited arbitrarily to 50.
	// if you nest deeper than this in your source code, you're insane.
	boolean[] write = new boolean[50];
	write[0] = true;
	boolean[] elsewrite = new boolean[50];
	elsewrite[0] = false;
	boolean neg = false;
	int i, type, len;
	int line = 0;
	String s, cmp, p, key, val;
	while ((s = in.readLine()) != null) {
	    line++;
	    cmp = s.trim();
	    if ( (cmp.startsWith("#")) && (cmp.length() > 1)) {
		cmp = cmp.substring(1).trim();
		// check for any comments within the comparison string
		i = cmp.indexOf("/");
		if (i > 0) {
		    cmp = cmp.substring(0,i).trim();
		}
		// now determine what kind of a condition it is
		if (cmp.startsWith(_IF)) {
		    hashlevel++;
		    write[hashlevel] = write[hashlevel-1];
		    elsewrite[hashlevel] = true;
		}
		type = getType(cmp);
		len = getTypeLength(type);
		if (debug) {
		    System.out.println
			("  {" + cmp + "} type:" + type + ":" + len);
		}
		if (type < T_ELSE) {
		    if ((write[hashlevel-1]) &&
			((type >= T_IFDEF) && (type < T_IFDEF + 10)) ) {
			// a flag in the form of: #ifdef <FLAG>
			p = cmp.substring(len + 1);
			if (hasFlag(p.trim())) {
			    write[hashlevel] = true;
			    elsewrite[hashlevel] = false;
			}
			else { write[hashlevel] = false; }
		    }
		    else if ((write[hashlevel-1]) &&
			     ((type >= T_IFNDEF) && (type < T_IFNDEF + 10))) {
			// a flag in the form of: #ifndef <FLAG>
			p = cmp.substring(len + 1);
			if ( ! hasFlag(p.trim())) {
			    write[hashlevel] = true;
			    elsewrite[hashlevel] = false;
			}
			else { write[hashlevel] = false; }
		    }
		    else if (write[hashlevel-1]) {
			// a condition
			p = cmp.substring(len + 1);
			if (p.indexOf("=") > 0) {
			    // it's a condition in form of
			    //   "condition == value" or "condtion != value"
			    neg = false;
			    i = p.indexOf("==");
			    if (i < 0) {         // check for != condition
				i = p.indexOf("!=");
				if (i >= 0) { neg = true; }
			    }
			    if ( (i > 0) && (i < (cmp.length() - 2)) ) {
				key = p.substring(0,i);
				val = p.substring(i+2);
				if (! neg) {
				    if (hasCondition(key.trim(),val.trim())) {
					write[hashlevel] = true;
					elsewrite[hashlevel] = false;
				    }
				    else { write[hashlevel] = false; }
				}
				else {
				    if (hasCondition(key.trim(),val.trim())) {
					write[hashlevel] = false;
				    }
				    else {
					write[hashlevel] = true;
					elsewrite[hashlevel] = false;
				    }
				}
			    }
			    else {
				throw new IOException("{" + inname +
   			        "[" + line + "]} badly formed condition: #"
				+ cmp);
			    }
			} // if (p.indexOf("=") > 0)
			else {
			    // check if it's a simple boolean condition
			    if ((p.equalsIgnoreCase("true")) ||
				(p.equals("1"))) {
				write[hashlevel] = true;
				elsewrite[hashlevel] = false;
			    }
			    else if ((p.equalsIgnoreCase("false")) ||
				     (p.equals("0"))) {
				write[hashlevel] = false;
			    }
			    else {
				// assume it's a flag
				if (hasFlag(p.trim())) {
				    write[hashlevel] = true;
				    elsewrite[hashlevel] = false;
				}
				else { write[hashlevel] = false; }
			    }
			}
		    }
		} // if (type < T_ELSE )
		else if (type == T_ELSE) {
		    if (hashlevel < 1) {
			throw new IOException("{" + inname +
			      "[" + line + "]} #else without #if: #" + cmp);
		    }
		    if ( (write[hashlevel-1]) && (elsewrite[hashlevel])) {
			write[hashlevel] = true;
		    }
		    else { write[hashlevel] = false; }
		}
		else if (type == T_END) {
		    elsewrite[hashlevel] = false;
		    write[hashlevel] = false;
		    hashlevel--;
		}
		else if (write[hashlevel]) {
		    // just write it
		    out.write(s + eof);
		}
	    } // if ( (cmp.startsWith("#")) && (cmp.length() > 1))
	    else if (write[hashlevel]) {
		out.write(s + eof);
	    }
	}

	in.close();
	out.close();
    }

    // }}}
    // {{{ hasCondition

    /**
     * Determine if we have the specified condition set
     *
     * @param    key  condition
     * @param    val  expected value for the condition
     * @return   true if we have the condition set to the specified value,
     *           false if not.
     */
    private boolean hasCondition(String key, String val)
    {
	if (conditions.size() > 0) {
	    String cmp = (String)conditions.get(key);
	    if (cmp != null ) {
		if (cmp.equals(val)) {
		    return true;
		}
	    }
	}
	return false;
    }
    // }}}
    // {{{ addCondition

    /**
     * Add a condition for pre-processing.
     *
     * @param    key  The condition (keyword)
     * @param    val  The condition value
     */
    private void addCondition(String key, String val)
    {
	conditions.put(key.trim(), val.trim());
    }

    // }}}
    // {{{ hasFlag

    /**
     * Determine if we have the specified flag set.
     *
     * @param    flag    the flag
     * @return   true    if the flag is set, false if not.
     */
    private boolean hasFlag(String flag)
    {
	String cmp;
	if ( ! flags.isEmpty()) {
	    return flags.contains(flag);
	}
	else { return false; }
    }

    // }}}
    // {{{ getType

    /**
     * Get hash-type of the given string.
     *
     * @param    s           input string to search
     * @return   the T_TYPE integer
     */
    private int getType(String s)
    {
	int type = T_NONE;
	if (s.startsWith(_IF)) {
	    if (s.startsWith(_IFDEF))  { type = T_IFDEF_IF; }
	    else if (s.startsWith(_IFNDEF)) { type = T_IFNDEF; }
	    else { type = T_IFCOND_IF; }
	}
	else if ( (s.startsWith(_ELSE)) || (s.startsWith(_ELIF)) ) {
	    if (s.startsWith(_ELSEIFDEF)) { type = T_IFDEF_ELSEIFDEF; }
	    else if (s.startsWith(_ELIFDEF)) { type = T_IFDEF_ELIFDEF; }
	    else if (s.startsWith(_ELSEIF)) { type = T_IFCOND_ELSEIF; }
	    else if (s.startsWith(_ELIF)) { type = T_IFCOND_ELIF; }
	    else { type = T_ELSE; }
	}
	else if (s.startsWith(_ENDIF)) {
	    type = T_END;
	}
	return type;
    }

    // }}}
    // {{{ getTypeLength

    /**
     * Get the length of given type.
     *
     * @param    type    the T_TYPE integer
     * @return   the string length of the hash-type
     */
    private int getTypeLength(int type)
    {
	int length = 0;
	switch(type) {
	case T_IFDEF_IF:         length = _IFDEF.length(); break;
	case T_IFDEF_ELIFDEF:    length = _ELIFDEF.length(); break;
	case T_IFDEF_ELSEIFDEF:  length = _ELSEIFDEF.length(); break;
	case T_IFCOND_IF:        length = _IF.length(); break;
	case T_IFCOND_ELIF:      length = _ELIF.length(); break;
	case T_IFCOND_ELSEIF:    length = _ELSEIF.length(); break;
	case T_IFNDEF:           length = _IFNDEF.length(); break;
	case T_ELSE:             length = _ELSE.length(); break;
	case T_END:              length = _ENDIF.length(); break;
	}
	return length;
    }

    // }}}
    // {{{ setProperties

    /**
     * Set default properties.
     */
    private void setProperties()
    {
	// Add flags and conditions for pre-processing.
	addCondition(_JDK,jdk);
	if (junit) {
	    flags.addElement(_JUNIT);
	}
	// set default EOF
	if (filetype.equals(_UNIX)) {
	    eof = "\n";
	}
	else if (filetype.equals(_WINDOWS)) {
	    eof = "\r\n";
	}
	else {
	    eof = System.getProperty("line.separator");
	}
    }

    // }}}
    // {{{ printProperties

    /**
     * Output settable properties.
     */
    private void printProperties()
    {
	System.out.println("conditions: ");
	String s;
	Enumeration e = conditions.keys();
	for (int i = 0; i < conditions.size(); i++) {
	    s = (String)e.nextElement();
	    System.out.println("  " + s + " = " + (String)conditions.get(s));
	}
	System.out.println("flags:");
	if (flags.size() == 0) {
	    System.out.println("  (no flags defined)");
	}
	else {
	    for (int i = 0; i < flags.size(); i++) {
		System.out.println("  " + (String)flags.elementAt(i));
	    }
	}

	System.out.println("filetype = " + filetype);
	System.out.println("extin = " + extin);
	System.out.println("extout = " + extout);
    }

    // }}}
    // {{{ execute

    /**
     * This is the method called by Ant to implement the task.
     * Note that Ant will first call any indicated 'set' methods
     * before 'execute.'
     * For example, if your Ant task call looks like this:
     *    <qfpp junit="true" jdk="12"/>
     * Then Ant will first call the 'setJunit' and 'setJdk' methods
     * before calling 'execute.'
     */
    public void execute()
    {
	setProperties();
	if (verbose) { printProperties(); }

	String file, inname, outname;
	// parse each file
	try {
	    DirectoryScanner ds =
		fileset.getDirectoryScanner(fileset.getProject());
	    String[] files = ds.getIncludedFiles();

	    System.out.println("Pre-processing " + files.length
			       + " files under " +
		fileset.getDir(fileset.getProject()).getAbsolutePath());

	    for (int i = 0; i < files.length; i++) {
		if (files[i].endsWith("." + extin)) {
		    outname = files[i].substring
			(0, files[i].length() - 4) + "." + extout;
		    inname = files[i];
		    if (verbose) {
			System.out.println((i+1) + ". "
			       + inname + " > " + outname);
		    }
		    processFile(inname,outname);
		}
		else {
		    if (verbose) {
			System.err.println(files[i] +
			 " ignored, does not match input extension ."
			 + extin);
		    }
		}
	    }
	}
	catch (IOException e) {
	    System.err.println(e);
	}
    }

    // }}}
}
