// {{{ copyright

/********************************************************************
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Component;
import java.awt.Container;

import javax.swing.JMenu;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.tree.IndexBasedTreeAdapter;

// }}}

/**
 * This is a TreeAdapter for the AWT Component hierarchy. It assumes that all
 * nodes passed are Components. <p>
 *
 * Special care is taken to handle JMenus. These have a logical hierarchy,
 * meaning that JMenuItems are permanent children of a JMenu and a physical
 * hierarchy in which the Items are added to the menu only while it is popped
 * up. <p>
 *
 * By default the logical menu structure is assumed. If you want to traverse
 * the component tree's physical structure, use the {@link
 * #AWTTreeAdapter(boolean) AWTTreeAdapter(boolean)} constructor with a false
 * argument. <p>
 *
 * Warning: The logical structure is inconsistent in that the parent menu of a
 * MenuItem cannot be determined if the menu is not posted. This doesn't
 * prevent traversing the hierarchy though.
 *
 * @author      Gregor Schmid
 * @since       0.98.1
 */
public class AWTTreeAdapter
    extends IndexBasedTreeAdapter
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.gui.AWTTreeAdapter");

    /**
     * Whether to treat JMenus specially.
     */
    private boolean specialJMenu = true;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ AWTTreeAdapter()

    /**
     * Create a new AWTTreeAdapter.
     */
    public AWTTreeAdapter ()
    {
    }

    // }}}
    // {{{ AWTTreeAdapter(boolean)

    /**
     * Create a new AWTTreeAdapter.
     *
     * @param   specialJMenu    Whether to treat JMenus specially (see class
     *                          documentation). Default is true.
     *
     * @since   0.99.0
     */
    public AWTTreeAdapter (boolean specialJMenu)
    {
        this.specialJMenu = specialJMenu;
    }

    // }}}

    //----------------------------------------------------------------------
    // TreeAdapter interface
    //----------------------------------------------------------------------
    // {{{ getParentNode

    /**
     * Get the parent of a node.
     *
     * @param   child   The node to query for the parent.
     *
     * @return  The parent of the node.
     */
    public Object getParentNode(Object child)
    {
        return ((Component) child).getParent();
    }

    // }}}
    // {{{ isLeaf

    /**
     * Query whether a node is a leaf.
     *
     * @param   node    The node to query.
     *
     * @return  True if the node is a leaf.
     */
    public boolean isLeaf(Object node)
    {
        return ! (node instanceof Container);
    }

    // }}}

    //----------------------------------------------------------------------
    // Index based
    //----------------------------------------------------------------------
    // {{{ getChildCount

    /**
     * Get the number of children of a node.
     *
     * @param   parent  The node to query.
     *
     * @return  The number of children of the node.
     */
    public int getChildCount(Object parent)
    {
        if (specialJMenu
            && (parent instanceof JMenu)) {
            return ((JMenu) parent).getMenuComponentCount();
        }
        if (parent instanceof Container) {
            return ((Container) parent).getComponentCount();
        }
        return 0;
    }

    // }}}
    // {{{ getChildNode

    /**
     * Get a child of a node.
     *
     * @param   parent  The node to query.
     * @param   index   The index of the child.
     *
     * @return  The child at the index.
     *
     * @throws   IndexOutOfBoundsException If the index is invalid.
     */
    public Object getChildNode(Object parent, int index)
         throws IndexOutOfBoundsException
    {
        if (specialJMenu
            && (parent instanceof JMenu)) {
            return ((JMenu) parent).getMenuComponents()[index];
        }
        if (! (parent instanceof Container)) {
            throw new IndexOutOfBoundsException
                ("parent " + parent + " is not a Container");
        }
        return ((Container) parent).getComponents()[index];
    }

    // }}}
    // {{{ getIndexOfChildNode

    /**
     * Get the index of a child in a node's list of children.
     *
     * @param   parent  The node to query.
     * @param   child   The child to look for.
     *
     * @return  The index of the child or -1 if it is not a child of this node.
     */
    public int getIndexOfChildNode(Object parent, Object child)
    {
        if (specialJMenu
            && (parent instanceof JMenu)) {
            Component[] children = ((JMenu) parent).getMenuComponents();
            for (int i = 0; i < children.length; i++) {
                if (children[i].equals(child)) {
                    return i;
                }
            }
        }
        if (parent instanceof Container) {
            Component[] children = ((Container) parent).getComponents();
            for (int i = 0; i < children.length; i++) {
                if (children[i].equals(child)) {
                    return i;
                }
            }
        }
        return -1;
    }

    // }}}

}
