// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.lang.reflect.InvocationTargetException;

import java.awt.ActiveEvent;
import java.awt.im.InputContext;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Label;
import java.awt.MenuComponent;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.PaintEvent;
import java.awt.event.WindowEvent;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.Action;
import javax.swing.KeyStroke;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JPopupMenu;
import javax.swing.JPasswordField;
import javax.swing.JRootPane;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.JViewport;
import javax.swing.Scrollable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.text.JTextComponent;
import javax.swing.text.Keymap;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeModel;

import de.qfs.lib.gui.AWTTreeAdapter;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.tree.TreeUtil;
import de.qfs.lib.util.Misc;
import de.qfs.lib.util.Reflector;

// }}}

/**
 * A collection of utilities and Swing problem workarounds that go beyond
 * {@link javax.swing.SwingUtilities SwingUtilities}.
 *
 * @author      Gregor Schmid
 */
public abstract class SwingUtil
{
    // {{{ variables

    /**
     * Identifier for no events.
     */
    public final static int NO_EVENTS = 0;

    /**
     * Identifier for all events.
     */
    public final static int ALL_EVENTS = ~0;

    /**
     * Identifier for paint events.
     */
    public final static int PAINT_EVENTS = 1;

    /**
     * Identifier for input events.
     */
    public final static int INPUT_EVENTS = 2;

    /**
     * Identifier for the rest of the events. Must include all unused types
     * for compatibility with possible future extensions.
     */
    public final static int OTHER_EVENTS = ~3;

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger ("de.qfs.lib.gui.SwingUtil");

    /**
     * Keymap with the original binding for enter.
     */
    private static Keymap defaultKeymap;

    /**
     * ID counter for DoneEvent ids.
     */
    private static int doneID;

    // }}}

    //----------------------------------------------------------------------
    // EventLoop
    //----------------------------------------------------------------------
    // {{{ invokeAndWait

  /**
   * Invoke a thread inside the swing EventLoop and wait for it to finish.
   *
   * @param     runnable The Runnable to invoke.
   *
   * @return    True if everything was OK, false, if the current Thread was
   *            interrupted while waititing for the EventLoop.
   */
    public static boolean invokeAndWait (Runnable runnable)
    {
        if (SwingUtilities.isEventDispatchThread()) {
            runnable.run();
        } else {
            try {
                SwingUtilities.invokeAndWait(runnable);
            } catch (InvocationTargetException ex) {
                if (ex.getTargetException() instanceof RuntimeException) {
                    throw (RuntimeException) ex.getTargetException();
                } else {
                    // impossible ?!?
                    ex.getTargetException().printStackTrace();
                }
            } catch (InterruptedException ex) {
                return false;
            }
        }
        return true;
    }

    // }}}
    // {{{ interface UnsafeRunnable

    /**
     * This interface is similar to Runnable, except that the run method
     * may return a value and throw any kind of Exception.
     */
    public static interface UnsafeRunnable
    {
        /**
         * Execute the unsafe runnable.
         *
         * @return      Whatever the UnsafeRunnable decides to return.
         *
         * @throws      Exception       If anything goes wrong.
         */
        public Object run() throws Exception;
    }

    // }}}
    // {{{ withEvents

    /**
     * Execute some code in a background Thread, while continuing to handle
     * AWT Events from the current Thread. InputEvents like mouse or keyboard
     * input can optionally be discarded (useful together with a wait cursor).
     * This works only if called from an AWT EventLoop Thread. When called from
     * some other Thread, the Runnable simply gets executed directly. In this
     * case there is no way to block input events.
     *
     * @param   runnable        The UnsafeRunnable to execute.
     * @param   noInput         If true, InputEvents will be discarded.
     *
     * @return  The result from runnable's execution.
     *
     * @throws  RuntimeException        If runnable's run method
     *          throws a RuntimeException, it is ismply passed on
     * @throws  InvocationTargetException       If runnable's run method*
     *          throws an Exception.
     *
     * @deprecated      Please use
     *                  {@link #withEvents(SwingUtil.UnsafeRunnable,int,int)
     *                  withEvents} instead.
     */
    public static Object withEvents(final UnsafeRunnable runnable,
                                    final boolean noInput)
        throws InvocationTargetException
    {
        if (noInput) {
            return withEvents(runnable, PAINT_EVENTS | OTHER_EVENTS,
                              INPUT_EVENTS);
        } else {
            return withEvents(runnable, ALL_EVENTS, NO_EVENTS);
        }
    }

    /**
     * Execute some code in a background thread, while continuing to handle
     * AWT events from the current tread. <p>
     *
     * This method distinguishes between three kinds of events identified by
     * <dl>
     * <dt>PAINT_EVENTS</dt><dd>All kinds of paint events.</dd>
     * <dt>INPUT_EVENTS</dt><dd>All kind of input events, e.g. keyboard or
     * mouse events.</dd>
     * <dt>OTHER_EVENTS</dt><dd>All the rest.</dd>
     * </dl>
     * For each of these kinds of events, there are three possible ways to
     * handle them:
     * <dl>
     * <dt>dispatch</dt><dd>Dispatch the event immediately as it arrives. This
     * is the default for all events.</dd>
     * <dt>delay</dt><dd>Delay the event until the background thread is
     * finished. Useful if handling the event could interfere with the
     * background thread.</dd>
     * <dt>discard</dt>
     * <dd> Discard the event as if it never happended. Makes sense for input
     * events if a wait cursor is being displayed while the background thread
     * is active.</dd>
     * </dl>
     *
     * This works only if called from an AWT EventLoop Thread. When called from
     * some other Thread, the Runnable simply gets executed directly. In this
     * case there is no way to block or delay events. <p>
     *
     * This method is more or less reentrant, meaning that recusive calls don't
     * fail horribly, but they can lead to confusion since the last call will
     * determine the delayed and discarded event types for all calls.
     *
     * @param   runnable        The UnsafeRunnable to execute.
     * @param   delay           Binary OR of event types that should be
     *                          delayed.
     * @param   discard         Binary OR of event types that should be
     *                          discarded.
     *
     * @return  The result from runnable's execution.
     *
     * @throws  RuntimeException        If runnable's run method
     *          throws a RuntimeException, it is ismply passed on
     * @throws  InvocationTargetException       If runnable's run method*
     *          throws an Exception.
     */
    public static Object withEvents(final UnsafeRunnable runnable,
                                    int delay, int discard)
        throws InvocationTargetException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "withEvents(UnsafeRunnable,int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "runnable: " + runnable + ", " +
                       "delay: " + delay + ", " +
                       "discard: " + discard);
        }

        // I hate this hack...
        final Exception[] exHolder = new Exception[1];
        final Object[] oHolder = new Object[1];
        final EventLock lock = new EventLock ();
        lock.setLocked(true);
        new Thread ("SwingUtil-withEventsThread") {
            public void run()
            {
                try {
                    oHolder[0] = runnable.run();
                } catch (Exception ex) {
                    exHolder[0] = ex;
                }

                Toolkit.getDefaultToolkit().getSystemEventQueue()
                    .postEvent(new DoneEvent(lock));
            }
        }.start();
        poll(delay, discard, lock);

        if (exHolder[0] != null) {
            if (exHolder[0] instanceof RuntimeException) {
                throw (RuntimeException) exHolder[0];
            }
            throw new InvocationTargetException(exHolder[0]);
        }
        return oHolder[0];
    }
    // }}}
    // {{{ poll

    /**
     * Helper method for polling and dispatching AWT events.
     *
     * @param   delay   Flags for events to delay.
     * @param   discard Flags for events to discard.
     * @param   lock    The lock that determines when polling is finished.
     */
    private static void poll(int delay, int discard, EventLock lock)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "poll(int,int,EventLock)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "delay: " + delay + ", " +
                       "discard: " + discard + ", " +
                       "lock: " + lock);
        }
        java.awt.EventQueue queue = EventQueue.instance();
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "poll(int,int,EventLock)",
                       "queue: " + queue);
        }

        boolean done = false;
        List eventStack = new ArrayList ();
        while (!done) {
            if (SwingUtilities.isEventDispatchThread()) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "poll(int,int,EventLock)",
                               "getting next event");
                }
                try {
                    AWTEvent event = queue.getNextEvent();
                    AWTEvent orig = event;
                    if (event instanceof EventQueue.WrapperEvent) {
                        event = ((EventQueue.WrapperEvent) event)
                            .getWrappedEvent();
                    }
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "poll(int,int,EventLock)",
                                   "event: " + event);
                    }
                    if (event instanceof DoneEvent) {
                        doDispatch(orig, event);
                    } else if (event instanceof InputEvent) {
                        if ((discard & INPUT_EVENTS) != 0) {
                            ((InputEvent) event).consume();
                        } else if ((delay & INPUT_EVENTS) != 0) {
                            eventStack.add(event);
                        } else {
                            try {
                                doDispatch(orig, event);
                            } catch (RuntimeException ex) {
                                if (logger.level >= Log.ERR) {
                                    logger.log("poll(int,int,EventLock)", ex);
                                }
                            }
                        }
                    } else if (event instanceof PaintEvent) {
                        if ((discard & PAINT_EVENTS) != 0) {
                            // ignore
                        } else if ((delay & PAINT_EVENTS) != 0) {
                            eventStack.add(event);
                        } else {
                            try {
                                doDispatch(orig, event);
                            } catch (RuntimeException ex) {
                                if (logger.level >= Log.ERR) {
                                    logger.log("poll(int,int,EventLock)", ex);
                                }
                            }
                        }
                    } else {
                        if ((discard & OTHER_EVENTS) == OTHER_EVENTS) {
                            // ignore
                        } else if ((delay & OTHER_EVENTS) == OTHER_EVENTS) {
                            eventStack.add(event);
                        } else {
                            try {
                                doDispatch(orig, event);
                            } catch (RuntimeException ex) {
                                if (logger.level >= Log.ERR) {
                                    logger.log("poll(int,int,EventLock)", ex);
                                }
                            }
                        }
                    }
                } catch (InterruptedException ex) {
                }
            } else {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "poll(int,int,EventLock)",
                               "waiting for lock");
                }
                synchronized(lock) {
                    try {
                        lock.wait(100);
                    } catch (InterruptedException ex) {
                    }
                }
            }
            done = ! lock.isLocked();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "poll(int,int,EventLock)",
                           "done: " + done);
            }
        }
        // put the other DoneEvents back on the Queue
        if (eventStack.size() >0) {
            for (Iterator i = eventStack.iterator(); i.hasNext();) {
                AWTEvent pevent = (AWTEvent) i.next();
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "poll(int,int,EventLock)",
                               "pevent: " + pevent);
                }
                queue.postEvent(pevent);
            }
        }
    }

    // }}}
    // {{{ doDispatch

    /**
     * Helper method to dispatch one event.
     *
     * @param   orig    The original event.
     * @param   event   The same as orig or the true event, if orig is a
     *                  WrapperEvent.
     */
    private static void doDispatch(AWTEvent orig, AWTEvent event)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "doDispatch(AWTEvent,AWTEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "orig: " + orig + ", " +
                       "event: " + event);
        }
        if (orig instanceof EventQueue.WrapperEvent) {
            ((EventQueue.WrapperEvent) orig).dispatch();
        } else if (EventQueue.isInstalled()) {
            ((EventQueue) EventQueue.instance()).dispatchEvent(event);
        } else {
            Object src = event.getSource();
            if (event instanceof ActiveEvent) {
                ((ActiveEvent)event).dispatch();
            } else if (src instanceof Component) {
                ((Component)src).dispatchEvent(event);
            } else if (src instanceof MenuComponent) {
                ((MenuComponent)src).dispatchEvent(event);
            }
        }
    }


    // }}}
    // {{{ class DoneEvent

    /**
     * Helper Event that signals the end of AWT Event polling.
     */
    private static class DoneEvent
        extends AWTEvent
        implements ActiveEvent
    {
        /**
         * Dummy source for the event.
         */
        private static Label dummy = new Label();

        /**
         * The lock for the event.
         */
        private EventLock lock;

        /**
         * Create a new DoneEvent.
         *
         * @param       lock      The event lock.
         */
        public DoneEvent(EventLock lock)
        {
            super(dummy, AWTEvent.RESERVED_ID_MAX);
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "DoneEvent(EventLock)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "lock: " + lock);
            }
            this.lock = lock;
        }

        /**
         * Get a String representation for the DoneEvent.
         */
        public String toString()
        {
            return "DoneEvent[lock=" + lock + "]";
        }

        public void dispatch()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "dispatch()", "");
            }
            lock.setLocked(false);
        }
    }

    // }}}
    // {{{ class EventLock

    private static class EventLock
    {
        // {{{ Variables

        /**
         * The locked state.
         */
        private boolean locked;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ EventLock

        /**
         * Create a new EventLock.
         */
        public EventLock ()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "EventLock()", "");
            }
        }

        // }}}

        // {{{ isLocked

        /**
         * Get the locked state of the EventLock.
         *
         * @return      The locked state of the EventLock.
         */
        public final synchronized boolean isLocked()
        {
            return locked;
        }

        // }}}
        // {{{ setLocked

        /**
         * Set the locked state of the EventLock.
         *
         * @param       locked  The locked state to set.
         */
        public final synchronized void setLocked(boolean locked)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "setLocked(boolean)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "locked: " + locked);
            }
            this.locked = locked;
            notifyAll();
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Keymaps
    //----------------------------------------------------------------------
    // {{{ prepareKeymaps

    /**
     * Prepare Keymaps of JTextField and JPasswordField, so that the
     * return key invokes a possible default button. This can be undone for
     * individual TextComponents with {@link #setDefaultEnterBinding
     * setDefaultEnterBinding}.
     */
    public static void prepareKeymaps()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "prepareKeymaps()", "");
        }
        if (defaultKeymap == null) {
            JTextField tf = new JTextField ();
            Keymap tfKeymap = tf.getKeymap();
            KeyStroke enter = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
            Action defaultEnterBinding = tfKeymap.getAction(enter);
            if (defaultEnterBinding != null) {
                tfKeymap.removeKeyStrokeBinding(enter);
                new JPasswordField ().getKeymap()
                    .removeKeyStrokeBinding(enter);
                defaultKeymap = tf.addKeymap("old", tfKeymap);
                defaultKeymap.addActionForKeyStroke
                    (enter, defaultEnterBinding);
            }
        }
    }

    // }}}
    // {{{ setDefaultEnterBinding

    /**
     * Restore the default binding for the return key, if it was changed
     * by {@link #prepareKeymaps prepareKeymaps}.
     *
     * @param   text    The TextComponent for which to restore the binding.
     */
    public static void setDefaultEnterBinding(JTextComponent text)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setDefaultEnterBinding(JTextComponent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "text: " + text);
        }
        text.setKeymap(defaultKeymap);
    }

    // }}}

    //----------------------------------------------------------------------
    // Close windows on escape
    //----------------------------------------------------------------------
    // {{{ closeOnEscape

    /**
     * Add a KeyboardListener to a Window that posts a WINDOW_CLOSING event
     * on it when Escape is pressed. Useful for modal dialogs.
     *
     * @param   win
     */
    public static void closeOnEscape(final Window win)
    {
        win.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == e.VK_ESCAPE && e.getModifiers() == 0) {
                    Toolkit.getDefaultToolkit().getSystemEventQueue()
                        .postEvent(new WindowEvent
                                   (win,
                                    WindowEvent.WINDOW_CLOSING));
                    e.consume();
                }
            }
        });
        // Add an additional binding for the root pane. Appears to be
        // necessary with JDK 1.4
        JRootPane root = null;
        if (win instanceof JFrame) {
            root = ((JFrame) win).getRootPane();
        }
        if (win instanceof JDialog) {
            root = ((JDialog) win).getRootPane();
        }
        if (root != null) {
            ActionListener al = new ActionListener () {
                public void actionPerformed(ActionEvent e) {
                    Toolkit.getDefaultToolkit().getSystemEventQueue()
                        .postEvent(new WindowEvent
                                   (win,
                                    WindowEvent.WINDOW_CLOSING));
                }
            };
            root.registerKeyboardAction
                (al, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
                 root.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tree Scroll bugfix
    //----------------------------------------------------------------------
    // {{{ constrainScroll

    /**
     * Constrain the scrolling range of a {@link javax.swing.JScrollPane
     * JScrollPane} so that it never scrolls to a negative view position. This
     * helps fix a bug in the Swing 1.1.1 action for PageDown in a {@link
     * javax.swing.JTree JTree} that places the tree at the bottom of the
     * JScrollPane if its size is less than the visible area. <p>
     *
     * This method replaces the JScrollPane's {@link javax.swing.JViewport
     * JViewport} with one that checks the argument to {@link
     * javax.swing.JViewport#setViewPosition setViewPosition}.
     *
     * @param   scrollPane      The JScrollPane to adapt.
     */
    public static void constrainScroll(JScrollPane scrollPane)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "constrainScroll(JScrollPane)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "scrollPane: " + scrollPane);
        }

        ConstrainedViewport cv = new ConstrainedViewport ();
        JViewport vp = scrollPane.getViewport();
        Point pos = vp.getViewPosition();
        cv.setView(vp.getView());
        cv.setViewPosition(pos);
        scrollPane.setViewport(cv);
    }

    // }}}
    // {{{ class ConstrainedViewport

    /**
     * This class extends JViewport to restrict the arguments passed to
     * {@link JViewport#setViewPosition setViewPosition} to nonnegative
     * coordinates.
     */
    public static class ConstrainedViewport extends JViewport
    {
        /**
         * Create a new ConstrainedViewport.
         */
        public ConstrainedViewport ()
        {
        }

        /**
         * Sets the view coordinates that appear in the upper left
         * hand corner of the viewport, does nothing if there's no view.
         *
         * @param p  a Point object giving the upper left coordinates
         */
        public void setViewPosition(Point p)
        {
            Point p2 = new Point ( p.x < 0 ? 0 : p.x, p.y < 0 ? 0: p.y);
            super.setViewPosition(p2);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Popup helper
    //----------------------------------------------------------------------
    // {{{ showPopup

    /**
     * Bring up a JPopupMenu but make sure that it is fully visible on the
     * screen (if at all possible). Optionally it can be constrained to its
     * parent Window as well.
     *
     * @param   menu    The JPopupMenu to show.
     * @param   parent  The parent Component of the menu.
     * @param   x       The targeted x coordinate.
     * @param   y       The targeted y coordinate.
     * @param   constrain Whether the menu should be constrained to its parent
     *                  window.
     */
    public static void showPopup(JPopupMenu menu, Component parent,
                                 int x, int y, boolean constrain)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "showPopup(JPopupMenu,Component,int,int,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "menu: " + menu + ", " +
                       "parent: " + parent + ", " +
                       "x: " + x + ", " +
                       "y: " + y + ", " +
                       "constrain: " + constrain);
        }

        if (parent == null) {
            // not our job
            menu.show(parent, x, y);
            return;
        }

        menu.pack();
        Dimension size = menu.getPreferredSize();
        Window win;
        Point pos;
        if (parent instanceof Window) {
            win =  (Window) parent;
            pos = new Point (x, y);
        } else {
            win = SwingUtilities.windowForComponent(parent);
            pos = SwingUtilities.convertPoint(parent, x, y, win);
            x = pos.x;
            y = pos.y;
        }
        Point wPos = win.getLocation();

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "x: " + x + ", y: " + y + ", wPos: " + wPos +
                       ", size: " + size);
        }
        if (constrain) {
            Dimension wSize = win.getSize();

            if (x + size.width > wSize.width) {
                x = wSize.width - size.width;
            }
            if (x < 0) {
                x = 0;
            }
            if (y + size.height > wSize.height) {
                y = wSize.height - size.height;
            }
            if (y < 0) {
                y = 0;
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG,
                           "showPopup(JPopupMenu,Component,int,int,boolean)",
                           "Window constrain x: " + x + ", y: " + y);
            }
        }
        Dimension sSize = Toolkit.getDefaultToolkit().getScreenSize();
        x += wPos.x;
        y += wPos.y;
        if (x + size.width > sSize.width) {
            x = sSize.width - size.width;
        }
        if (x < 0) {
            x = 0;
        }
        if (y + size.height > sSize.height) {
            y = sSize.height - size.height;
        }
        if (y < 0) {
            y = 0;
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "Screen constrain x: " + x + ", y: " + y);
        }
        pos = SwingUtilities.convertPoint(win, x - wPos.x, y -wPos.y, parent);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG,
                       "showPopup(JPopupMenu,Component,int,int,boolean)",
                       "Converted back x: " + pos.x + ", y: " + pos.y);
        }
        menu.show(parent, pos.x, pos.y);
    }

    // }}}

    //----------------------------------------------------------------------
    // Scrolling
    //----------------------------------------------------------------------
    // {{{ scrollUpLine

    /**
     * Scroll the contents of a scrollPane up by one line if possible. Note
     * that this action typically corresponds to pressing the DOWN arrow key.
     *
     * @param   scrollPane      The scroll pane to scroll.
     */
    public static void scrollUpLine(JScrollPane scrollPane)
    {
        JViewport vp = scrollPane.getViewport();
        Component view = vp == null ? null : vp.getView();
        if (! (view instanceof Scrollable)) {
            return;
        }

        Rectangle visRect = vp.getViewRect();
        Dimension vSize = view.getSize();
        int amount = ((Scrollable)view).getScrollableUnitIncrement
            (visRect, SwingConstants.VERTICAL, -1);
        visRect.y += amount;
        if (visRect.y + visRect.height > vSize.height) {
            visRect.y = Math.max(0, vSize.height - visRect.height);
        }
        vp.setViewPosition(visRect.getLocation());
    }

    // }}}
    // {{{ scrollDownLine

    /**
     * Scroll the contents of a scrollPane down by one line if possible. Note
     * that this action typically corresponds to pressing the UP arrow key.
     *
     * @param   scrollPane      The scroll pane to scroll.
     */
    public static void scrollDownLine(JScrollPane scrollPane)
    {
        JViewport vp = scrollPane.getViewport();
        Component view = vp == null ? null : vp.getView();
        if (! (view instanceof Scrollable)) {
            return;
        }

        Rectangle visRect = vp.getViewRect();
        int amount = ((Scrollable)view).getScrollableUnitIncrement
            (visRect, SwingConstants.VERTICAL, -1);
        visRect.y -= amount;
        if (visRect.y < 0) {
            visRect.y = 0;
        }
        vp.setViewPosition(visRect.getLocation());
    }

    // }}}

    //----------------------------------------------------------------------
    // Focus
    //----------------------------------------------------------------------
    // {{{ disableScrollBarFocus

    /**
     * Traverse a component hierarchy and make all JScrollBars non-focusable.
     * Ignored unless run with JDK 1.4 and above.
     *
     * @param   com     The root of the hierarchy.
     *
     * @since   0.99.0
     */
    public static void disableScrollBarFocus(Component com)
    {
        if (Misc.isJavaVersionAtLeast(1, 4, 0)) {
            TreeUtil.visitTree
                (new AWTTreeAdapter (), new FocusDisabler(), com);
        }
    }

    // }}}
    // {{{ focusLater

    /**
     * A delayed version of the {@link java.awt.Component#requestFocus
     * requestFocus} method, a workaround for some weird JDK 1.4.1/Motif focus
     * behavior that messes up focus requests after closing a modal dialog.
     *
     * @param   com     The component to request the focus for.
     *
     * @since   0.99.0
     */
    public static void focusLater(final Component com)
    {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                com.requestFocus();
            }
        });
    }


    // }}}

    //----------------------------------------------------------------------
    // Cleanup
    //----------------------------------------------------------------------
    // {{{ cleanup

    /**
     * Break up a Component hierarchy to increase chances for garbage
     * collection. Swing has quite a few static caches that can delay or
     * even prevent garbage collection. This method tries to remove as many
     * references as possible.
     *
     * Also implements a workaround fow a problem with some JDKs that leaves
     * active InputContexts around that prevent a window from being disposed.
     *
     * @param   c       The root of the hierarchy to break up.
     */
    public static void cleanup(Component c)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "cleanup(Component)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "c: " + c);
        }

        Component [] children = null;
        if (c instanceof JMenu) {
            children = ((JMenu) c).getMenuComponents();
        } else
        if (c instanceof Container) {
            children = ((Container) c).getComponents();
        }
        if (children != null) {
            for (int i = 0; i < children.length; i++) {
                cleanup(children[i]);
                ((Container) c).remove(children[i]);
            }
        }
        if (c instanceof JMenu) {
            ((JMenu) c).removeAll();
        }
        if (c.getParent() != null) {
            c.getParent().remove(c);
        }
    }

    // }}}
    /**
     * There's a problem: The AWT keeps around all the owned dialogs and
     * disposes these when the frame is disposed. For reasons unknown some may
     * still have an active InputContext which can cause an
     * IllegalStateException, so we have to "deactivate" these contexts.
     *
     * For JDK 1.1 this is a NOOP.
     *
     * @param   win     The window whose InputContexts to clear.
     *
     * @since   0.99.0
     */
    public static void deactivateInputContexts(Window win)
    {
        InputContext ic = win.getInputContext();
        if (ic != null) {
            Object current =
                Reflector.safeGet(ic, "currentClientComponent", true);
            if (current != null) {
                ic.removeNotify((Component) current);
            }
        }
        Window[] wins = win.getOwnedWindows();
        for (int i = 0; wins != null && i < wins.length; i++) {
            ic = wins[i].getInputContext();
            if (ic != null) {
                Object current =
                    Reflector.safeGet(ic, "currentClientComponent", true);
                if (current != null) {
                    ic.removeNotify((Component) current);
                }
            }
        }
    }

    //----------------------------------------------------------------------
    // Tree expand/collapse all
    //----------------------------------------------------------------------
    // {{{ expandAll

    /**
     * Expand a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to expand.
     * @param   path    Path to the node to start at.
     * @param   limit   Maximum number of nodes to expand.
     *
     * @return  True if everything was fine, false if the limit was exceeded
     * and the operation aborted.
     */
    public static boolean expandAll(JTree tree, TreePath path, int limit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "expandAll(JTree,TreePath,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path + ", " +
                       "limit: " + limit);
        }

        return _expandAll(tree, path, new int[] {limit});
    }

    // }}}
    // {{{ collapseAll

    /**
     * Collapse a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to collapse.
     * @param   path    Path to the node to start at.
     * @param   limit   Maximum number of nodes to expand.
     *
     * @return  True if everything was fine, false if the limit was exceeded
     * and the operation aborted.
     */
    public static boolean collapseAll(JTree tree, TreePath path, int limit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "expandAll(JTree,TreePath)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path);
        }

        return _collapseAll(tree, path, new int[] {limit});
    }

    // }}}
    // {{{ _expandAll

    /**
     * Expand a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to expand.
     * @param   path    Path to the node to start at.
     * @param   limit   Maximum number of nodes to expand.
     *
     * @return  True if everything was fine, false if the limit was exceeded
     * and the operation aborted.
     */
    private static boolean _expandAll(JTree tree, TreePath path, int[] limit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "expandAll(JTree,TreePath,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path + ", " +
                       "limit: " + limit);
        }

        Object node = path.getLastPathComponent();
        TreeModel model = tree.getModel();
        if (model.isLeaf(node)) {
            return true;
        }

        tree.expandPath(path);
        if (--limit[0] <= 0) {
            return false;
        }
        int num = model.getChildCount(node);
        for (int i = 0; i < num; i++) {
            if (! _expandAll(tree, path.pathByAddingChild
                            (model.getChild(node, i)), limit)) {
                return false;
            }
        }
        return true;
    }

    // }}}
    // {{{ _collapseAll

    /**
     * Collapse a tree node and all its child nodes recursively.
     *
     * @param   tree    The tree whose nodes to collapse.
     * @param   path    Path to the node to start at.
     * @param   limit   Maximum number of nodes to expand.
     *
     * @return  True if everything was fine, false if the limit was exceeded
     * and the operation aborted.
     */
    private static boolean _collapseAll(JTree tree, TreePath path, int[] limit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "_collapseAll(JTree,TreePath,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "tree: " + tree + ", " +
                       "path: " + path + ", " +
                       "limit: " + limit);
        }

        Object node = path.getLastPathComponent();
        TreeModel model = tree.getModel();
        if (model.isLeaf(node)) {
            return true;
        }
        int num = model.getChildCount(node);
        for (int i = 0; i < num; i++) {
            if (! _collapseAll(tree, path.pathByAddingChild
                              (model.getChild(node, i)), limit)) {
                return false;
            }
        }
        tree.collapsePath(path);
        return --limit[0] > 0;
    }

    // }}}

    //----------------------------------------------------------------------
    // Auto size table column
    //----------------------------------------------------------------------
    // {{{ autoSizeTableColumn

    /**
     * Resize a table column so its fits exactly the broadest of its visible
     * cells. This will set the columns width as well as its preferred width.
     * Only works if the table's parent is a JViewport, i.e. the table is
     * nested in a JScrollPane.
     *
     * @param   table   The table that contains the column.
     * @param   column  The colunm index in table coordinates.
     */
    public static void autoSizeTableColumn(JTable table, int column)
    {
        if (table.getRowCount() == 0
            || ! (table.getParent() instanceof JViewport)) {
            return;
        }

        JViewport vp = (JViewport) table.getParent();
        Rectangle rect = vp.getViewRect();

        int start = table.rowAtPoint(new Point(0, rect.y));
        int end = table.rowAtPoint(new Point(0, rect.y + rect.height));
        if (end < 0) {
            end = table.getRowCount() - 1;
        }

        int size = 0;
        int margin = table.getColumnModel().getColumnMargin();
        for (int i = start; i <= end; i++) {
            Component com = table.prepareRenderer
                (table.getCellRenderer(i, column), i, column);
            if (com instanceof JComponent) {
                size = Math.max
                    (size,
                     margin + ((JComponent) com).getPreferredSize().width);
            } else {
                size = Math.max(size, margin + com.getSize().width);
            }
        }

        table.getColumnModel().getColumn(column).setPreferredWidth(size);
        table.getColumnModel().getColumn(column).setWidth(size);
        table.revalidate();
    }

    // }}}

    //----------------------------------------------------------------------
    // Miscellaneous
    //----------------------------------------------------------------------
    // {{{ centerWindow

    /**
     * Place a window in the center of the screen.
     *
     * @param   win     The window.
     *
     * @since   0.96.0
     */
    public static void centerWindow(Window win)
    {
        Dimension sSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension wSize = win.getSize();

        win.setLocation((sSize.width - wSize.width) / 2,
                        (sSize.height - wSize.height) / 2);
    }

    // }}}
    // {{{ centerWindowOnComponent

    /**
     * Place a window in the center of a component.
     *
     * @param   win     The window.
     * @param   com     The component to center on.
     * @param   constrain       Whether to constrain the window to the visible
     *                  screen.
     *
     * @since   0.99.0
     */
    public static void centerWindowOnComponent(Window win, Component com,
                                               boolean constrain)
    {
        Dimension wSize = win.getSize();
        Dimension cSize = com == null ? win.getToolkit().getScreenSize()
            : com.getSize();
        Point cPos = com == null ? new Point (0,0) : com.getLocationOnScreen();

        int x = cPos.x + (cSize.width - wSize.width) / 2;
        int y = cPos.y + (cSize.height - wSize.height) / 2;

        if (constrain) {
            Dimension sSize = Toolkit.getDefaultToolkit().getScreenSize();
            x = Math.max (0, Math.min(x, sSize.width - wSize.width));
            y = Math.max (0, Math.min(y, sSize.height - wSize.height));
        }
        win.setLocation(x, y);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class FocusDisabler

    /**
     * Visitor that disables the focus for JDK 1.4 JScrollBars.
     */
    public static class FocusDisabler
    {
        //--------------------------------------------------------------
        // Constructor
        //--------------------------------------------------------------
        // {{{ FocusDisabler

        /**
         * Create a new FocusDisabler.
         */
        public FocusDisabler ()
    {
    }

    // }}}

        // {{{ visit(Object)

        /**
         * Fallback
         *
         * @param       o       Ignored.
         *
         * @since       0.99.0
         */
        public void visit(Object o)
        {
        }

        // }}}
        // {{{ visit

        /**
         * Visit a JScrollPane. Disable the focus of its scroll bars.
         *
         * @param       sp     The scrollpane.
         *
         * @since       0.99.0
         */
        public void visit(JScrollPane sp) {
            Reflector.safeCall(sp.getHorizontalScrollBar(), "setFocusable",
                               new Class[] {Boolean.TYPE},
                               new Object[] {Boolean.FALSE});
            Reflector.safeCall(sp.getVerticalScrollBar(), "setFocusable",
                               new Class[] {Boolean.TYPE},
                               new Object[] {Boolean.FALSE});

        }

        // }}}
    }

    // }}}
    // {{{ class PoppableEventQueue

    /**
     * EventQueue that can be popped.
     */
    private static class PoppableEventQueue
        extends java.awt.EventQueue
        implements Runnable
    {
        public void run() {
            super.pop();
        }

    }

    // }}}

}


