// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Rectangle;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import java.util.Properties;

import javax.swing.BorderFactory;
import javax.swing.JTextArea;
import javax.swing.JScrollPane;
import javax.swing.border.Border;
import javax.swing.text.BadLocationException;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.config.Configurable;
import de.qfs.lib.config.Configuration;

// }}}

/**
 * Scroll pane with a text area that draws a border around itself, when
 * the text area has the focus.
 *
 * @author      Gregor Schmid
 */
public class TextView
        extends JScrollPane
    implements FocusListener, Configurable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.gui.TextView");

    /**
     * Optional size limit for the view.
     */
    private int limit = -1;

    /**
     * The non focus border.
     */
    private Border defaultBorder;

    /**
     * The border used when the TextView has the focus.
     */
    private Border focusBorder;

    /**
     * The JTextArea of the TextView.
     */
    private JTextArea textArea;

    /**
     * Current position of the view.
     */
    private int currentPos;

    /**
     * Listner for scrolling.
     */
    private KeyAdapter scrollListener = new KeyAdapter () {
        public void keyPressed (KeyEvent e) {
            if (e.getKeyCode() == e.VK_UP &&
                e.getModifiers() == 0) {
                SwingUtil.scrollDownLine(TextView.this);
                e.consume();
            } else if (e.getKeyCode() == e.VK_DOWN &&
                       e.getModifiers() == 0) {
                SwingUtil.scrollUpLine(TextView.this);
                e.consume();
            }
        }
    };

    //----------------------------------------------------------------------
    // Attributes for the Configurable interface
    //----------------------------------------------------------------------
    /**
     * The configurable name.
     */
    private String configurableName = "TextView";

    /**
     * Whether to scroll to the end after appending text.
     */
    private boolean scrollOnAppend = true;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ TextView()

    /**
     * Create a new TextView.
     */
    public TextView ()
    {
        this (null);
    }

    // }}}
    // {{{ TextView(String)

    /**
     * Create a new TextView.
     *
     * @param       contents        The text to display in the view.
     */
    public TextView (String contents)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "TextView(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "contents: " + contents);
        }

        textArea = new JTextArea (contents) {
            public boolean isManagingFocus()
            {
                return false;
            }
        };
        textArea.setLineWrap(true);
        textArea.setEditable(false);
        textArea.addFocusListener(this);

        setViewportView(textArea);

        defaultBorder = BorderFactory.createCompoundBorder
            (BorderFactory.createEmptyBorder(1, 1, 1, 1),
             BorderFactory.createLoweredBevelBorder());
        focusBorder = BorderFactory.createCompoundBorder
            (Borders.STIPPLE,
             BorderFactory.createLoweredBevelBorder());
        this.setBorder(defaultBorder);

        textArea.addKeyListener(scrollListener);
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getTextArea

    /**
     * Get the JTextArea of the TextView.
     *
     * @return      The JTextArea of the TextView.
     */
    public JTextArea getTextArea()
    {
        return textArea;
    }

    // }}}
    // {{{ clear

    /**
     * Convenience method to clear the JTextArea.
     */
    public void clear()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "clear()", "");
        }
        textArea.setText(null);
    }

    // }}}
    // {{{ append

    /**
     * Convenience method to add text to the JTextArea.
     *
     * @param   text    The text to add.
     */
    public void append(String text)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "append(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "text: " + text);
        }
        // boolean atEnd = true;
        if (isShowing()) {
            // atEnd = getViewport().getViewPosition().y +
                // getViewport().getExtentSize().height >=
                // textArea.getHeight();
            // System.err.println("atEnd: " + atEnd);
            // System.err.println("y: " + getViewport().getViewPosition().y);
            // System.err.println("h: " + getViewport().getExtentSize().height);
            // System.err.println("h2: " + textArea.getHeight());
            currentPos = textArea.viewToModel(getViewport().getViewPosition());
            // if (currentPos >= 0) {
                // textArea.setCaretPosition(currentPos);
            // }
        }
        textArea.append(text);
        enforceLimit();
        if (isShowing()) {
            if (scrollOnAppend) {
                int len = textArea.getDocument().getLength();
                textArea.setCaretPosition(len);
                try {
                    Rectangle rect = textArea.modelToView(len);
                    rect.x = 0;
                    textArea.scrollRectToVisible(rect);
                } catch (BadLocationException ex) {
                    // Shouldn't happen
                    if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "append(String)", ex);
                    }
                }
            } else {
                textArea.setCaretPosition(currentPos);
            }
        }
    }



    // }}}
    // {{{ getLimit

    /**
     * Get the size limit of the TextView.
     *
     * @return      The limit of the TextView.
     */
    public final int getLimit()
    {
        return limit;
    }

    // }}}
    // {{{ setLimit

    /**
     * Set the size limit of the TextView.
     *
     * @param   limit   The limit to set. 0 or negative means no limit.
     */
    public final void setLimit(int limit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setLimit(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "limit: " + limit);
        }
        this.limit = limit;
        enforceLimit();
    }

    // }}}
    // {{{ isScrollOnAppend

/**
 * Get the scrollOnAppend state of the TextView.
 *
 * @return      The scrollOnAppend state of the TextView.
 */
    public final boolean isScrollOnAppend()
    {
        return scrollOnAppend;
    }

    // }}}
    // {{{ setScrollOnAppend

    /**
     * Set the scrollOnAppend state of the TextView.
     *
     * @param   scrollOnAppend  The scrollOnAppend state to set.
     */
    public final void setScrollOnAppend(boolean scrollOnAppend)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setScrollOnAppend(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "scrollOnAppend: " + scrollOnAppend);
        }
        this.scrollOnAppend = scrollOnAppend;
    }

    // }}}

    //----------------------------------------------------------------------
    // The Configurable interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableName

    /**
     * Get the configurableName of the TextView.
     *
     * @return  The configurableName of the TextView.
     */
    public final String getConfigurableName()
    {
        return configurableName;
    }

    // }}}
    // {{{ setConfigurableName

    /**
     * Set the configurableName of the TextView.
     *
     * @param   configurableName        The configurableName to set.
     */
    public final void setConfigurableName(String configurableName)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableName(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurableName: " + configurableName);
        }
        this.configurableName = configurableName;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     *
     * @return The Configurable's state.
     */
    public Properties getConfigurableState()
    {
        Properties ret = new Properties ();
        ret.put("scrollOnAppend", "" + scrollOnAppend);
        ret.put("wrapLines", "" + textArea.getLineWrap());
        return ret;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     *
     * @param   state   The state to set.
     */
    public void setConfigurableState(Properties state)
    {
        if (state.containsKey("scrollOnAppend")) {
            scrollOnAppend = new Boolean (state.getProperty("scrollOnAppend"))
                .booleanValue();
        }
        if (state.containsKey("wrapLines")) {
            textArea.setLineWrap(new Boolean (state.getProperty("wrapLines"))
                                 .booleanValue());
        }
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ enforceLimit

    /**
     * Enforce the size limit for the TextView.
     *
     * @since   0.99.0
     */
    private void enforceLimit()
    {
        if (limit > 0) {
            int len = textArea.getDocument().getLength();
            if (len > limit) {
                try {
                    textArea.getDocument().remove(0, len - limit);
                    currentPos = Math.max(0, currentPos - (len - limit));
                } catch (BadLocationException ex) {
                    // Shouldn't happen
                    if (logger.level >= Log.ERR) {
                        logger.log(Log.ERR, "enforceLimit()", ex);
                    }
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The FocusListener interface
    //----------------------------------------------------------------------
    // {{{ focusGained

    /**
     * Draw a focus border around the TextView.
     */
    public void focusGained(FocusEvent e)
    {
        this.setBorder(focusBorder);
    }


    // }}}
    // {{{ focusLost

    /**
     * Remove the focus border from the TextView.
     */
    public void focusLost(FocusEvent e)
    {
        this.setBorder(defaultBorder);
    }

    // }}}

}
