// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A ChoiceOption is an Option that is edited with the help of a {@link
 * ChoiceOptionEdit ChoiceOptionEdit}.
 *
 * @author      Gregor Schmid
 */
public class ChoiceOption
    extends Option
    implements ConfigurableOption
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -10121595319748204L;

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.ChoiceOption");

    /**
     * Whether the option is integer based.
     */
    private boolean intBased;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ ChoiceOption()

    /**
     * Create a new ChoiceOption.
     */
    public ChoiceOption ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ChoiceOption()", "");
        }
    }

    // }}}
    // {{{ ChoiceOption(String,Object)

    /**
     * Create a new ChoiceOption.
     *
     * @param   name    The name of the ChoiceOption.
     * @param   value   The initial value.
     */
    public ChoiceOption(String name, Object value)
    {
        super(name, value);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ChoiceOption(String,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
    }

    // }}}
    // {{{ ChoiceOption(String,int)

    /**
     * Create a new integer based ChoiceOption.
     *
     * @param   name    The name of the ChoiceOption.
     * @param   value   The initial value.
     */
    public ChoiceOption(String name, int value)
    {
        super(name, new Integer (value));
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ChoiceOption(String,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
        intBased = true;
    }

    // }}}

    //----------------------------------------------------------------------
    // Getters and setters
    //----------------------------------------------------------------------
    // {{{ isIntBased

    /**
     * Get the intBased state of the ChoiceOption.
     *
     * @return  The intBased state of the ChoiceOption.
     */
    public final boolean isIntBased()
    {
        return intBased;
    }

    // }}}

    //----------------------------------------------------------------------
    // The ConfigurableOption interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableValue

    /**
     * Get the Option's value as a String.
     *
     * @return  The Option's value as a String.
     */
    public String getConfigurableValue()
    {
        Object value = getValue();
        return value == null ? "" : value.toString();
    }

    // }}}
    // {{{ setConfigurableValue

    /**
     * Set the Option's value from a String.
     *
     * @param   val     The value to set.
     */
    public void setConfigurableValue(String val)
    {
        if (isIntBased()) {
            try {
                setValue(val.equals("") ? null : new Integer (val));
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "setConfigurableValue(String)", ex);
                }
            }
        } else {
            setValue(val.equals("") ? null : val);
        }
    }

    // }}}
}
