// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.Enum;

// }}}

/**
 * An EnumOption's value type is an Enum. It is edited with the help of an
 * {@link EnumOptionEdit EnumOptionEdit}.
 *
 * @author      Gregor Schmid
 */
public class EnumOption
    extends Option
    implements ConfigurableOption
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -10121595319748204L;

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.EnumOption");

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ EnumOption()

    /**
     * Create a new EnumOption.
     */
    public EnumOption ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "EnumOption()", "");
        }
    }

    // }}}
    // {{{ EnumOption(String,Enum)

    /**
     * Create a new EnumOption.
     *
     * @param   name    The name of the EnumOption.
     * @param   value   The initial value.
     */
    public EnumOption(String name, Enum value)
    {
        super(name, value);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "EnumOption(String,Enum)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The ConfigurableOption interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableValue

    /**
     * Get the Option's value as a String.
     *
     * @return  The Option's value as a String.
     */
    public String getConfigurableValue()
    {
        Enum value = (Enum) getValue();
        return value == null ? "" :
            value.getTypeName() + ":" + value.getValue();
    }

    // }}}
    // {{{ setConfigurableValue

    /**
     * Set the Option's value from a String.
     *
     * @param   val     The value to set.
     */
    public void setConfigurableValue(String val)
    {
        int idx = val.lastIndexOf(":");
        if (idx > 0) {
            try {
                setValue(Enum.lookup(val.substring(0, idx),
                                     Integer.parseInt
                                     (val.substring(idx + 1))));
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "setConfigurableValue(String)", ex);
                }
                setValue(null);
            }
        } else {
            setValue(null);
        }
    }

    // }}}
}
