// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;

import java.util.Vector;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.UIManager;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.Enum;
import de.qfs.lib.util.Misc;
import de.qfs.lib.util.MapResourceBundle;

// }}}

// {{{ doc

/**
 * An OptionEdit for a {@link EnumOption EnumOption}. It is represented by
 * a {@link javax.swing.JComboBox JComboBox} and uses the following resources:
 * <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.type</code></TD>
 * <TD>The type of the Enum (requird, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.numchoices</code></TD>
 * <TD>The number of choices in the edit.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.choice<em>n</em>.label</code></TD>
 * <TD>The label for the n<sup>th</sup> choice of the edit, where n is
 * between 1 and numchoices.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.choice<em>n</em>.value</code></TD>
 * <TD>The Enum value for the n<sup>th</sup> choice of the edit, where n is
 * between 1 and numchoices (optional, default n).</TD>
 * </TR>
 * </TABLE>
 *
 *
 * @author      Gregor Schmid
 */

// }}}
public class EnumOptionEdit
    extends OptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.EnumOptionEdit");

    /**
     * The label of the component.
     */
    protected JLabel label;

    /**
     * The ComboBox Component of the EnumOptionEdit.
     */
    protected JComboBox combo;

    /**
     * The Enum type for the option.
     */
    protected String type;

    /**
     * The possible choices for the JComboBox.
     */
    protected Object[] choices;

    /**
     * The possible values for the JComboBox.
     */
    protected int[] values;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ EnumOptionEdit

    /**
     * Create a new EnumOptionEdit. Don't call this directly, use {@link
     * OptionEdit#makeEdit OptionEdit.makeEdit} instead.
     */
    public EnumOptionEdit ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "EnumOptionEdit()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the OptionEdit. Called from {@link #makeEdit makeEdit}.
     */
    public void init()
    {
        setLayout(new BorderLayout ());
        label = new JLabel
            (getString ("label", "Missing: " + option.getName()));
        label.setName("Label-" + option.getName());
        String mnemo = getString("mnemonic", null);
        if (mnemo != null && mnemo.length() > 0) {
            label.setDisplayedMnemonic(mnemo.charAt(0));
        }
        add(label, BorderLayout.NORTH);

        type = getString("type", null);
        if (type == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "getEnums()",
                           "No type for EnumOptionEdit " + option.getName());
            }
            type = "";
        }

        getChoices();
        combo = new JComboBox (choices);
        label.setLabelFor(combo);
        combo.setName("ComboBox-" + option.getName());
        combo.setEditable(false);

        add (combo, BorderLayout.CENTER);
        update(option);
        disabled = getBoolean("disabled", false);
        setEnabled(!disabled);
    }

    // }}}
    // {{{ isModified

    /**
     * Query whether the edit has been modified, i.e. the current value of the
     * edit differs from the option's value.
     *
     * @return  True if the values in the edit and option differ.
     *
     * @since   0.99.0
     */
    public boolean isModified()
    {
        return Misc.equalOrNull(option.getValue(), getCurrentValue());
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "checkValue()", "");
        }
        if (! option.checkValue(this, getCurrentValue())) {
            combo.requestFocus();
            return false;
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }
        option.setValue(getCurrentValue());
    }

    // }}}
    // {{{ update

    /**
     * Update the display for a possibly new Option.
     *
     * @param   option The new option.
     */
    public void update(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        this.option = option;
        if (! (option.getValue() instanceof Enum)) {
            combo.setSelectedIndex(0);
        } else {
            int val = ((Enum) option.getValue()).getValue();
            for (int i = 0; i < values.length; i++) {
                if (values[i] == val) {
                    combo.setSelectedIndex(i);
                    break;
                }
            }

        }
    }

    // }}}

    // {{{ getComboBox

    /**
     * Get the JComboBox used by the EnumOptionEdit.
     *
     * @return  The edit's combo box.
     *
     * @since   0.98.1
     */
    public JComboBox getComboBox()
    {
        return combo;
    }


    // }}}

    //----------------------------------------------------------------------
    // Override some methods from JPanel
    //----------------------------------------------------------------------
    // {{{ setEnabled

    /**
     * Enable or diesable the EnumOptionEdit.
     *
     * @param   enabled True to enable, false to disable.
     */
    public void setEnabled(boolean enabled)
    {
        if (disabled) {
            enabled = false;
        }
        combo.setEnabled(enabled);
        label.setForeground(UIManager.getColor
                            (enabled ? "textText" : "textInactiveText"));
    }

    // }}}
    // {{{ isEnabled

    /**
     * Query whether the EnumOptionEdit is enabled.
     *
     * @return  True if it is enabled.
     */
    public boolean isEnabled()
    {
        return combo.isEnabled();
    }

    // }}}
    // {{{ requestFocus

    /**
     * Request the focus for the EnumOptionEdit.
     */
    public void requestFocus()
    {
        combo.requestFocus();
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ getChoices

    /**
     * Create the possible choices from the resources.
     */
    protected void getChoices()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getChoices()", "");
        }

        Vector choices = new Vector();
        Vector values = new Vector();

        int num = getInt("numchoices", -1);
        if (num < 0) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "getEnums()",
                           "No numchoices for EnumOptionEdit " +
                           option.getName());
                return;
            }
        }

        for (int i = 1; i <= num; i++) {
            String choice = getString("choice" + i + ".label", null);
            if (choice == null) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "getEnums()",
                               "No label for choice " + i +
                               " in EnumOptionEdit " + option.getName());
                }
                continue;
            }
            choices.addElement(choice);
            String val = getString("choice" + i + ".value", "" + i);
            try {
                values.addElement(new Integer (val));
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "getEnums()",
                               "Bad value for choice " + i +
                               " in EnumOptionEdit " + option.getName());
                }
                values.addElement(new Integer (i));
            }
        }

        this.choices = new Object [choices.size()];
        choices.copyInto(this.choices);
        this.values = new int[values.size()];
        for (int i = 0; i < this.values.length; i++) {
            this.values[i] = ((Integer) values.elementAt(i)).intValue();
        }

    }

    // }}}
    // {{{ getCurrentValue

    /**
     * Get the currently selected value.
     *
     * @return  The currently selected value.
     *
     * @since   0.99.0
     */
    protected Enum getCurrentValue()
    {
        int idx = combo.getSelectedIndex();
        if (idx < 0 || idx >= values.length) {
            return null;
        }
        return Enum.lookup(type, values[idx]);
    }

    // }}}
}
