// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;

import java.io.File;

import java.util.StringTokenizer;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

import de.qfs.lib.gui.Message;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.ExtensionFileFilter;

// }}}

// {{{ doc

/**
 * An OptionEdit for a {@link FileOption FileOption}. It can also be used
 * in combination with a {@link StringOption StringOption} by setting the
 * resource <code>option.<em>name</em>.edit.class</code> to the value
 * <code>de.qfs.lib.option.FileOptionEdit</code>.
 *
 * It uses the following resources: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.select</code></TD>
 * <TD>Whether to select the whole text when the edit gains focus (optional,
 * default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.action</code></TD>
 * <TD>The type of operation for the file dialog. Values are "open" or "save"
 * for a standard open or save dialog. Any other value is used as the text of
 * the approve button (optional, default "open").</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.title</code></TD>
 * <TD>The title of the file dialog (optional, default is Swing
 * dependent).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.numfilters</code></TD>
 * <TD>The number of file filters for the dialog (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.filter&lt;n&gt;.extensions</code></TD>
 * <TD>The extensions for the n-th file filter, starting from 1. Extensions
 * must include the '.' character. Multiple extensions are separated by
 * whitespace.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.filter&lt;n&gt;.desc</code></TD>
 * <TD>The description for the n-th file filter, starting from 1.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.allfilter</code></TD>
 * <TD>Whether to add the all-files file filter to the file dialog (optional,
 * default true if no other filters are defined, false otherwise).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.askoverwrite</code></TD>
 * <TD>Whether to ask for permission to overwrite a file that already exists
 * (optional, default true if type is "save", false otherwise).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */

// }}}
public class FileOptionEdit
    extends SelectorOptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.FileOptionEdit");

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ FileOptionEdit

    /**
     * Create a new FileOptionEdit.
     */
    public FileOptionEdit ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the edit.
     */
    public void init()
    {
        text = new JTextField () {
            public boolean isFocusTraversable()
            {
                return isEditable() && this.isEnabled();
            }
            public boolean isFocusable() {
                return true;
            }
        };
        super.init();
        label.setLabelFor(text);
        text.setName("Text-" + option.getName());
        add(text, BorderLayout.CENTER);
        if (selectOnFocus && ! disabled) {
            text.addFocusListener(this);
            text.addMouseListener(this);
        }
        if (option.getValue() != null) {
            text.setText((option instanceof FileOption)
                         ? ((File) option.getValue()).getPath()
                         : option.getValue().toString());
            text.setCaretPosition(text.getDocument().getLength());
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Standard methods
    //----------------------------------------------------------------------
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        String val = text.getText();
        File file = null;
        if (!val.equals("")) {
            file = new File (val);
            if (option instanceof FileOption) {
                if (((FileOption) option).isMustExist()
                    && ! file.exists()) {
                    Message.showMessage(this, "option.error.noSuchFile");
                    text.requestFocus();
                    return false;
                }
                if (((FileOption) option).getFileType() ==
                        FileOption.DIRECTORY_TYPE
                    && file.exists()
                    && ! file.isDirectory()) {
                    Message.showMessage(this, "option.error.notADirectory");
                    text.requestFocus();
                    return false;
                }
                if (((FileOption) option).getFileType() == FileOption.FILE_TYPE
                    && file.exists()
                    && ! file.isFile()) {
                    Message.showMessage(this, "option.error.notAFile");
                    text.requestFocus();
                    return false;
                }
            }
            if(file.exists()
               && getBoolean("askoverwrite",
                             "save".equals(getString("action", "open")))) {
                String res = getString("askoverwrite.message", null) == null
                    ? "option.error.askOverwrite"
                    : "option." + option.getName() + ".askoverwrite";
                if (Message.showMessage(this, res,
                                        new Object[] {file.getPath()})
                    != JOptionPane.OK_OPTION) {
                    text.requestFocus();
                    return false;
                }
            }
        }
        if (! option.checkValue(this,
                                (option instanceof FileOption)
                                ? (Object) file : (Object) val)) {
            text.requestFocus();
            return false;
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }
        String val = text.getText();
        option.setValue(val.equals("") ? null :
                        ((option instanceof FileOption)
                         ? (Object) new File (val) : (Object) val));
    }

    // }}}

    //----------------------------------------------------------------------
    // Show the dialog
    //----------------------------------------------------------------------
    // {{{ actionPerformed

    /**
     * Notify the edit that the button was pressed.
     *
     * @param   e       Ignored.
     */
    public void actionPerformed(ActionEvent e)
    {
        File file = showFileDialog(getText());
        if (file != null) {
            setText(file.getPath());
        }
        text.requestFocus();
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper method
    //----------------------------------------------------------------------
    // {{{ askForFile

    /**
     * Ask the user for the file to load or save to.
     *
     * @param   current    The current file or null.
     *
     * @return  The new file.
     */
    protected File showFileDialog(String current)
    {
        File file = null;
        if (current != null && current.length() > 0) {
            file = new File (current);
        }


        JFileChooser chooser = new JFileChooser ();

        //------------------------------------------------------------------
        // Set up the choosers filters
        //------------------------------------------------------------------
        int numFilters = getInt("numfilters", 0);
        if (numFilters > 0) {
            ExtensionFileFilter dflt = null;
            for (int i = 1; i <= numFilters; i++) {
                String desc = getString("filter" + i + ".desc", null);
                String ext = getString("filter" + i + ".extensions", null);
                if (desc == null || ext == null) {
                    break;
                }
                int len = 0;
                for (StringTokenizer st = new StringTokenizer (ext, " \t");
                     st.hasMoreElements();) {
                    st.nextElement();
                    len++;
                }
                if (len == 0) {
                    break;
                }
                String[] exts = new String[len];
                len = 0;
                for (StringTokenizer st = new StringTokenizer (ext, " \t");
                     st.hasMoreTokens();) {
                    exts[len++] = st.nextToken();
                }
                // Add ignoreCase flag?
                ExtensionFileFilter filter = new ExtensionFileFilter
                    (desc, exts, true, false);
                if (dflt == null) {
                    dflt = filter;
                }
                chooser.addChoosableFileFilter(filter);
            }
            if (dflt != null
                && ! getBoolean("allfilter", true)) {
                chooser.removeChoosableFileFilter
                    (chooser.getAcceptAllFileFilter());
                chooser.setFileFilter(dflt);
            }
        }

        //------------------------------------------------------------------
        // The title of the dialog
        //------------------------------------------------------------------
        String title = getString("title", null);
        if (title != null) {
            chooser.setDialogTitle(title);
        }

        //------------------------------------------------------------------
        // Set current directory
        //------------------------------------------------------------------
        chooser.setCurrentDirectory
            (file == null
                 ? new File (new File (new File (".").getAbsolutePath())
                             .getParent())
                 : file);
        if (file != null) {
            chooser.setSelectedFile(file);
        }

        //------------------------------------------------------------------
        // Select directory?
        //------------------------------------------------------------------
        if ((option instanceof FileOption)
            && ((FileOption) option).getFileType() == FileOption.DIRECTORY_TYPE) {
            chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            chooser.removeChoosableFileFilter
                (chooser.getAcceptAllFileFilter());
            chooser.setFileFilter(new FileFilter () {
                public boolean accept (File f) {
                    return f.isDirectory();
                }
                public String getDescription() {
                    return Option.getResources().getString
                        ("option.fileoption.directoryfilter.desc", "Directories");
                }
            });
        }

        //------------------------------------------------------------------
        // Load or save?
        //------------------------------------------------------------------
        String action = getString("action", "open");
        int ret;
        if ("open".equals(action)) {
            ret = chooser.showOpenDialog(this);
        } else if ("save".equals(action)) {
            ret = chooser.showSaveDialog(this);
        } else {
            ret = chooser.showDialog(this, action);
        }
        if (ret != chooser.APPROVE_OPTION) {
            return null;
        }

        file = chooser.getSelectedFile();
        if (file == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "showFileDialog(String)",
                           "File chooser returned null file");
            }
            return null;
        }

        //------------------------------------------------------------------
        // Check again for directories
        //------------------------------------------------------------------
        if ((option instanceof FileOption)
            && ((FileOption) option).getFileType()
                == FileOption.DIRECTORY_TYPE) {
            if (file == null || ! file.exists()) {
                file = chooser.getCurrentDirectory();
            }
        }

        //------------------------------------------------------------------
        // Maybe auto-complete
        //------------------------------------------------------------------
        FileFilter filter = chooser.getFileFilter();
        if (file.getName().indexOf(".") < 0
            && filter instanceof ExtensionFileFilter) {
            file = ((ExtensionFileFilter) filter).forceExtension(file);
        }

        return file;
    }

    // }}}
}
