// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;
import java.awt.Dimension;

import java.util.StringTokenizer;

import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.UIManager;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Special edit that displays a (possibly multi-line) label.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public class LabelOptionEdit
    extends OptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.LabelOptionEdit");

    /**
     * Whether the label shoud be displayed disabled.
     */
    protected boolean disabled;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ LabelOptionEdit

    /**
     * Create a new LabelOptionEdit.
     */
    public LabelOptionEdit ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "LabelOptionEdit()", "");
        }
    }

    // }}}

    // {{{ init

    /**
     * Initialize the OptionEdit. This method is called by the {@link
     * #makeEdit makeEdit} factory method after the option member has been
     * set. <p>
     *
     * Must be implemented by a derived class.
     */
    protected void init()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }
        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        disabled = getBoolean ("disabled", false);
        update(option);
    }

    // }}}
    // {{{ update(Option)

    /**
     * Update the display of the edit component after a possible change of its
     * Option or the Option's value. <p>
     *
     * Must be implemented by a derived class.
     *
     * @param   option The (possibly new) option.
     */
    public void update(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }

        this.option = option;
        removeAll();
        if (option.getValue() != null
            && ! (option.getValue() instanceof String)) {
            // empty
            return;
        }
        String val = (String) option.getValue();
        if (val == null) {
            val = getString("label", null);
        }
        while (val != null) {
            int pos = val.indexOf("\n");
            if (pos >= 0) {
                if (pos == 0) {
                    add(new Empty ());
                } else {
                    JLabel label = new JLabel (val.substring(0, pos));
                    label.setForeground(UIManager.getColor
                            (disabled ? "textInactiveText" : "textText"));
                    add(label);
                }
                val = pos == val.length() - 1 ? null : val.substring(pos + 1);
            } else {
                if ("".equals(val)) {
                    add (new Empty ());
                } else {
                    JLabel label = new JLabel (val);
                    label.setForeground(UIManager.getColor
                            (disabled ? "textInactiveText" : "textText"));
                    add(label);
                }
                val = null;
            }
        }
        invalidate();
    }

    // }}}
    // {{{ isModified

    /**
     * Query whether the edit has been modified, i.e. the current value of the
     * edit differs from the option's value.
     *
     * @return  False.
     *
     * @since   0.99.0
     */
    public boolean isModified()
    {
        return false;
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus. The OptionEdit should check the value
     * against its own constraints as well as pass it to {@link
     * Option#checkValue Option.checkValue}. <p>
     *
     * @return  True.
     */
    public boolean checkValue()
    {
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option. <p>
     *
     * Noop.
     */
    public void readValue()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Override some methods from JPanel
    //----------------------------------------------------------------------
    // {{{ setEnabled

    /**
     * Enable or disable the SimpleOptionEdit.
     *
     * @param   enabled True to enable, false to disable.
     */
    public void setEnabled(boolean enabled)
    {
        if (disabled) {
            enabled = false;
        }

        Component[] labels = getComponents();
        for (int i = 0; i < labels.length; i++) {
            if (labels[i] instanceof JLabel) {
                ((JLabel) labels[i]).setForeground
                    (UIManager.getColor
                     (enabled ? "textText" : "textInactiveText"));
            }
        }
    }

    // }}}
    //----------------------------------------------------------------------
    // Helper class
    //----------------------------------------------------------------------
    // {{{ class Empty

    /**
     * An Empty component with the height of a JLabel.
     */
    private static class Empty
       extends Component
    {
        private int height;

        public Empty ()
        {
            JLabel dummy = new JLabel ("Xyz");
            height = dummy.getPreferredSize().height;
        }

        public Dimension getPreferredSize() {
            return new Dimension(1, height);
        }
    }

    // }}}
}
