// {{{ copyright

/********************************************************************
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Insets;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;

import javax.swing.AbstractButton;
import javax.swing.Box;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.KeyStroke;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Base class for OptionEdits that need to display an editable String value
 * with a selector button next to it, e.g. a FileOptionEdit.
 *
 * Concrete subclasses must override {@link #createButton createButton} or
 * configure the button in {@link #init init} after calling super.init() and
 * implement {@link #actionPerformed actionPerformed}.
 *
 * The label of the selector button can also be set with the resource
 * <code>option.<em>name</em>.select.label</code>.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public abstract class SelectorOptionEdit
    extends SimpleOptionEdit
    implements ActionListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.SelectorOptionEdit");

    /**
     * The select... button.
     */
    protected AbstractButton select;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ SelectorOptionEdit

    /**
     * Create a new SelectorOptionEdit.
     */
    public SelectorOptionEdit ()
    {
    }

    // }}}

    // {{{ init

    /**
     * Initialize the OptionEdit.
     */
    public void init()
    {
        // Create the button first so setEnable will work
        select = createButton();
        select.addActionListener(this);
        super.init();

        // Place the button to the right of the text field.
        Box box = Box.createHorizontalBox();
        box.add(Box.createHorizontalStrut(5));
        box.add(select);
        add(box, BorderLayout.EAST);

        text.registerKeyboardAction
            (this,
             KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.SHIFT_MASK),
             WHEN_FOCUSED);
        text.registerKeyboardAction
            (this,
             KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.ALT_MASK),
             WHEN_FOCUSED);
    }

    // }}}
    // {{{ setEnabled

    /**
     * Enable or diesable the SelectorOptionEdit.
     *
     * @param   enabled True to enable, false to disable.
     */
    public void setEnabled(boolean enabled)
    {
        if (disabled) {
            enabled = false;
        }

        super.setEnabled(enabled);
        select.setEnabled(enabled);
    }

    // }}}

    // {{{ createButton

    /**
     * Create the select button.
     *
     * @return  The new button.
     */
    protected AbstractButton createButton()
    {
        JButton ret = new JButton () {
            public boolean isFocusTraversable() {
                return false;
            }
        };
        ret.setName("Selector-" + option.getName());
        ret.setRequestFocusEnabled(false);
        String text = getString("select.label", "...");
        Icon icon = Option.getResources().getIcon
            ("option." + option.getName() + ".select.icon", null);
        if (icon != null) {
            ret.setText(null);
            ret.setIcon(icon);
            ret.setMargin(new Insets(0, 0, 0, 0));
        } else {
            ret.setText(text);
            String mnemo = getString("select.mnemonic", null);
            if (mnemo != null && mnemo.length() > 0) {
                ret.setMnemonic(mnemo.charAt(0));
            }
            Insets ins = ret.getMargin();
            ret.setMargin(new Insets(0, ins.left, 0, ins.right));
        }
        String tip = getString("select.tooltip", null);
        if (tip != null) {
            ret.setToolTipText(tip);
        }
        return ret;
    }

    // }}}
}
