// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JLabel;
import javax.swing.UIManager;
import javax.swing.text.JTextComponent;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.Misc;

// }}}

/**
 * Common baseclass for similar kinds of OptionEdits that are built from
 * a JLabel and a JTextComponent.
 *
 * @author      Gregor Schmid
 */
public class SimpleOptionEdit
    extends OptionEdit
    implements FocusListener, MouseListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.SimpleOptionEdit");

    /**
     * The label of the component.
     */
    protected JLabel label;

    /**
     * The text of the component.
     */
    protected JTextComponent text;

    /**
     * Whether to select everything on a FOCUS_GAINED event.
     */
    protected boolean selectOnFocus;

    /**
     * Whether a FOCUS_GAINED event can cause a selection.
     */
    protected boolean canFocus = true;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ SimpleOptionEdit

    /**
     * Create a new SimpleOptionEdit. Don't call this directly, use
     * {@link OptionEdit#makeEdit OptionEdit.makeEdit}.
     */
    protected SimpleOptionEdit()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the OptionEdit. Called from {@link #makeEdit makeEdit}.
     */
    public void init()
    {
        setLayout(new BorderLayout ());
        label = new JLabel (getString("label", ""));
              // "Missing: " + option.getName()));
        label.setName("Label-" + option.getName());

        String mnemo = getString("mnemonic", null);
        if (mnemo != null && mnemo.length() > 0) {
            label.setDisplayedMnemonic(mnemo.charAt(0));
        }
        disabled = getBoolean("disabled", false);
        label.setForeground(UIManager.getColor
                            (disabled ? "textInactiveText" : "textText"));
        selectOnFocus = getBoolean("select", false);

        add(label, BorderLayout.NORTH);
    }

    // }}}
    // {{{ isModified

    /**
     * Query whether the edit has been modified, i.e. the current value of the
     * edit differs from the option's value.
     *
     * @return  True if the values in the edit and option differ.
     *
     * @since   0.99.0
     */
    public boolean isModified()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isModified()", "");
        }

        String val = text.getText();
        if ("".equals(val)) {
            if ("".equals(option.getValue())) {
                return false;
            }
            val = null;
        }

        if (logger.level >= Log.DBG) {
            try {
                logger.log(Log.DBG, "isModified()",
                           "val: " +
                           (val == null ? null : Misc.toHex(val.getBytes()))+
                           "\ntext:" +
                           Misc.toHex(text.getDocument().getText
                                      (0, text.getDocument().getLength())
                                      .getBytes()));
            } catch (Exception ex) {
            }
        }
        String oval = option.getValue( ) == null ? null
            : option.getValue().toString();
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "isModified()",
                       "option: " +
                       (oval == null ? null : Misc.toHex(oval.getBytes())));
        }


        boolean ret = ! Misc.equalOrNull(val, oval);

            // val == null ? option.getValue() != null :
            // option.getValue() == null ||
            // ! val.equals(option.getValue().toString());
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "isModified()",
                       "ret: " + ret);
        }
        return ret;
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        String val = text.getText();
        if (! option.checkValue(this, "".equals(val) ? null : val)) {
            text.requestFocus();
            return false;
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }
        String val = text.getText();
        option.setValue(val.equals("") ? null : val);
    }

    // }}}
    // {{{ update

    /**
     * Update the display for a possibly new Option.
     *
     * @param   option The new option.
     */
    public void update(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        this.option = option;
        Object val = option.getValue();
        text.setText(val == null ? "" : val.toString());
        text.setCaretPosition(0);
        text.revalidate();
    }

    // }}}

    // {{{ getText

    /**
     * Get the text of the SimpleOptionEdit.
     *
     * @return  The text of the SimpleOptionEdit.
     */
    public String getText()
    {
        return text != null ? text.getText() : null;
    }

    // }}}
    // {{{ setText

    /**
     * Set the text of the SimpleOptionEdit.
     *
     * @param   text    The text to set.
     *
     * @since   0.98.1
     */
    public void setText(String text)
    {
        if (this.text != null) {
            this.text.setText(text);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The FocusListener interface
    //----------------------------------------------------------------------
    // {{{ focusGained

    /**
     * Notify the edit of a FOCUS_GAINED event. Can cause a selection of the
     * whole text.
     *
     * @param   event   The event.
     *
     * @since   0.99.0
     */
    public void focusGained(FocusEvent event)
    {
        if (selectOnFocus && ! event.isTemporary() && canFocus) {
            canFocus = false;
            text.setSelectionStart(0);
            text.setSelectionEnd(text.getDocument().getLength());
        }
    }

    // }}}
    // {{{ focusLost

    /**
     * Notify the edit of a FOCUS_LOST event. If the event is permanent, reset
     * the focus so it can cause a selection again.
     *
     * @param   event   The event.
     *
     * @since   0.99.0
     */
    public void focusLost(FocusEvent event)
    {
        if (! event.isTemporary()) {
            canFocus = true;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The MouseListener interface
    //----------------------------------------------------------------------
    // {{{ mousePressed

    /**
     * Noop.
     *
     * @param   event   Ignored.
     *
     * @since   0.99.0
     */
    public void mousePressed(MouseEvent event)
    {
    }

    // }}}
    // {{{ mouseReleased

    /**
     * Noop.
     *
     * @param   event   Ignored.
     *
     * @since   0.99.0
     */
    public void mouseReleased(MouseEvent event)
    {
    }

    // }}}
    // {{{ mouseClicked

    /**
     * Maybe reset the selection.
     *
     * @param   event   Ignored.
     *
     * @since   0.99.0
     */
    public void mouseClicked(MouseEvent event)
    {
        if (! Misc.isJavaVersionAtLeast(1, 4, -1)) {
            text.setCaretPosition
                (text.viewToModel(new Point(event.getX(), event.getY())));
        }
    }

    // }}}
    // {{{ mouseEntered

    /**
     * Noop.
     *
     * @param   event   Ignored.
     *
     * @since   0.99.0
     */
    public void mouseEntered(MouseEvent event)
    {
    }

    // }}}
    // {{{ mouseExited

    /**
     * Noop.
     *
     * @param   event   Ignored.
     *
     * @since   0.99.0
     */
    public void mouseExited(MouseEvent event)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Override some methods from JPanel
    //----------------------------------------------------------------------
    // {{{ setEnabled

    /**
     * Enable or disable the SimpleOptionEdit.
     *
     * @param   enabled True to enable, false to disable.
     */
    public void setEnabled(boolean enabled)
    {
        if (disabled) {
            enabled = false;
        }

        text.setEditable(enabled);
        // Metal L&F uses a gray background for non-editable text fields which
        // looks horrible and breaks the consistent look for TextFields and
        // TextAreas, so we override it.
        text.setBackground(UIManager.getColor("TextField.background"));
        if (enabled) {
            label.setForeground(UIManager.getColor("textText"));
        } else {
            label.setForeground(UIManager.getColor("textInactiveText"));
        }
    }

    // }}}
    // {{{ isEnabled

    /**
     * Query whether the SimpleOptionEdit is enabled.
     *
     * @return  True if it is enabled.
     */
    public boolean isEnabled()
    {
        return text.isEditable();
    }

    // }}}
    // {{{ requestFocus

    /**
     * Request the focus for the SimpleOptionEdit.
     */
    public void requestFocus()
    {
        text.requestFocus();
    }

    // }}}
}
