// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputMethodEvent;
import java.awt.event.InputMethodListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.io.IOException;

import java.util.HashSet;
import java.util.Set;

import java.util.EventObject;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JScrollPane;
import javax.swing.JLabel;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableCellEditor;

import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.gui.Borders;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.WindowsTableCellRenderer;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.Misc;
import de.qfs.lib.util.Reflector;

// }}}

// {{{ doc

/**
 * JTable based Edit for a TableOption. It uses the following resources:
 * <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.width</code></TD>
 * <TD>Preferred width of the edit (optional, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.height</code></TD>
 * <TD>Preferred height of the edit (optional, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.weight</code></TD>
 * <TD>The weight (a double) to use for the edit when placed in the {@link
 * de.qfs.lib.gui.Grid Grid} of an {@link OptionGroup OptionGroup} (optional,
 * default 1.0).</TD>
 * </TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.configure</code></TD>
 * <TD>Whether a {@link de.qfs.lib.config.ConfigWrapper ConfigWrapper}
 * should be created for the table and registered with the {@link
 * de.qfs.lib.config.Configuration Configuration} under the name
 * <code><em>name</em>.table</code> (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canedit</code></TD>
 * <TD>Whether the cell values can be edited (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.caninsert</code></TD>
 * <TD>Whether new rows can be added (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canremove</code></TD>
 * <TD>Whether rows can be deleted (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canmove</code></TD>
 * <TD>Whether rows can be moved up or down (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.directinsert</code></TD>
 * <TD>Whether new rows are inserted directly (true) or created via an
 * {@link OptionDialog OptionDialog} (false) (optional, default
 * <code>(!canedit)</code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.directedit</code></TD>
 * <TD>Whether typing a key starts editing the focused cell. This is the
 * default value for the whole table and can be overridden for each column
 * separately (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.autosize</code></TD>
 * <TD>Whether a double-click on the table header will resize the column to
 * fit its values (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.selectrow</code></TD>
 * <TD>Whether to select the whole row (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.multirow</code></TD>
 * <TD>Whether to support multiple row selection (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.cutpaste</code></TD>
 * <TD>Whether to support cut/copy/paste (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.columns</code></TD>
 * <TD>Whitespace separated list of column identifiers (required).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.name</code></TD>
 * <TD>The name of the column to display in the table's header (optional,
 * default <code><em>id</em></code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.type</code></TD>
 * <TD>The type of the column. Currently supported are <code>string</code>,
 * <code>taxt, </code><code>boolean</code>and <code>int</code> (optional,
 * default string).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.cwidth</code></TD>
 * <TD>The initial width of the column (optional, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.nullok</code></TD>
 * <TD>Whether empty values are legal in this column (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.negativeok</code></TD>
 * <TD>Whether negative values are legal in this column (optional, default
 * false, only for columns of type <code>int</code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.default</code></TD>
 * <TD>Default value for new rows (optional, default null).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.disabled</code></TD>
 * <TD>If true, values in this column cannot be edited (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.directedit</code></TD>
 * <TD>Whether typing a key starts editing the focused cell (optional, default
 * taken from the value for the whole table).</TD>
 * </TR>
 * </TABLE>
 * <p>
 *
 * To edit a row's values in a dialog, an {@link OptionDialog OptionDialog}
 * with the same name as the edit's option is used. You can customize this
 * dialog in the usual way. The {@link OptionGroup OptionGroup} for this
 * dialog also has the same name. If the resource
 * <code>option.group.<em>name</em>.type</code> is not defined, default values
 * are set to create a plain group with an edit for each column. The labels of
 * these edits are copied from the table hader. If you define the resource
 * <code>option.group.<em>name</em>.type</code> you take full responsibility
 * for all of the groups resources.
 *
 * @author      Gregor Schmid
 * @since       0.98.1
 */

// }}}
public class TableOptionEdit
    extends OptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.TableOptionEdit");

    /**
     * Key for forward TAB traversal.
     */
    private static HashSet forwardKeys;

    /**
     * Key for forward TAB traversal.
     */
    private static HashSet backwardKeys;

    static {
        if (Misc.isJavaVersionAtLeast(1, 4, -1)) {
            forwardKeys = new HashSet ();
            forwardKeys.add(KeyStroke.getKeyStroke
                            (KeyEvent.VK_TAB, 0));
            forwardKeys.add(KeyStroke.getKeyStroke
                            (KeyEvent.VK_TAB, KeyEvent.CTRL_MASK));
            backwardKeys = new HashSet ();
            backwardKeys.add(KeyStroke.getKeyStroke
                             (KeyEvent.VK_TAB, KeyEvent.SHIFT_MASK));
            backwardKeys.add(KeyStroke.getKeyStroke
                             (KeyEvent.VK_TAB,
                              KeyEvent.CTRL_MASK | KeyEvent.SHIFT_MASK));
        }
    }

    /**
     * The label of the edit.
     * @serial
     */
    protected JLabel label;

    /**
     * The table of the edit.
     * @serial
     */
    protected Table table;

    /**
     * The config wrapper for the table.
     */
    protected ConfigWrapper wrapper;

    /**
     * The columns of the table.
     */
    protected Column columns[];

    /**
     * Whether the edit is currently enabled.
     * @serial
     */
    protected boolean enabled = true;

    /**
     * Whether editing via dialog is allowed.
     */
    protected boolean canedit;

    /**
     * Whether adding rows is allowed.
     */
    protected boolean caninsert;

    /**
     * Whether removing rows is allowed.
     */
    protected boolean canremove;

    /**
     * Whether moving rows is allowed.
     */
    protected boolean canmove;

    /**
     * Whether to insert new rows directly.
     */
    protected boolean directinsert;

    /**
     * Whether to start editing immediately on a keypress.
     */
    protected boolean directedit;

    /**
     * Whether to select the whole row.
     */
    protected boolean selectrow;

    /**
     * Whether to allow multiple row selection.
     */
    protected boolean multirow;

    /**
     * Whether to support cut/copy/paste
     */
    protected boolean cutpaste;

    /**
     * The y-weight of the table.
     */
    protected double weight = 1.0;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ TableOptionEdit

    /**
     * Create a new TableOptionEdit.
     */
    public TableOptionEdit ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the Edit.
     */
    protected void init()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }
        String cols = getString("columns", null);
        if (cols == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "init()",
                           "No columns for TableOptionEdit " +
                           option.getName());
            }
            return;
        }

        // needed as default for the columns
        directedit = getBoolean("directedit", true);

        int cnt = 0;
        for (StringTokenizer st = new StringTokenizer (cols);
             st.hasMoreTokens(); st.nextToken()) {
            cnt++;
        }
        columns = new Column[cnt];
        cnt = 0;
        for (StringTokenizer st = new StringTokenizer (cols);
             st.hasMoreTokens();) {
            columns[cnt++] = new Column (st.nextToken());
        }

        canedit = getBoolean("canedit", true);
        caninsert = getBoolean("caninsert", true);
        canremove = getBoolean("canremove", true);
        canmove = getBoolean("canmove", true);
        directinsert = getBoolean("directinsert", ! canedit);
        selectrow = getBoolean("selectrow", false);
        multirow = selectrow && getBoolean("multirow", false);
        cutpaste = getBoolean("cutpaste", false);

        setLayout(new BorderLayout ());

        disabled = getBoolean("disabled", false);

        table = createTable();

        String name = getString("label", null);
        if (name != null) {
            label = new JLabel (name);
            label.setName("Label-" + option.getName());
            add(label, BorderLayout.NORTH);
            label.setForeground(UIManager.getColor
                                (disabled ? "textInactiveText" : "textText"));
            label.setLabelFor(table);
            String mnemo = getString("mnemonic", null);
            if (mnemo != null && mnemo.length() > 0) {
                label.setDisplayedMnemonic(mnemo.charAt(0));
            }
        }

        Component buttons = table.getButtons();
        Box box = Box.createVerticalBox();
        box.add(buttons);
        JScrollPane scroll = new JScrollPane (table);
        scroll.setName("ScrollPane-" + option.getName());
        box.add(scroll);
        add(box, BorderLayout.CENTER);

        setValue(option.getValue());
        table.setName("Table-" + option.getName());
        table.getTableHeader().setName("TableHeader-" + option.getName());

        int width = getInt("width", 50);
        int height = getInt("height", 100);
        setPreferredSize(new Dimension (width, height));
        if (getBoolean("configure", false)) {
            wrapper = ConfigWrapper.makeWrapper
                (table, option.getName() + ".table");
            Configuration.instance().register(wrapper);
        }

        String sWeight = getString("weight", null);
        if (sWeight != null) {
            try {
                weight = new Double(sWeight).doubleValue();
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "init()", ex);
                }
            }
        }

        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}
    // {{{ update

    /**
     * Update the display for a possibly new Option.
     *
     * @param   option The new option.
     */
    public void update(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        this.option = option;
        Object val = option.getValue();
        if (wrapper != null) {
            Configuration.instance().unregister(wrapper, true);
        }
        setValue(val);
        if (getBoolean("configure", false)) {
            wrapper = ConfigWrapper.makeWrapper
                (table, option.getName() + ".table");
            Configuration.instance().register(wrapper);
        }
    }

    // }}}
    // {{{ isModified

    /**
     * Query whether the edit has been modified, i.e. the current value of the
     * edit differs from the option's value.
     *
     * @return  True if the values in the edit and option differ.
     *
     * @since   0.99.0
     */
    public boolean isModified()
    {
        int rows = table.getRowCount();
        int columns = table.getColumnModel().getColumnCount();

        if (rows > 0 && columns > 0) {
            if (! (option.getValue() instanceof Object[][])) {
                return true;
            }
            Object [][] old = (Object[][]) option.getValue();
            if (old.length != rows) {
                return true;
            }
            for (int i = 0; i < rows; i++) {
                if (old[i].length != columns) {
                    return true;
                }
                for (int j = 0; j < columns; j++) {
                    Object val = table.getModel().getValueAt(i, j);
                    if ("".equals(val)) {
                        val = null;
                    }
                    if (val == null
                        ? (old[i][j] != null && ! "".equals(old[i][j]))
                        : ! val.equals(old[i][j])) {
                        return true;
                    }
                }
            }
            return false;
        } else {
            return option.getValue() != null;
        }
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the ValueEdit. If the value is
     * not valid, request the focus and bring up a dialog box explaining why.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "checkValue()", "");
        }
        if (table.isEditing()) {
            table.getCellEditor().stopCellEditing();
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the current value of the table.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }


        if (table.isEditing()) {
            table.getCellEditor().stopCellEditing();
        }
        int rows = table.getRowCount();
        int columns = table.getColumnModel().getColumnCount();

        if (rows > 0 && columns > 0) {
            Object[][] val = new Object[rows][];
            for (int i = 0; i < val.length; i++) {
                val[i] = new Object[columns];
                for (int j = 0; j < columns; j++) {
                    val[i][j] = table.getModel().getValueAt(i, j);
                }
            }
            option.setValue(val);
        } else {
            option.setValue(null);
        }
    }

    // }}}
    // {{{ getWeight

    /**
     * Get the y-weight for the TableOptionEdit.
     *
     * @return  1.0
     */
    public double getWeight()
    {
        return weight;
    }

    // }}}

    // {{{ setEnabled

    /**
     * Set the enabled flag.
     *
     * @param   enabled The flag to set.
     */
    public void setEnabled(boolean enabled)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setEnabled(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "enabled: " + enabled +
                       ", disabled: " + disabled);
        }
        if (disabled) {
            enabled = false;
        }

        this.enabled = enabled;
        if (table != null) {
            table.setEnabled(enabled);
        }
        if (label != null) {
            if (enabled) {
                label.setForeground(UIManager.getColor("textText"));
            } else {
                label.setForeground(UIManager.getColor("textInactiveText"));
            }
        }
    }

    // }}}
    // {{{ isEnabled

    /**
     * See if the ValueEdit is enabled.
     *
     * @return  True if it is enabled.
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    // }}}
    // {{{ requestFocus

    /**
     * Request the focus for the ValueEdit.
     */
    public void requestFocus()
    {
        table.requestFocus();
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods (overridables)
    //----------------------------------------------------------------------
    // {{{ setValue

    /**
     * Set the Edit's value. Same format as the data parameter in
     * {@link #TableOptionEdit the constructor}.
     *
     * @param   value   The value of the Edit.
     */
    public void setValue (Object value)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setValue(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "value: " + value);
        }
        boolean initial = table.getColumnCount() == 0;
        Object[][] data =
            (value == null ? new Object[0][] : (Object[][]) value);

        table.setModel (new Model (data, columns.length));
        if (initial) {
            for (int i = 0; i < columns.length; i++) {
                if (columns[i].width > 0) {
                    table.getColumnModel().getColumn(i)
                        .setPreferredWidth(columns[i].width);
                }
            }
        }
        if (table.getRowCount() > 0) {
            table.setRowSelectionInterval(0,0);
            table.setColumnSelectionInterval(0,0);
        } else {
            table.clearSelection();
        }
        for (int i = 0; i < columns.length; i++) {
            table.getColumnModel().getColumn(i).setIdentifier
                (columns[i].id);
        }

    }

    // }}}
    // {{{ createTable

    /**
     * Create the table for the option edit. Factory method for derived
     * classes.
     *
     * @return  The table for the TableOptionEdit.
     *
     * @since   0.99.0
     */
    protected Table createTable()
    {
        return new Table ();
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class Column

    /**
     * Structure that holds column data.
     */
    protected class Column
    {
        // {{{ variables

        /**
         * The identifier for the column.
         */
        public String id;

        /**
         * The name to display.
         */
        public String name;

        /**
         * The class of the column's values.
         */
        public Class clazz;

        /**
         * Default value for a new cell.
         */
        public Object dflt;

        /**
         * An option for the column's value.
         */
        public Option option;

        /**
         * Whether the column is disabled - i.e. not editable.
         */
        public boolean disabled;

        /**
         * Whether the column should start editing on any keypress.
         */
        public boolean directedit;

        /**
         * Default width of the column.
         */
        public int width = -1;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Column

        /**
         * Crate a new column.
         *
         * @param       id    The column identifier.
         */
        public Column(String id)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "Column(String)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "id: " + id);
            }

            String col =
                TableOptionEdit.this.option.getName() + ".column." + id;
            this.id = id;
            name = getString("column." + id + ".name", id);

            String type = getString("column." + id + ".type", "string");
            if ("string".equals(type)) {
                clazz = String.class;
                option = new StringOption
                    (col, null,
                     getBoolean("column." + id + ".nullok", false));
            } else if ("text".equals(type)) {
                clazz = String.class;
                option = new TextOption
                    (col, null,
                     getBoolean("column." + id + ".nullok", true));
            } else if ("int".equals(type)) {
                clazz = Integer.class;
                option = new IntegerOption
                    (col, 0,
                     getBoolean("column." + id + ".nullok", false),
                     getBoolean("column." + id + ".negativeok", false));
            } else if ("boolean".equals(type)) {
                clazz = Boolean.class;
                option = new BooleanOption
                    (col, false);
            }

            String df = getString("column." + id + ".default", null);
            if (df != null) {
                try {
                    dflt = clazz.getConstructor(new Class[] {String.class})
                        .newInstance(new String[] {df});
                } catch (Exception ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log(Log.ERR, "Column(String)", ex);
                    }
                }
            }

            disabled = getBoolean("column." + id + ".disabled", false);
            width = getInt("column." + id + ".cwidth", -1);

            this.directedit = getBoolean("column." + id + ".directedit",
                                    TableOptionEdit.this.directedit);

            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "Column(String)",
                           "New column: " + this);
            }

        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ toString

        /**
         * Get a string representation of the Column.
         *
         * @return      The string for the column.
         */
        public String toString()
        {
            return super.toString() + "[" +
                "id=" + id +
                ",name=" + name +
                ",clazz=" + clazz +
                ",dflt=" + dflt +
                ",disabled=" + disabled +
                ",directedit=" + directedit +
                "]";
        }

        // }}}
    }

    // }}}
    // {{{ class Model

    /**
     * The table model used by the edit.
     */
    protected class Model
        extends DefaultTableModel
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Model

        /**
         * Create a new Model.
         *
         * @param   data    The table data.
         * @param   col     The number of columns (javac doesn't allow access
         *                  to columns in this constructor).
         */
        public Model (Object[][] data, int col)
        {
            super(data, new Object[col]);
            Vector ids = new Vector ();
            for (int i = 0; i < columns.length; i++) {
                ids.addElement(columns[i].id);
            }
            setColumnIdentifiers(ids);

        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ getColumnClass

        /**
         * Get the class for a column.
         *
         * @param   column  The column index.
         *
         * @return  The class for the column.
         */
        public Class getColumnClass(int column)
        {
            return columns[column].clazz;
        }

        // }}}
        // {{{ getColumnName

        /**
         * Get the name for a column.
         *
         * @param   column  The column index.
         *
         * @return  The name for the column.
         */
        public String getColumnName(int column)
        {
            return columns[column].name;
        }

        // }}}
    }

    // }}}
    // {{{ class Table

    protected class Table
        extends JTable
    {
        // {{{ variables

        /**
         * The handler for the new/del/up/down actions.
         */
        protected ActionHandler handler;

        /**
         * The "edit row" button.
         */
        protected JButton editButton;

        /**
         * The "insert new row" button.
         */
        protected JButton newButton;

        /**
         * The "remove row" button.
         */
        protected JButton removeButton;

        /**
         * The "move row up" button.
         */
        protected JButton upButton;

        /**
         * The "move row down" button.
         */
        protected JButton downButton;

        /**
         * The box with the buttons.
         */
        protected Box buttonBox;

        /**
         * The dialog for the row edit.
         */
        protected OptionDialog editDialog;

        /**
         * The OptionSet for the dialog.
         */
        protected OptionSet rowOptions;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Table

        /**
         * Create a new Table.
         */
        public Table ()
        {
            JTableHeader header = new JTableHeader (getColumnModel()) {
                public boolean isFocusTraversable()
                {
                    return false;
                }
            };
            header.setRequestFocusEnabled(false);
            setTableHeader(header);
            setRowSelectionAllowed(selectrow);
            setAutoResizeMode(AUTO_RESIZE_OFF);
            // Renderers
            setDefaultRenderer(String.class, new WindowsTableCellRenderer ());
            WindowsTableCellRenderer wr = new WindowsTableCellRenderer ();
            wr.setHorizontalAlignment(JLabel.RIGHT);
            setDefaultRenderer(Integer.class, wr);
            setDefaultRenderer(Boolean.class, new BooleanRenderer ());
            // Editors
            setDefaultEditor(String.class, new StringEditor ());
            setDefaultEditor(Integer.class, new NumberEditor ());
            setDefaultEditor(Boolean.class, new BooleanEditor (this));
            // Override buggy F2 key binding
            this.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if ((getColumnClass(column) == String.class
                         || getColumnClass(column) == Integer.class)
                        && editCellAt(row, column, e)) {
                        Component other = Table.this.hasFocus()
                            ? getEditorComponent() : Table.this;
                        if (other != null) {
                            other.requestFocus();
                        }
                    }
                }
            }, "F2",
            KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0),
            WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            // Direct edit
            this.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    if (row >= getRowCount()) {
                        // Happens
                        return;
                    }
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if (columns[column].directedit
                        && e.getKeyChar() != 0
                        && e.getKeyChar() != 65535
                        && ! Character.isISOControl(e.getKeyChar())
                        && e.getKeyCode() != KeyEvent.VK_ENTER
                        && e.getKeyCode() != KeyEvent.VK_ESCAPE
                        && (e.getModifiers() &
                            (e.ALT_MASK | e.CTRL_MASK | e.META_MASK)) == 0
                        && (getColumnClass(column) == String.class
                            || getColumnClass(column) == Integer.class)
                        && editCellAt(row, column, e)) {
                        Component other = Table.this.hasFocus()
                            ? getEditorComponent() : Table.this;
                        if (other != null) {
                            other.requestFocus();
                        }
                        if (getCellEditor() != null) {
                            getCellEditor().shouldSelectCell(e);
                        }
                    }
                }
                public void keyTyped(KeyEvent e) {
                    // This is fixed in JDK 1.4
                    if (isEditing()
                        && ! Misc.isJavaVersionAtLeast(1, 4, -1)
                        && (getCellEditor() instanceof Editor)) {
                        ((Editor) getCellEditor()).insertChar
                            (e.getKeyChar());
                    }
                }
            });
            this.addInputMethodListener(new InputMethodListener() {
                public void inputMethodTextChanged(InputMethodEvent e)
                {
                    if (isEditing()
                        && (getCellEditor() instanceof Editor)) {
                        for (int i = 0; i < e.getCommittedCharacterCount();
                             i++) {
                            ((Editor) getCellEditor()).insertChar
                                (e.getText().next());
                        }
                    }
                }

                public void caretPositionChanged(InputMethodEvent e)
                {
                    // Noop
                }
            });
            // Edit booleans with ' ' key binding
            this.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if (getColumnClass(column) == Boolean.class
                        && editCellAt(row, column, e)) {
                        BooleanEditor editor = (BooleanEditor)
                            getCellEditor();
                        JCheckBox cb = (JCheckBox) editor.getComponent();
                        cb.setSelected(! cb.isSelected());
                        editor.stopCellEditing();
                    }
                }
            }, "space",
            KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0),
            WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

            // Remove default escape and return bindings.
            this.unregisterKeyboardAction
                (KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
            this.unregisterKeyboardAction
                (KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
            // Special version for JDK 1.3
            if (Misc.isJavaVersionAtLeast(1, 3, -1)) {
                Object inputMap = Reflector.safeCall
                    (this, "getInputMap", Integer.TYPE,
                     new Integer (WHEN_ANCESTOR_OF_FOCUSED_COMPONENT));
                if (inputMap != null) {
                    Reflector.safeCall
                        (inputMap, "put",
                         new Class[] {KeyStroke.class, Object.class},
                         new Object[] {
                             KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
                             "dummy"});
                    Reflector.safeCall
                        (inputMap, "put",
                         new Class[] {KeyStroke.class, Object.class},
                         new Object[] {
                             KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
                             "dummy"});
                }
            }
            // Special 1.4 hack
            if (Misc.isJavaVersionAtLeast(1, 4, -1)) {
                Reflector.safeCall(this, "setFocusTraversalKeys",
                                   new Class[] {Integer.TYPE, Set.class},
                                   new Object[] {
                                       // FORWARD_TRAVERSAL_KEYS
                                       new Integer (0),
                                       forwardKeys});
                Reflector.safeCall(this, "setFocusTraversalKeys",
                                   new Class[] {Integer.TYPE, Set.class},
                                   new Object[] {
                                       // BACKWARD_TRAVERSAL_KEYS
                                       new Integer (1),
                                       backwardKeys});
            }
            handler = new ActionHandler ();

            if (canedit) {
                this.registerKeyboardAction
                    (handler, "edit",
                     KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "edit",
                     KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,
                                            KeyEvent.ALT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (caninsert) {
                this.registerKeyboardAction
                    (handler, "insert",
                     KeyStroke.getKeyStroke(KeyEvent.VK_INSERT,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (canremove) {
                this.registerKeyboardAction
                    (handler, "remove",
                     KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (canmove) {
                if (multirow) {
                    this.registerKeyboardAction
                        (handler, "up",
                         KeyStroke.getKeyStroke(KeyEvent.VK_UP,
                                                KeyEvent.SHIFT_MASK |
                                                KeyEvent.CTRL_MASK),
                         WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                    this.registerKeyboardAction
                        (handler, "down",
                         KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,
                                                KeyEvent.SHIFT_MASK |
                                                KeyEvent.CTRL_MASK),
                         WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                } else {
                    this.registerKeyboardAction
                        (handler, "up",
                         KeyStroke.getKeyStroke(KeyEvent.VK_UP,
                                                KeyEvent.SHIFT_MASK),
                         WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                    this.registerKeyboardAction
                        (handler, "down",
                         KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,
                                                KeyEvent.SHIFT_MASK),
                         WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                }
            }
            if (multirow) {
                this.registerKeyboardAction
                    (handler, "ctrl-up",
                     KeyStroke.getKeyStroke(KeyEvent.VK_UP,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "ctrl-down",
                     KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "ctrl-left",
                     KeyStroke.getKeyStroke(KeyEvent.VK_LEFT,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "ctrl-right",
                     KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "toggle",
                     KeyStroke.getKeyStroke(KeyEvent.VK_SPACE,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (cutpaste) {
                this.registerKeyboardAction
                    (handler, "cut",
                     KeyStroke.getKeyStroke(KeyEvent.VK_X,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "copy",
                     KeyStroke.getKeyStroke(KeyEvent.VK_C,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "paste",
                     KeyStroke.getKeyStroke(KeyEvent.VK_V,
                                            KeyEvent.CTRL_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (getBoolean("autosize", true)) {
                getTableHeader().addMouseListener(new HeaderListener ());
            }
        }

        // }}}

        //------------------------------------------------------------------
        // Getters and setters
        //------------------------------------------------------------------
        // {{{ getButtons

        /**
         * Get the component with the buttons.
         *
         * @return  The button box.
         */
        public Component getButtons()
        {
            if (buttonBox == null) {
                buttonBox = Box.createHorizontalBox();
                if (caninsert) {
                    newButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    newButton.setName("NewButton-" + option.getName());
                    newButton.setRequestFocusEnabled(false);
                    newButton.setMargin(new Insets (0,0,0,0));
                    newButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.new.icon", null));
                    newButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.new.tooltip", null));
                    newButton.setActionCommand("insert");
                    newButton.addActionListener(handler);
                    buttonBox.add(newButton);
                }
                if (canedit) {
                    editButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    editButton.setName("EditButton-" + option.getName());
                    editButton.setRequestFocusEnabled(false);
                    editButton.setMargin(new Insets (0,0,0,0));
                    editButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.edit.icon", null));
                    editButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.edit.tooltip", null));
                    editButton.setActionCommand("edit");
                    editButton.addActionListener(handler);
                    buttonBox.add(editButton);
                }
                if (canremove) {
                    removeButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    removeButton.setName("RemoveButton-" + option.getName());
                    removeButton.setRequestFocusEnabled(false);
                    removeButton.setMargin(new Insets (0,0,0,0));
                    removeButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.remove.icon", null));
                    removeButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.remove.tooltip", null));
                    removeButton.setActionCommand("remove");
                    removeButton.addActionListener(handler);
                    buttonBox.add(removeButton);
                }
                if (canmove) {
                    upButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    upButton.setName("UpButton-" + option.getName());
                    upButton.setRequestFocusEnabled(false);
                    upButton.setMargin(new Insets (0,0,0,0));
                    upButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.up.icon", null));
                    upButton.setToolTipText
                        (Option.getResources().getString
                         (multirow ? "option.table.row.upmulti.tooltip"
                          : "option.table.row.up.tooltip", null));
                    upButton.setActionCommand("up");
                    upButton.addActionListener(handler);
                    buttonBox.add(upButton);

                    downButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    downButton.setName("DownButton-" + option.getName());
                    downButton.setRequestFocusEnabled(false);
                    downButton.setMargin(new Insets (0,0,0,0));
                    downButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.down.icon", null));
                    downButton.setToolTipText
                        (Option.getResources().getString
                         (multirow ? "option.table.row.downmulti.tooltip"
                          : "option.table.row.down.tooltip", null));
                    downButton.setActionCommand("down");
                    downButton.addActionListener(handler);
                    buttonBox.add(downButton);
                }
                buttonBox.add(Box.createHorizontalGlue());
            }
            return buttonBox;
        }

        // }}}

        //------------------------------------------------------------------
        // Actions
        //------------------------------------------------------------------
        // {{{ moveUp

        /**
         * Move the selected row up by one.
         */
        protected void moveUp()
        {
            int row = getAnchorRow();
            if (row > 0) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).moveRow(row, row, row - 1);
                setRowSelectionInterval(row - 1, row - 1);
            }
        }

        // }}}
        // {{{ moveDown

        /**
         * Move the selected row down by one.
         */
        protected void moveDown()
        {
            int row = getAnchorRow();
            if (row >= 0 && row < getRowCount() - 1) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).moveRow(row, row, row + 1);
                setRowSelectionInterval(row + 1, row + 1);
            }
        }

        // }}}
        // {{{ insertRow

        /**
         * Create and insert a new row.
         */
        protected void insertRow()
        {
            Object[] data;
            if (isEditing()) {
                if (getCellEditor() != null) {
                    if (! getCellEditor().stopCellEditing()) {
                        getCellEditor().cancelCellEditing();
                    }
                }
            }
            int row = Math.min(getRowCount(), getAnchorRow() + 1);
            if (canedit && ! directinsert) {
                initDialog(-1);
                int ret = editDialog.doModal
                    (SwingUtilities.windowForComponent(this), true);
                SwingUtil.focusLater(this);
                if (ret != 0) {
                    return;
                }
                data = new Object[columns.length];
                for (int i = 0; i < columns.length; i++) {
                    data[i] = rowOptions.getValue
                        (option.getName() + ".column." + columns[i].id, null);
                }
            } else {
                data = new Object[columns.length];
                for (int i = 0; i < columns.length; i++) {
                    data[i] = columns[i].dflt;
                }
                SwingUtil.focusLater(this);
            }

            ((Model) getModel()).insertRow(row, data);
            setRowSelectionInterval(row, row);
            int col = getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();
            col = Math.max(0, col);
            Table.this.scrollRectToVisible(getCellRect(row, col, true));
        }

        // }}}
        // {{{ removeRow

        /**
         * Remove the selected row.
         */
        protected void removeRow()
        {
            int row = getAnchorRow();
            if (row >= 0 && row < getRowCount()) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).removeRow(row);
                if (getRowCount() == 0) {
                    clearSelection();
                } else if (row == getRowCount()) {
                    setRowSelectionInterval(row - 1, row - 1);
                } else {
                    setRowSelectionInterval(row, row);
                }
            }
        }

        // }}}
        // {{{ editRow

        /**
         * Bring up a dialog for the selected row.
         */
        protected void editRow()
        {
            int row = getAnchorRow();
            if (row >= 0 && row < getRowCount()) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        if (! getCellEditor().stopCellEditing()) {
                            getCellEditor().cancelCellEditing();
                        }
                    }
                }
                initDialog(row);
                if (editDialog.doModal(SwingUtilities.windowForComponent(this),
                                       true) == 0) {
                    for (int i = 0; i < columns.length; i++) {
                        ((Model) getModel()).setValueAt
                            (rowOptions.getValue
                             (option.getName() + ".column." + columns[i].id,
                              null),
                             row, i);
                    }
                }
                SwingUtil.focusLater(this);
            }
        }

        // }}}
        // {{{ ctrlUp

        /**
         * Move the focus up without changing the selection.
         *
         * @since       0.99.0
         */
        protected void ctrlUp()
        {
            int row = getAnchorRow();
            if (row > 0) {
                getSelectionModel().setAnchorSelectionIndex(row - 1);
            }
        }

        // }}}
        // {{{ ctrlDown

        /**
         * Move the focus down without changing the selection.
         *
         * @since       0.99.0
         */
        protected void ctrlDown()
        {
            int row = getAnchorRow();
            if (row < getRowCount() - 1) {
                getSelectionModel().setAnchorSelectionIndex(row + 1);
            }
        }

        // }}}
        // {{{ ctrlLeft

        /**
         * Move the focus left without changing the selection.
         *
         * @since       0.99.0
         */
        protected void ctrlLeft()
        {
            int col = getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();
            if (col > 0) {
                getColumnModel().getSelectionModel()
                    .setAnchorSelectionIndex(col - 1);
            }
        }

        // }}}
        // {{{ ctrlRight

        /**
         * Move the focus right without changing the selection.
         *
         * @since       0.99.0
         */
        protected void ctrlRight()
        {
            int col = getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();
            if (col < getColumnCount() - 1) {
                getColumnModel().getSelectionModel()
                    .setAnchorSelectionIndex(col + 1);
            }
        }

        // }}}
        // {{{ toggleSelection

        /**
         * Toggle the selection of the focus row.
         *
         * @since       0.99.0
         */
        protected void toggleSelection()
        {
            int row = getAnchorRow();
            if (getSelectionModel().isSelectedIndex(row)) {
                getSelectionModel().removeSelectionInterval(row, row);
            } else {
                getSelectionModel().addSelectionInterval(row, row);
            }
        }

        // }}}
        // {{{ cut

        /**
         * Delete the selected rows and copy them to the clipboard.
         *
         * @since       0.99.0
         */
        protected void cut()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "cut()", "");
            }

            copy();
            int row = getAnchorRow();
            int rows = getRowCount();
            for (int i = rows - 1; i >= 0; i--) {
                if (getSelectionModel().isSelectedIndex(i)) {
                    ((Model) getModel()).removeRow(i);
                }
                if (i < row) {
                    row--;
                }
            }

            getSelectionModel().clearSelection();
            if (getRowCount() > 0) {
                getSelectionModel().setSelectionInterval(row, row);
            }
            getSelectionModel().setAnchorSelectionIndex(row);
        }

        // }}}
        // {{{ copy

        /**
         * Copy the selected rows to the clipboard.
         *
         * @since       0.99.0
         */
        protected void copy()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "copy()", "");
            }

            int columns = getColumnCount();
            int rows = getRowCount();

            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "copy()",
                           "columns: " + columns +
                           ", rows: " + rows);
            }

            if (rows <= 0 || columns <= 0) {
                return;
            }

            ListSelectionModel lsm = getSelectionModel();
            int srows = 0;

            for (int i = 0; i < rows; i++) {
                if (lsm.isSelectedIndex(i)) {
                    srows++;
                }
            }

            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "copy()",
                           "srows: " + srows);
            }

            Object[][] val = null;

            if (srows > 0) {
                val = new Object[srows][];
                int row = 0;
                for (int i = 0; i < rows && row < srows; i++) {
                    if (lsm.isSelectedIndex(i)) {
                        val[row] = copyRow(i);
                        row++;
                    }
                }
            }

            int row = getAnchorRow();
            int col = getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();

            String sel = "";
            if (row >= 0 && col >= 0) {
                Object tmp = getValueAt(row, col);
                if (tmp != null) {
                    sel = tmp.toString();
                }
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "copy()",
                           "sel: " + sel);
            }

            RowSelection rs = new RowSelection (val, sel);
            Toolkit.getDefaultToolkit().getSystemClipboard()
                .setContents(rs, rs);

        }

        // }}}
        // {{{ paste

        /**
         * Paste rows from the clipboard.
         *
         * @since       0.99.0
         */
        protected void paste()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "paste()", "");
            }
            Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();

            Object[][] data = null;
            try {
                Transferable transfer = cb.getContents(null);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "paste()",
                               "transfer: " + transfer);
                }
                if (transfer != null) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "unclip()",
                                   "contents: " + transfer);
                    }
                    if (transfer.isDataFlavorSupported(RowSelection.flavor)) {
                        data = (Object[][]) transfer.getTransferData
                            (RowSelection.flavor);
                    }
                }
            } catch (UnsupportedFlavorException ex) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "paste()", ex);
                }
            } catch (java.io.IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("unclip()", ex);
                }
            }
            if (data != null) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                int srow = getAnchorRow();
                int row = Math.min(getRowCount(), srow + 1);
                boolean mustClear = true;
                for (int i = 0; i < data.length; i++) {
                    if (columns.length == data[i].length) {
                        Object[] rowdata = pasteRow(data[i]);
                        if (rowdata != null) {
                            if (mustClear) {
                                getSelectionModel().clearSelection();
                                mustClear = false;
                            }
                            ((Model) getModel()).insertRow(row, rowdata);
                        }
                        getSelectionModel().addSelectionInterval(row, row);
                        row++;
                    }
                }
                getSelectionModel().setAnchorSelectionIndex(srow);
            }
        }

        // }}}

        //------------------------------------------------------------------
        // JTable methods
        //------------------------------------------------------------------
        // {{{ isManagingFocus

        public boolean isManagingFocus()
        {
            return false;
        }

        // }}}
        // {{{ isFocusTraversable

        public boolean isFocusTraversable()
        {
            return this.isEnabled();
        }

        // }}}
        // {{{ isRequestFocusEnabled

        public boolean isRequestFocusEnabled()
        {
            return this.isEnabled();
        }

        // }}}
        // {{{ isCellEditable

        public boolean isCellEditable(int row, int column)
        {
            return this.isEnabled()
                && ! columns[column].disabled
                && super.isCellEditable(row, column);
        }

        // }}}
        // {{{ setEnabled

        /**
         * Set the enabled state of the table. Also changes the buttons.
         *
         * @param   enabled The new enabled state.
         */
        public void setEnabled(boolean enabled)
        {
            super.setEnabled(enabled);
            if (buttonBox != null) {
                if (editButton != null) {
                    editButton.setEnabled(enabled);
                }
                if (newButton != null) {
                    newButton.setEnabled(enabled);
                }
                if (removeButton != null) {
                    removeButton.setEnabled(enabled);
                }
                if (upButton != null) {
                    upButton.setEnabled(enabled);
                }
                if (downButton != null) {
                    downButton.setEnabled(enabled);
                }
            }
        }

        // }}}

        //------------------------------------------------------------------
        // Helper methods (overridables)
        //------------------------------------------------------------------
        // {{{ initDialog

        protected void initDialog(int row)
        {
            String name = option.getName() + ".row";
            if (rowOptions == null) {
                rowOptions = new OptionSet (name);
                for (int i = 0; i < columns.length; i++) {
                    rowOptions.add(columns[i].option);
                }
            }
            for (int i = 0; i < columns.length; i++) {
                if (row >= 0) {
                    rowOptions.getOption
                        (option.getName() + ".column." + columns[i].id)
                        .setValue(getModel().getValueAt(row, i));
                } else {
                    rowOptions.getOption
                        (option.getName() + ".column." + columns[i].id)
                        .setValue(columns[i].dflt);
                }
            }
            if (editDialog == null) {
                // fixup option resources
                Properties props = new Properties ();
                if (Option.getResources().getString
                    ("option.group." + name + ".type", null) == null) {
                    props.put("option.group." + name + ".type", "plain");
                    StringBuffer sb = new StringBuffer ();
                    for (int i = 0; i < columns.length; i++) {
                        if (i > 0) {
                            sb.append(' ');
                        }
                        sb.append(option.getName());
                        sb.append(".column.");
                        sb.append(columns[i].id);
                        props.put("option." + option.getName() + ".column." +
                                  columns[i].id + ".label", columns[i].name);
                    }
                    props.put("option.group." + name + ".members",
                              sb.toString());
                }
                if (Option.getResources().getString
                    ("dialog." + name + ".title", null) == null) {
                    props.put("dialog." + name + ".title",
                              Option.getResources().getString
                              ("option.table.row.dialog.title",
                               "Edit table row"));
                }
                if (props.size() > 0) {
                    Option.getResources().addResources(props);
                }
                Window win = SwingUtilities.windowForComponent(this);
                if (win instanceof Frame) {
                    editDialog = new OptionDialog
                        (name, (Frame) win, rowOptions, name);
                } else if (win instanceof Dialog) {
                    editDialog = new OptionDialog
                        (name, (Dialog) win, rowOptions, name);
                } else {
                    editDialog = new OptionDialog
                        (name, (Frame) null, rowOptions, name);
                }
                SwingUtil.closeOnEscape(editDialog);
            } else {
                editDialog.update(rowOptions);
            }
        }

        // }}}
        // {{{ getAnchorRow

        /**
         * Get the row with the focus.
         *
         * @return      The row with the focus.
         *
         * @since       0.99.0
         */
        public int getAnchorRow()
        {
            return getSelectionModel().getAnchorSelectionIndex();
        }

        // }}}
        // {{{ copyRow

        /**
         * Copy the data of a single row for storage on the clibborad.
         *
         * @param       row     The row index.
         *
         * @return      The data to store.
         *
         * @since       0.99.0
         */
        protected Object[] copyRow(int row)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "copyRow(int)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "row: " + row);
            }
            int columns = getColumnCount();
            Object[] ret = new Object[columns];
            for (int i = 0; i < columns; i++) {
                Object tmp = getModel().getValueAt(row, i);
                try {
                    ret[i] = Reflector.call(tmp, "clone");
                } catch (Exception ex) {
                    ret[i] = tmp;
                }
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "copy()",
                               "ret[" + i + "]: " +
                               ret[i]);
                }
            }
            return ret;
        }

        // }}}
        // {{{ pasteRow

        /**
         * Copy (and optionally transform) the data for one row pasted from
         * the clipboard.
         *
         * @param       data    The pasted row data.
         *
         * @return      The transformed data or null if incompatible.
         *
         * @since       0.99.0
         */
        protected Object[] pasteRow(Object[] data)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "pasteRow(Object[])",
                           logger.level < Log.MTDDETAIL ? "" :
                           "data: " + data);
            }

            for (int i = 0; i < data.length; i++) {
                if (data[i] != null
                    && ! columns[i].clazz.isAssignableFrom
                    (data[i].getClass())) {
                    if (logger.level >= Log.MSG) {
                        logger.log(Log.MSG, "pasteRow(Object[])",
                                   "Incompatible rows, data[" + i + "]: " +
                                   data[i]);
                    }
                    return null;
                }
            }

            Object[] rowdata = (Object[]) data.clone();
            for (int i = 0; i < data.length; i++) {
                try {
                    rowdata[i] = Reflector.call(data[i], "clone");
                } catch (Exception ex) {
                }
            }

            return rowdata;
        }

        // }}}

        //------------------------------------------------------------------
        // Helper classes
        //------------------------------------------------------------------
        // {{{ class ActionHandler

        /**
         * Action listener for the table row actions.
         */
        protected class ActionHandler
            implements ActionListener
        {
            // {{{ actionPerformed

            /**
             * Perform the action.
             *
             * @param       e     The event containing the details.
             */
            public void actionPerformed(ActionEvent e)
            {
                String name = e.getActionCommand();
                if (logger.level >= Log.MTD) {
                    logger.log(Log.MTD, "actionPerformed(ActionEvent)",
                               logger.level < Log.MTDDETAIL ? "" :
                               "name: " + name);
                }
                if ("up".equals(name)) {
                    moveUp();
                } else if ("down".equals(name)) {
                    moveDown();
                } else if ("insert".equals(name)) {
                    insertRow();
                } else if ("remove".equals(name)) {
                    removeRow();
                } else if ("edit".equals(name)) {
                    editRow();
                } else if ("ctrl-up".equals(name)) {
                    ctrlUp();
                } else if ("ctrl-down".equals(name)) {
                    ctrlDown();
                } else if ("ctrl-left".equals(name)) {
                    ctrlLeft();
                } else if ("ctrl-right".equals(name)) {
                    ctrlRight();
                } else if ("toggle".equals(name)) {
                    toggleSelection();
                } else if ("cut".equals(name)) {
                    cut();
                } else if ("copy".equals(name)) {
                    copy();
                } else if ("paste".equals(name)) {
                    paste();
                }
            }

            // }}}
        }

        // }}}
        // {{{ class HeaderListener

        /**
         * Listener for mouse clicks on the Table's header.
         */
        private class HeaderListener extends MouseAdapter
        {
            // /**
             // * Delay changes so multiple clicks can be detected.
             // */
            // private DelayedAction delay = new DelayedAction (300);

            /**
             * Handle a mouse click.
             *
             * @param       e       The event contining the details.
             */
            public void mouseClicked(MouseEvent e)
            {
                // if (hlLogger.level >= Log.MTD) {
                    // hlLogger.log(Log.MTD, "mouseClicked(MouseEvent)",
                               // hlLogger.level < Log.MTDDETAIL ? "" :
                               // "e: " + e);
                // }
                if(e.getClickCount() == 2) {
                    final int column = Table.this.getColumnModel()
                        .getColumnIndexAtX(e.getX());
                    final int ccol =
                        Table.this.convertColumnIndexToModel(column);
                    if (column != -1) {
                        SwingUtil.autoSizeTableColumn(Table.this, column);
                    }
                }
            }
        }

        // }}}
    }

    // }}}
    // {{{ class BooleanRenderer

    /**
     * Rendere for boolean values.
     */
    private class BooleanRenderer
        extends JPanel
        implements TableCellRenderer
    {
        // {{{ variables

        /**
         * The CheckBox that renders the value.
         */
        private JCheckBox cb;

        // }}}

        // {{{ BooleanRenderer

        /**
         * Create a new BooleanRenderer.
         */
        public BooleanRenderer ()
        {
            this.setLayout(new BorderLayout());
            cb = new JCheckBox ();
            cb.setHorizontalAlignment(cb.CENTER);
            this.add(cb, BorderLayout.CENTER);
            this.setBorder(BorderFactory.createLineBorder(Color.black));
        }

        // }}}
        // {{{ getTableCellRendererComponent

        /**
         * Get the component that renders the value.
         *
         * @param   table   The table to paint.
         * @param   value   The value of the cell.
         * @param   isSelected     Whether the cell is selected.
         * @param   hasFocus       Whether the cell has the focus - ignored.
         * @param   row     The cell's row index.
         * @param   column  The cell's column index.
         *
         * @return      This
         */
        public Component getTableCellRendererComponent
            (JTable table, Object value,
             boolean isSelected, boolean hasFocus, int row, int column) {
	    if (isSelected) {
	        cb.setForeground(table.getSelectionForeground());
	        cb.setBackground(table.getSelectionBackground());
	    }
	    else {
	        cb.setForeground(table.getForeground());
	        cb.setBackground(table.getBackground());
	    }
            cb.setSelected((value instanceof Boolean)
                            && ((Boolean)value).booleanValue());
            this.setBorder(hasFocus ? Borders.STIPPLE : null);
            return this;
        }

        // }}}
    }

    // }}}
    // {{{ class Editor

    /**
     * Baseclass for the cell editors.
     */
    protected class Editor
        extends DefaultCellEditor
    {
        // {{{ variables

        /**
         * The column being edited.
         */
        Column column;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Editor(JTextField)

        /**
         * Create a new Editor.
         *
         * @param       com     The actual editor component.
         */
        public Editor (JTextField com)
        {
            super (com);
        }

        // }}}
        // {{{ Editor(JCheckBox)

        /**
         * Create a new Editor.
         *
         * @param       com     The actual editor component.
         */
        public Editor (JCheckBox com)
        {
            super (com);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
            this.column = columns[table.convertColumnIndexToModel(column)];
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ isCellEditable

        /**
         * Query whether a cell is editable.
         *
         * @param       anEvent The event initiating the edit.
         *
         * @return      False if anEvent is null, otherwise the result of the
         *              superclass' method.
         */
        public boolean isCellEditable(EventObject anEvent)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "isCellEditable(EventObject)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "anEvent: " + anEvent);
            }

            // Don't just start editing on any keypress
            if (anEvent == null) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "isCellEditable(EventObject)",
                               "isCellEditable: ignore");
                }
                return false;
            }
            return super.isCellEditable(anEvent);
        }

        // }}}
        // {{{ shouldSelectCell

        /**
         * Query whether the cell should be selected.
         *
         * @param       anEvent The event that started the editing.
         *
         * @return      Depends on the superclass.
         */
        public boolean shouldSelectCell(EventObject anEvent)
        {
            if (anEvent instanceof KeyEvent) {
                if (editorComponent instanceof JTextField) {
                    ((JTextField) editorComponent).setText(null);
                }
            }
            return super.shouldSelectCell(anEvent);
        }

        // }}}

        //------------------------------------------------------------------
        // Special methods
        //------------------------------------------------------------------
        // {{{ insertChar

        /**
         * Insert a character into the editor. Called if the table receives
         * a KEY_TYPED or INPUT_METHOD_TEXT_CHANGED event after creating the
         * Editor.
         *
         * @param       c    The character to insert.
         *
         * @since       0.99.0
         */
        public void insertChar(char c)
        {
            if (! Character.isISOControl(c)
                && (getComponent() instanceof JTextField)) {
                JTextField field = (JTextField) getComponent();
                try {
                    field.getDocument().insertString
                        (field.getCaretPosition(), "" + c, null);
                } catch (Exception ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log(Log.ERR, "insertString(String)", ex);
                    }
                }
            }
        }

        // }}}
    }

    // }}}
    // {{{ class BooleanEditor

    /**
     * Editor for Boolean table cell.
     */
    protected class BooleanEditor
        extends Editor
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ BooleanEditor

        /**
         * Create a new BooleanEditor.
         */
        public BooleanEditor (JTable table)
        {
            super (new JCheckBox ());
            JCheckBox cb = (JCheckBox) editorComponent;
            cb.setName("BooleanEditor-" + option.getName());
            cb.setForeground(table.getForeground());
            cb.setBackground(table.getBackground());
            cb.setHorizontalAlignment(cb.CENTER);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ shouldSelectCell

        /**
         * Query whether the cell should be selected.
         *
         * @param       anEvent The event that started the editing.
         *
         * @return      False if the event is a mouse event.
         */
        public boolean shouldSelectCellXXX(EventObject anEvent)
        {
            if (anEvent instanceof MouseEvent) {
                return false;
            }
            return super.shouldSelectCell(anEvent);
        }

        // }}}
    }

    // }}}
    // {{{ class StringEditor

    /**
     * Editor for String table cell.
     */
    public class StringEditor
        extends Editor
    {
        // {{{ variables

	Object value;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ StringEditor

        /**
         * Create a new StringEditor.
         */
        public StringEditor ()
        {
            super (new JTextField ());
            JTextField tf = (JTextField) editorComponent;
            tf.setName("StringEditor-" + option.getName());
            // Reset binding for return key
            tf.addKeyListener(new KeyAdapter () {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER
                        && e.getModifiers() == 0) {
                        stopCellEditing();
                        e.consume();
                    }
                }
            });
            // Bind escape to cancel editing
            tf.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelCellEditing();
                }
            },
            KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
            WHEN_FOCUSED);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ stopCellEditing

	public boolean stopCellEditing()
        {
	    value = super.getCellEditorValue();
	    if (! column.option.checkValue
                (null, "".equals(value) ? null : value)) {
		((JComponent) getComponent()).setBorder
                    (BorderFactory.createCompoundBorder
                     (BorderFactory.createLineBorder(Color.red),
                      BorderFactory.createEmptyBorder(1, 1, 1, 1)));
		return false;
	    }
	    return super.stopCellEditing();
	}

        // }}}
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
	    this.value = null;
            this.column = columns[column];
	    ((JComponent)getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.black),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ getCellEditorValue

	public Object getCellEditorValue()
        {
	    return value;
	}

        // }}}
    }

    // }}}
    // {{{ class NumberEditor

    /**
     * Editor for Number table cell.
     */
    public class NumberEditor
        extends Editor
    {
        // {{{ variables

	Class[] argTypes = new Class [] {String.class};
	java.lang.reflect.Constructor constructor;
	Object value;
        Column column;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ NumberEditor

        /**
         * Create a new NumberEditor.
         */
        public NumberEditor ()
        {
            super (new JTextField ());
            JTextField tf = (JTextField) editorComponent;
            tf.setName("NumberEditor-" + option.getName());
            tf.setHorizontalAlignment(tf.RIGHT);
            // This listener is necessary for Swing 1.1.1
            tf.addKeyListener(new KeyAdapter () {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER
                        && e.getModifiers() == 0) {
                        stopCellEditing();
                        e.consume();
                    }
                }
            });
            // Bind escape to cancel editing
            tf.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelCellEditing();
                }
            },
            KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
            WHEN_FOCUSED);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ stopCellEditing

	public boolean stopCellEditing()
        {
	    String s = (String) super.getCellEditorValue();
            boolean ok = false;
	    try {
		value = "".equals(s) ? null
                    : constructor.newInstance(new Object[]{s});
                ok = column.option.checkValue(null,  value);
	    }
	    catch (Exception e) {
            }
            if (! ok) {
		((JComponent) getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.red),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
		return false;
	    }
	    return super.stopCellEditing();
	}

        // }}}
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
	    this.value = null;
            this.column = columns[column];
	    ((JComponent)getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.black),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
	    try {
		Class type = table.getColumnClass(column);
		// Since our obligation is to produce a value which is
		// assignable for the required type it is OK to use the
		// String constructor for columns which are declared
		// to contain Objects. A String is an Object.
		if (type == Object.class) {
		    type = String.class;
		}
		constructor = type.getConstructor(argTypes);
	    }
	    catch (Exception e) {
		return null;
	    }
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ getCellEditorValue

	public Object getCellEditorValue()
        {
	    return value;
	}

        // }}}
    }

    // }}}
    // {{{ class RowSelection

    public static class RowSelection
        extends StringSelection
        implements Transferable, ClipboardOwner
    {
        // {{{ Variables

        /**
         * The data flavor for rows.
         */
        private static DataFlavor flavor =
            new DataFlavor (RowSelection.class, "qflib row selection");

        /**
         * The copied data.
         */
        private Object[] data;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ RowSelection

        /**
         * Create a new RowSelection.
         */
        public RowSelection (Object[][] data, String fallback)
        {
            super(fallback);
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "RowSelection(Object[][],String)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "data: " + data + ", " +
                           "fallback: " + fallback);
            }
            this.data = data;
        }

        // }}}

        //------------------------------------------------------------------
        // The Transferable interface
        //------------------------------------------------------------------
        // {{{ getTransferData

        /**
         * Returns an object which represents the data to be transferred.
         *
         * @param   flavor  The flavor to get the data for.
         *
         * @return  The transfer data.
         */
        public Object getTransferData(DataFlavor flavor)
            throws IOException, UnsupportedFlavorException
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "getTransferData(DataFlavor)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "flavor: " + flavor +
                           "[" + flavor.getHumanPresentableName() +
                           ",mime=" + flavor.getMimeType() +
                           "]");
            }
            if (this.flavor.equals(flavor)) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "getTransferData(DataFlavor)",
                               "flavor is RowSelection.flavor");
                }
                return data;
            }
            return super.getTransferData(flavor);
        }

        // }}}
        // {{{ getTransferDataFlavors

        /**
         * Returns an array of DataFlavor objects indicating the flavors
         * the data can be provided in.
         *
         * @return  The available data flavors.
         */
        public DataFlavor[] getTransferDataFlavors()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "getTransferDataFlavors()", "");
            }
            if (data == null) {
                return super.getTransferDataFlavors();
            }
            DataFlavor[] dflt = super.getTransferDataFlavors();
            DataFlavor[] ret = new DataFlavor[dflt.length + 1];
            System.arraycopy(dflt, 0, ret, 0, dflt.length);
            ret[dflt.length] = flavor;
            return ret;
        }

        // }}}
        // {{{ isDataFlavorSupported

        /**
         * Returns whether or not the specified data flavor is supported
         * for this object.
         *
         * @param   flavor  The flavor to check.
         *
         * @return  True, if the flavor is supported.
         */
        public boolean isDataFlavorSupported(DataFlavor flavor)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "isDataFlavorSupported(DataFlavor)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "flavor: " + flavor);
            }
            return (data != null && this.flavor.equals(flavor))
                || super.isDataFlavorSupported(flavor);
        }

        // }}}
    }

    // }}}
}
