// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Dialog;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

// {{{ doc

/**
 * This specialized {@link OptionDialog OptionDialog} is used to to display a
 * sequence of Option panes which the user can travers with a "Next" and a
 * "Back" button.
 *
 * To be useful at all, the main group of the dialog must be a {@link
 * WizardOptionGroup WizardOptionGroup}.
 *
 * By default the WizardDialog provides the buttons "Back", "Next" and
 * "Cancel". If the first pane is displayed, the "Back" button is disabled and
 * on the final pane the "Next" button changes to "Finish". The resources
 * names for the "Back" and "Next" buttons are <code>wizard.back</code> and
 * <code>wizard.next</code>. If you use custom buttons other than these you
 * must* implement all the pane switching yourself.
 *
 * Please see the documentation of the {@link OptionDialog OptionDialog} and
 * {@link de.qfs.lib.gui.ModalDialog ModalDialog} classes for the resources
 * used.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */

// }}}
public class WizardDialog
    extends OptionDialog
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.WizardDialog");

    /**
     * The default button names.
     */
    protected final static String[] wizardButtons =
    {"wizard.back", "wizard.next", "cancel"};

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ WizardDialog(String,Frame,OptionSet,String)

    /**
     * Create a new WizardDialog.
     *
     * @param   name    The resource name of the dialog.
     * @param   parent  The parent Frame.
     * @param   options The Options to edit.
     * @param   root    The resource name of the root of the OptionGroups. If
     *                  null, the configurable name of the OptionSet will be
     *                  used.
     *
     * @since   0.98.0
     */
    public WizardDialog (String name, Frame parent, OptionSet options,
                         String root)
    {
        super(name, parent, options, root);
        if (Option.getResources().getString(resName("name"), null) == null) {
            setName("WizardDialog-" + root + "-" +
                    options.getConfigurableName());
        }
    }

    // }}}
    // {{{ WizardDialog(String,Dialog,OptionSet,String)

    /**
     * Create a new WizardDialog.
     *
     * @param   name    The resource name of the dialog.
     * @param   parent  The parent Dialog.
     * @param   options The Options to edit.
     * @param   root    The resource name of the root of the OptionGroups. If
     *                  null, the configurable name of the OptionSet will be
     *                  used.
     *
     * @since   0.98.0
     */
    public WizardDialog (String name, Dialog parent, OptionSet options,
                         String root)
    {
        super(name, parent, options, root);
        if (Option.getResources().getString(resName("name"), null) == null) {
            setName("WizardDialog-" + root + "-" +
                    options.getConfigurableName());
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ init

    public void init(MapResourceBundle rb)
    {
        super.init(rb);
        if (getButton("wizard.next") != null) {
            getButton("wizard.next").addActionListener(createNextHandler());
        }
        if (getButton("wizard.back") != null) {
            getButton("wizard.back").addActionListener(createBackHandler());
        }
        enableButtons();
    }

    // }}}
    // {{{ initButtonNames

    /**
     * Get the names of the dialog's buttons. The default action is to read
     * the names from the resources.
     *
     * @param   rb      The resources to use.
     *
     * @return  The button names.
     */
    protected String[] initButtonNames(MapResourceBundle rb)
    {
        String buttonRes = rb.getString(resName("buttons"), null);
        if (buttonRes == null) {
            return wizardButtons;
        }
        return super.initButtonNames(rb);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ createNextHandler

    /**
     * Create an action listener for the "Next" button.
     *
     * @return  An action listener for the "Next" button.
     */
    protected ActionListener createNextHandler()
    {
        return new NextHandler ();
    }

    // }}}
    // {{{ createBackHandler

    /**
     * Create an action listener for the "Back" button.
     *
     * @return  An action listener for the "Back" button.
     */
    protected ActionListener createBackHandler()
    {
        return new BackHandler ();
    }

    // }}}
    // {{{ enableButtons

    /**
     * Enable/disable the buttons and update their labels according to the
     * current pane.
     */
    protected void enableButtons()
    {
        if (optionGroup instanceof WizardOptionGroup) {
            if (getButton("wizard.next") != null) {
                String label = Option.getResources().getString
                    ("dialog.button.wizard." +
                     (((WizardOptionGroup) optionGroup).isLast()
                      ? "finish" : "next") +
                     ".label", "Next");
                String mnemo = Option.getResources().getString
                    ("dialog.button.wizard." +
                     (((WizardOptionGroup) optionGroup).isLast()
                      ? "finish" : "next") +
                     ".mnemonic", "n");
                getButton("wizard.next").setText(label);
                getButton("wizard.next").setMnemonic(mnemo.charAt(0));
                getButton("wizard.next").setDefaultCapable(true);
            }
            if (getButton("wizard.back") != null) {
                getButton("wizard.back").setEnabled
                    (! ((WizardOptionGroup) optionGroup).isFirst());
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class NextHandler

    /**
     * Default ActionListener for the "Next" burron.
     */
    protected class NextHandler
        implements ActionListener
    {
        public void actionPerformed (ActionEvent e)
        {
            if (optionGroup instanceof WizardOptionGroup) {
                boolean finish = ((WizardOptionGroup) optionGroup).isLast();
                if (((WizardOptionGroup) optionGroup).nextPane()) {
                    if (finish) {
                        setVisible(false);
                    } else {
                        enableButtons();
                    }
                }
            }
        }
    }

    // }}}
    // {{{ class BackHandler

    /**
     * Default ActionListener for the "Back" burron.
     */
    protected class BackHandler
        implements ActionListener
    {
        public void actionPerformed (ActionEvent e)
        {
            if (optionGroup instanceof WizardOptionGroup) {
                ((WizardOptionGroup) optionGroup).previousPane();
                enableButtons();
            }
        }
    }

    // }}}
}
