// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;

import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.FocusManager;
import javax.swing.JComponent;
import javax.swing.JPanel;

import de.qfs.lib.gui.Grid;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * A WizardOptionGroup displays a sequence of option panes like they are
 * typically used in wizard type dialogs.
 *
 * When switching to the next group, the current values are checked for
 * validity and stored in the respective Options. The methods {@link #nextPane
 * nextPane} and {@link #previousPane previousPane} are used to switch between
 * panes. If you need to perform custom tasks when switching, either override
 * these, or implement an {@link OptionValidator OptionValidator} or an {@link
 * java.util.Observer Observer} and register it with the relevant Option.
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.type</code></TD>
 * <TD>A WizardOptionGroup is designated by the type <code>wizard</code>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the group will be disabled and remain so even if
 * {@link OptionGroup#setEnabled setEnabled(true)} is called (optional,
 * default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.numpanes</code></TD>
 * <TD>The number of panes in the group.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.pane<em>n</em>.group</code></TD>
 * <TD>The name of the OptionGroup to put on the n<sup>th</sup> pane,
 * where n is between 1 and numpanes.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.frame</code></TD>
 * <TD>Whether the group should have a framed border (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.title</code></TD>
 * <TD>The title for a framed border of the group. Ignored if no frame is
 * specified(optional, default empty).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */

// }}}
public class WizardOptionGroup
    extends OptionGroup
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.option.WizardOptionGroup");

    /**
     * The JPanel Component.
     */
    protected JPanel pane;

    /**
     * The layout of the panel.
     */
    protected CardLayout layout;

    /**
     * The number of panes.
     */
    protected int numpanes;

    /**
     * The index of the current pane.
     */
    protected int current;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ WizardOptionGroup()

    /**
     * Create a new WizardOptionGroup.
     */
    public WizardOptionGroup ()
    {
    }

    // }}}
    // {{{ WizardOptionGroup(String,OptionSet)

    /**
     * Create a new WizardOptionGroup.
     *
     * @param   name    The name of the group.
     * @param   options The Options to edit.
     */
    public WizardOptionGroup (String name, OptionSet options)
    {
        super (name, options);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "WizardOptionGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        createMembers();
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getComponent

    /**
     * Get the AWT Component for the OptionGroup.
     *
     * @return  The Component.
     */
    public Component getComponent()
    {
        if (pane == null) {
            makePane();
        }
        return pane;
    }

    // }}}
    // {{{ nextPane

    /**
     * Switch to the next pane (or finish) if possible.
     *
     * @return  True if OK, false if aborted.
     */
    public boolean nextPane()
    {
        OptionGroup group = (OptionGroup) members.get(current);
        boolean ret = group.checkValues();
        if (! ret) {
            return false;
        }
        group.readValues();
        if (current < numpanes - 1) {
            current++;
            ((OptionGroup) members.get(current)).update(options);
            layout.next(pane);
            resetFocus();
        }
        return true;
    }

    // }}}
    // {{{ previousPane

    /**
     * Switch to the previous pane if possible.
     *
     * @return  True if OK, false if aborted.
     */
    public boolean previousPane()
    {
        if (current > 0) {
            current--;
            ((OptionGroup) members.get(current)).update(options);
            layout.previous(pane);
            resetFocus();
            return true;
        }
        return false;
    }

    // }}}

    // {{{ isFirst

    /**
     * Test whether the first pane is showing.
     *
     * @return  True if the first pane is currently showing.
     */
    public boolean isFirst()
    {
        return current == 0;
    }

    // }}}
    // {{{ isLast

    /**
     * Test whether the last pane is showing.
     *
     * @return  True if the last pane is currently showing.
     */
    public boolean isLast()
    {
        return current == numpanes - 1;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ createMembers

    /**
     * Create the panes for the TabbedOptionGroup.
     */
    protected void createMembers()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "createMembers()", "");
        }
        MapResourceBundle rb = Option.getResources();
        members = new ArrayList();

        numpanes = getInt("numpanes", -1);
        if (numpanes < 0) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "createMembers()",
                           "No numpanes for WizardOptionGroup " + name);
                return;
            }
        }

        for (int i = 1; i <= numpanes; i++) {
            String group = getString("pane" + i + ".group", null);
            if (group == null) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "createMembers()",
                               "No group for pane " + i +
                               " in WizardOptionGroup " + name);
                }
            } else {
                OptionGroup og = OptionGroup.makeGroup(group, options);
                if (og != null) {
                    og.setParentGroup(this);
                    members.add(og);
                }
            }
        }
    }

    // }}}
    // {{{ makePane

    /**
     * Create the JPanel for the pane members.
     */
    protected void makePane()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makePane()", "");
        }
        MapResourceBundle rb = Option.getResources();

        pane = new JPanel ();
        pane.setName("OptionGroup-" + name + "-" +
                     options.getConfigurableName());

        layout = new CardLayout ();
        pane.setLayout(layout);

        for (int i = 0; i < members.size(); i++) {
            OptionGroup group = (OptionGroup) members.get(i);
            Component com = group.getComponent();
            if (com instanceof Grid
                && ((Grid) com).needGlue()) {
                ((Grid) com).addGlue();
            }
            if (! (com instanceof JComponent)
                // Compound border doesn't quite work
                || ((JComponent) com).getBorder() != null) {
                JPanel panel = new JPanel ();
                panel.setLayout(new BorderLayout ());
                panel.add(com, BorderLayout.CENTER);
                com = panel;
            }
            ((JComponent) com).setBorder
                (BorderFactory.createEmptyBorder(5, 5, 5, 5));
            pane.add(com, group.getName());
        }

        if (disabled) {
            setEnabled(false);
        }

        makeBorder(pane);
    }

    // }}}
    // {{{ resetFocus

    /**
     * Reset the focus to the first component in the pane after switching
     * panes.
     */
    protected void resetFocus()
    {
        FocusManager.getCurrentManager().focusNextComponent(pane);
    }

    // }}}
}
