// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.Serializable;

import java.util.Hashtable;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * An implementation of the Enum pattern to provide enumerations for Java.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public class Enum
    implements Serializable, Cloneable , Comparable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =  new Logger ("de.qfs.lib.util.Enum");

    /**
     * The ID for serialization.
     */
    static final long serialVersionUID = -8444370131512049067L;

    /**
     * A registry for created enumerations. Required to avoid creation of
     * duplicate Enums and resolve objects during deserialization. Note that
     * the latter doesn't work with JDK 1.1.
     */
    private static Hashtable registry = new Hashtable ();

    /**
     * The enumeration value.
     */
    private int value;

    /**
     * The String representation.
     */
    private String name;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ Enum

    /**
     * Create a new Enum.
     *
     * @param   value   The value for the Enum.
     * @param   name    The Name for the Enum.
     *
     * @throws  IllegalArgumentException if an Enum of the same class with
     *          the same id is already registered.
     */
    public Enum (int value, String name)
        throws IllegalArgumentException
    {
        register(value);
        this.value = value;
        this.name = name;
    }

    // }}}

    //----------------------------------------------------------------------
    // Public static methods
    //----------------------------------------------------------------------
    // {{{ lookup

    /**
     * Get an Enum defined by its type and value.
     *
     * @param   type    The type as returned by getTypeName.
     * @param   value   The value of the enum.
     *
     * @return  The enum if known or null.
     */
    public static Enum lookup(String type, int value)
    {
        return (Enum) registry.get(type+value);
    }

    // }}}

    //----------------------------------------------------------------------
    // Miscellaneous
    //----------------------------------------------------------------------
    // {{{ getValue

    /**
     * Get the int value.
     *
     * @return  The int value.
     */
    public final int getValue()
    {
        return value;
    }

    // }}}
    // {{{ getName

    /**
     * Get the name of the Enum.
     *
     * @return  The name of the Enum.
     */
    public final String getName()
    {
        return name;
    }

    // }}}
    // {{{ toString

    /**
     * Get a String representation.
     *
     * @return  The String representation.
     */
    public String toString()
    {
        return getTypeName() + "[" + name + "]";
    }

    // }}}
    // {{{ clone

    /**
     * Cloning conflicts with the identity requirement, so the original Enum
     * is returned.
     *
     * @return  This.
     */
    public Object clone()
    {
        return this;
    }

    // }}}
    // {{{ readResolve

    /**
     * Resolve an Object during derserialization. Will only be called for
     * JDK 1.2 and up.
     *
     * @return  The registered Enum if it exists.
     */
    public Object readResolve()
    {
        String name = getTypeName() + value;
        synchronized (registry) {
            Object ret = registry.get(name);
            if (ret == null) {
                registry.put(name, this);
            }
            return ret == null ? this : ret;
        }
    }

    // }}}
    // {{{ equals

    /**
     * Compare two enums. For JDK 1.2 and above object identity is used. For
     * JDK 1.1 an additional test for type/value equality is made.
     *
     * @param   o       The object to compare to.
     *
     * @return  True if the Enums are equal.
     */
    public final boolean equals(Object o)
    {
        return this == o
            ;
    }

    // }}}
    // {{{ hashCode

    /**
     * Finalize hashCode to avoid accidental overriding.
     *
     * @return  The default hashCode.
     */
    public final int hashCode()
    {
        return super.hashCode();
    }

    // }}}
    // {{{ compareTo

    /**
     * Compare the Enum to another enum.
     *
     * @param   o       The enum to compare against.
     *
     * @return  <0 if value is less, 0 if equal or >0 if value is greate than
     *          the other enum's value.
     */
    public int compareTo(Object o)
    {
        return value - ((Enum) o).getValue();
    }

    // }}}

    //----------------------------------------------------------------------
    // Overridables
    //----------------------------------------------------------------------
    // {{{ getTypeName

    /**
     * Get the type name for the Enum. Can be implemented in a derived
     * class.
     *
     * @return  The name of the Enum type.
     */
    public String getTypeName()
    {
        return getClass().getName();
    }

    // }}}

    //----------------------------------------------------------------------
    // Private helpers
    //----------------------------------------------------------------------
    // {{{ register

    /**
     * Store an enum in the registry.
     *
     * @param   value   The value of the enum.
     *
     * @throws  IllegalArgumentException if an Enum of the same class with
     *          the same id is already registered.
     */
    private void register(int value)
        throws IllegalArgumentException
    {
        String name = getTypeName() + value;
        synchronized (registry) {
            if (registry.get(name) != null) {
                throw new IllegalArgumentException
                    ("Value " + value + " already defined for type "
                     + getTypeName());
            }
            registry.put(name, this);
        }
    }

    // }}}
}
