// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This is a special MultiMap with a list-based value set. It retains the
 * order in which values for a key are added. The result of getAll() is
 * derived from ArrayList.
 *
 * @author      Gregor Schmid
 */
public class ListBasedMultiMap
    extends MultiMap
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.ListBasedMultiMap");

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ ListBasedMultiMap

    /**
     * Create a new ListBasedMultiMap.
     */
    public ListBasedMultiMap()
    {
    }

    // }}}
    // {{{ ListBasedMultiMap

    /**
     * Create a new ListBasedMultiMap.
     *
     * @param   initialCapacity   The initial capacity of the
     *                            ListBasedMultiMap.
     */
    public ListBasedMultiMap(int initialCapacity)
    {
        super (initialCapacity);
    }

    // }}}
    // {{{ ListBasedMultiMap

    /**
     * Create a new ListBasedMultiMap and initialize it with all the mappings
     * from the passed map.
     *
     * @param   map     The map that initializes this ListBasedMultiMap.
     */
    public ListBasedMultiMap (Map map)
    {
        super (map);
    }

    // }}}

    //----------------------------------------------------------------------
    // Overridden methods
    //----------------------------------------------------------------------
    // {{{ createValueSet

    /**
     * Overidden to return list based set.
     */
    public ValueSet createValueSet()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "createValueSet()", "");
        }
        return new ArraySet ();
    }

    // }}}
    // {{{ copyResultSet

    /**
     * Create a new result set for the getAll method. Can be overridden to
     * provide a matching implementation.
     *
     * @param   result  The internal result set that should not be exposed.
     *
     * @return  A copy of the result set.
     */
    protected Set copyResultSet(ValueSet result)
    {
        return new ArraySet (result);
    }

    // }}}
    // {{{ createResultSet

    /**
     * Create a new result set for a single element result of the getAll
     * method. Can be overridden to provide a matching implementation.
     *
     * @param   result  The result element.
     *
     * @return  A new result set.
     */
    protected Set createResultSet(Object result)
    {
        ArraySet set = new ArraySet ();
        set.add(result);
        return set;
    }

    // }}}

    //----------------------------------------------------------------------
    // Implementation of the list-based value set.
    //----------------------------------------------------------------------
    // {{{ class ArraySet

    /**
     * Implementation of the list-based set. Overrides the add method to
     * prevent multiple ocurrence of the same object.
     */
    private class ArraySet
        extends ArrayList
        implements Set, ValueSet
    {
        //--------------------------------------------------------------
        // Constructor
        //--------------------------------------------------------------
        // {{{ ArraySet

        /**
         * Create a new ArraySet.
         */
        public ArraySet ()
        {
        }

        // }}}
        // {{{ ArraySet(Collection)

        /**
         * Create a new ArraySet.
         */
        public ArraySet (Collection c)
        {
            super (c);
        }

        // }}}

        //------------------------------------------------------------------
        // Overridden methods
        //------------------------------------------------------------------
        // {{{ add

        /**
         * Add an object to the end of the list. If the object is already
         * in the set, move it to the endo of the list.
         *
         * @param       o       The object to add.
         *
         * @return      True if the object was newly added.
         */
        public boolean add(Object o) {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "add(Object)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "o: " + o);
            }
            boolean ret = ! contains(o);
            super.remove(o);
            super.add(o);
            return ret;
        }

        // }}}
    }

    // }}}
}
