// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.Hashtable;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;


// }}}

// {{{ doc

/**
 * This is a cache that stores only a certain number of resources. If more
 * resources are added, only the most recently used ones are kept. <p>
 *
 * The MRUCache is based on a {@link MultiMap MultiMap}, so it can store
 * different kinds of resources distinguished by a key. The size limit of the
 * cache applies to the total number of resorces stored independent of their
 * type. <p>
 *
 * When a resource is retrieved with {@link #get(Object) get}, it is
 * automatically removed from the MRUCache. <p>
 *
 * If you just want to store resources of one kind, there is a simplified
 * interface of {@link #get() get()} and {@link #put(Object) put(Object)}
 * where you don't need to provide a key. <p>
 *
 * You cannot store the same resource more than once, not even using different
 * keys. <p>
 *
 * @author      Gregor Schmid
 * @since       0.98.0
 */

// }}}
public class MRUCache
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.MRUCache");

    /**
     * Unique key for simplified use.
     */
    private static Object uniqueKey = new Object ();

    /**
     * The map used for the cache.
     */
    private MultiMap cache;

    /**
     * The list of most recently used resources.
     */
    private Vector mruList;

    /**
     * Parallel vector for the keys of the resources on the mruList.
     */
    private Vector keys;

    /**
     * The maximum size of the MRU list.
     */
    private int mruSize;

    /**
     * The listeners for dropped resources.
     */
    private Vector listeners = new Vector ();

    // }}}

    // {{{ constructor

    /**
     * Create a new MRUCache.
     *
     * @param   size    The maximum number of resources to keep.
     */
    public MRUCache (int size)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "MRUCache(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "size: " + size);
        }
        mruSize = size;
        cache = new MultiMap (size);
        mruList = new Vector (size);
        keys = new Vector (size);
    }

    // }}}

    // {{{ get

    /**
     * Get a resource from the MRUCache.
     *
     * @return  An arbitrary resource or null if none is available.
     */
    public synchronized Object get()
    {
        return get(uniqueKey);
    }

    /**
     * Get a resource from the MRUCache.
     *
     * @param   key     The key for the resource.
     *
     * @return  An arbitrary resource stored for that key or null if none is
     *          available.
     */
    public synchronized Object get(Object key)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "get(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key);
        }
        Object resource = cache.get(key);
        if (resource != null) {
            cache.remove(key, resource);

            int idx = mruList.indexOf(resource);

            if (idx < 0) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "get(Object)",
                               "OOPS, object not on mruList");
                }
            } else {
                keys.removeElementAt(idx);
                mruList.removeElementAt(idx);
            }
        }
        return resource;
    }

    // }}}
    // {{{ put

    /**
     * Store a resource in the MRUCache.
     *
     * @param   resource The resource to store.
     */
    public synchronized void put(Object resource)
    {
        put(uniqueKey, resource);
    }

    /**
     * Store a resource in the MRUCache.
     *
     * @param   key      The key to store the resource under.
     * @param   resource The resource to store.
     */
    public synchronized void put(Object key, Object resource)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "put(Object,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "resource: " + resource);
        }

        int idx = mruList.indexOf(resource);
        if (idx >= 0) {
            if (logger.level >= Log.WRN) {
                logger.dumpStack(Log.WRN, "put(Object,Object)",
                                 "Object already in MRUCache: " + resource);
            }
            Object oldkey = keys.elementAt(idx);
            cache.remove(oldkey, resource);
            keys.removeElementAt(idx);
            mruList.removeElementAt(idx);
        }

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "put(Object,Object)",
                       "storing " + resource + "(hash: " +
                       resource.hashCode() +
                       ") in reverseMap");
        }

        if (mruList.size() >= mruSize) {
            Object dropped = mruList.elementAt(mruList.size() - 1);
            mruList.removeElementAt(mruList.size() - 1);
            Object oldkey = keys.elementAt(keys.size() - 1);
            keys.removeElementAt(keys.size() - 1);
            cache.remove(oldkey, dropped);
            fireResourceDropped(oldkey, dropped);
        }
        mruList.insertElementAt(resource, 0);
        keys.insertElementAt(key, 0);

        cache.put(key, resource);
    }

    // }}}
    // {{{ remove

    /**
     * Remove a resource from the cache.
     *
     * @param   resource        The resource to remove.
     */
    public void remove(Object resource)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "remove(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "resource: " + resource);
        }

        int idx = mruList.indexOf(resource);
        if (idx < 0) {
            if (logger.level >= Log.WRN) {
                logger.dumpStack(Log.WRN, "remove(Object)",
                                 "Object not in MRUCache: " + resource);
            }
            return;
        }
        Object oldkey = keys.elementAt(idx);
        cache.remove(oldkey, resource);
        keys.removeElementAt(idx);
        mruList.removeElementAt(idx);

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "remove(Object)",
                       "removing " + resource + "(hash: " +
                       resource.hashCode() +
                       ") from reverseMap");
        }
    }

    // }}}
    // {{{ contains

    /**
     * Test where the MRUCache contains a resource.
     *
     * @param   resource        The resource to look for.
     *
     * @return  True if the cache contains the resource.
     */
    public boolean contains(Object resource)
    {
        return mruList.contains(resource);
    }

    // }}}
    // {{{ getMaximumSize

    /**
     * Query the maximum size of the MRUCache.
     *
     * @return  The maximum number of resources stored by this cache.
     */
    public synchronized int getMaximumSize()
    {
        return mruSize;
    }

    // }}}
    // {{{ setMaximumSize

    /**
     * Set the maximum size of the MRUCache.
     *
     * @param size  The maximum number of resources allowed for this cache.
     */
    public synchronized void setMaximumSize(int size)
    {
        mruSize = size;

        while (mruList.size() > mruSize) {
            Object dropped = mruList.elementAt(mruList.size() - 1);
            mruList.removeElementAt(mruList.size() - 1);
            Object oldkey = keys.elementAt(keys.size() - 1);
            keys.removeElementAt(keys.size() - 1);
            cache.remove(oldkey, dropped);
            fireResourceDropped(oldkey, dropped);
        }
    }

    // }}}
    // {{{ getCurrentSize

    /**
     * Query the current size of the MRUCache.
     *
     * @return  The current number of resources stored by this cache.
     */
    public synchronized int getCurrentSize()
    {
        return mruList.size();
    }

    // }}}

    // {{{ addMRUCacheListener

    /**
     * Add a listener for dropped resources to the MRUCache.
     *
     * @param   listener        The listener to add.
     */
    public synchronized void addMRUCacheListener(MRUCacheListener listener)
    {
        // avoid duplicates
        listeners.removeElement(listener);
        listeners.insertElementAt(listener, 0);
    }

    // }}}
    // {{{ removeMRUCacheListener

    /**
     * Remove a listener for dropped resources from the MRUCache.
     *
     * @param   listener        The listener to remove.
     */
    public synchronized void removeMRUCacheListener(MRUCacheListener listener)
    {
        listeners.removeElement(listener);
    }

    // }}}
    // {{{ fireResourceDropped

    /**
     * Notify all listeners about a resource dropped from the cache.
     *
     * @param   key     The key the resource was stored under.
     * @param   resource        The dropped resource.
     */
    protected void fireResourceDropped(Object key, Object resource)
    {
        MRUCacheEvent event = new MRUCacheEvent (this, key, resource);

        Vector tmp;
        synchronized(this) {
            tmp = (Vector) listeners.clone();
        }
        for (int i = 0; i < tmp.size(); i++) {
            ((MRUCacheListener) tmp.elementAt(i)).resourceDropped(event);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------
}
