// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.StringTokenizer;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;


// }}}

/**
 * A collection of miscellaneous utility functions.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public abstract class Misc
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger ("de.qfs.lib.util.Misc");

    /**
     * Synchronization monitor.
     */
    private final static Object classlock = new Object ();

    /**
     * The characters for hexadecimal values.
     */
    private static final char[] nibbles = {
        '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
    };

    /**
     * Major version of java VM.
     */
    private static int majorVersion = -1;

    /**
     * Medium version of java VM.
     */
    private static int mediumVersion = -1;

    /**
     * Minor version of java VM.
     */
    private static int minorVersion = -1;

    // }}}

    //----------------------------------------------------------------------
    // Comparison
    //----------------------------------------------------------------------
    // {{{ equalOrNull

    /**
     * Test whether two objects are equal or both null.
     *
     * @param   o1      The first object.
     * @param   o2      The second object.
     *
     * @return  True if they are equal or both null.
     */
    public static boolean equalOrNull(Object o1, Object o2)
    {
        return o1 == null ? o2 == null : o1.equals(o2);
    }

    // }}}
    // {{{ equalOrEmpty

    /**
     * Test whether two Strings are equal or both empty or null.
     *
     * @param   s1      The first String.
     * @param   s2      The second String.
     *
     * @return  True if they are equal or both empty or null.
     */
    public static boolean equalOrEmpty(String s1, String s2)
    {
        return equalOrNull("".equals(s1) ? null : s1,
                           "".equals(s2) ? null : s2);
    }

    // }}}

    //----------------------------------------------------------------------
    // String methods
    //----------------------------------------------------------------------
    // {{{ lineSepToLineFeed

    /**
     * Convert system specific line separators in a String to just line feed
     * characters. Null Strings are OK.
     *
     * Any of the character combinations "\n", "\r\n", "\r" will be turned
     * into a single line-feed "\n".
     *
     * @param   text    The input string.
     *
     * @return  The converted string.
     */
    public static String lineSepToLineFeed(String text)
    {
        if (text == null) {
            return null;
        }
        StringBuffer ret = new StringBuffer ();
        boolean cr = false;
        for (StringTokenizer st = new StringTokenizer (text, "\r\n", true);
             st.hasMoreTokens(); ) {
            String token = st.nextToken();
            if ("\r".equals(token)) {
                if (cr) {
                    ret.append('\n');
                }
                cr = true;
            } else if ("\n".equals(token)) {
                ret.append('\n');
                cr = false;
            } else {
                if (cr) {
                    ret.append('\n');
                }
                ret.append(token);
                cr = false;
            }
        }
        if (cr) {
            ret.append('\n');
        }

        return ret.toString();
    }

    // }}}
    // {{{ countLines

    /**
     * Count the number of lines in a text String. Recognizes the line
     * separators \r, \n and \r\n.
     *
     * @param   text    The input text.
     *
     * @return  0 if text is NULL, empty or a single line break. The number of
     * lines otherwise.
     */
    public static int countLines(String text)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "countLines(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "text: " + text);
        }
        if (text == null) {
            return 0;
        }
        int lines = 0;
        boolean cr = false;
        boolean lf = false;
        boolean initial = true;
        for (StringTokenizer st = new StringTokenizer (text, "\r\n", true);
             st.hasMoreTokens(); ) {
            String token = st.nextToken();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "countLines(String)",
                           "token: " + token);
            }
            if ("\r".equals(token)) {
                if (cr || lf || initial) {
                    lines++;
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "countLines(String)",
                                   "lines: " + lines);
                    }
                }
                cr = true;
                lf = false;
            } else if ("\n".equals(token)) {
                if (lf || initial) {
                    lines++;
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "countLines(String)",
                                   "lines: " + lines);
                    }
                }
                lf = true;
                cr = false;
            } else {
                lines++;
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "countLines(String)",
                               "lines: " + lines);
                }
                cr = false;
                lf = false;
            }
            initial = false;
        }
        return lines;
    }

    // }}}
    // {{{ splitLines

    /**
     * Split a text into lines. Recognizes the line separators \r, \n and
     * \r\n.
     *
     * @param   text    The input text.
     *
     * @return An empty array if text is NULL or empty.
     * An array with the text lines otherwise. The result lines don't contain
     * line separators.
     */
    public static String[] splitLines(String text)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "splitLines(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "text: " + text);
        }
        if (text == null) {
            return new String[0];
        }
        Vector lines = new Vector ();
        boolean cr = false;
        boolean lf = false;
        boolean initial = true;
        for (StringTokenizer st = new StringTokenizer (text, "\r\n", true);
             st.hasMoreTokens(); ) {
            String token = st.nextToken();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "splitLines(String)",
                           "token: " + token);
            }
            if ("\r".equals(token)) {
                if (cr || lf || initial) {
                    lines.addElement("");
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "splitLines(String)",
                                   "added: \"\"");
                    }
                }
                cr = true;
                lf = false;
            } else if ("\n".equals(token)) {
                if (lf || initial) {
                    lines.addElement("");
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "splitLines(String)",
                                   "added: \"\"");
                    }
                }
                lf = true;
                cr = false;
            } else {
                lines.addElement(token);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "splitLines(String)",
                               "added: " + token);
                }
                lf = false;
                cr = false;
            }
            initial = false;
        }

        String[] result = new String[lines.size()];
        lines.copyInto(result);
        return result;
    }

    // }}}
    // {{{ quote(String,String)

    /**
     * Quote special characters in a String with a backslash.
     *
     * @param   data    The string.
     * @param   toQuote The characters to quote. Don't forget to add the
     *                  backslash itself.
     *
     * @return  The quoted String.
     */
    public static String quote(String data, String toQuote)
    {
        return quote(data, toQuote, '\\');
    }

    // }}}
    // {{{ quote(String,String,char)

    /**
     * Quote special characters in a String.
     *
     * @param   data    The string.
     * @param   toQuote The characters to quote. Don't forget to add the quote
     *                  character itself.
     * @param   quote   The quote character.
     *
     * @return  The quoted String.
     */
    public static String quote(String data, String toQuote, char quote)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "quote(String,String,char)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "data: " + data + ", " +
                       "toQuote: " + toQuote + ", " +
                       "quote: " + quote);
        }

        if (data == null) {
            return null;
        }

        if (toQuote == null) {
            return data;
        }

        StringBuffer sb = new StringBuffer ();
        for (int i = 0; i < data.length(); i++) {
            char c = data.charAt(i);
            if (toQuote.indexOf(c) >= 0) {
                sb.append(quote);
            }
            sb.append(c);
        }
        return sb.toString();
    }

    // }}}
    // {{{ unquote(String,String)

    /**
     * Unquote special characters in a String, quoted with a backslash.
     *
     * @param   data    The string.
     * @param   toQuote The characters to quote. Don't forget to add the
     *                  backslash itself.
     *
     * @return  The unquoted String.
     */
    public static String unquote(String data, String toQuote)
    {
        return unquote(data, toQuote, '\\');
    }

    // }}}
    // {{{ unquote(String,String,char)

    /**
     * Unquote special characters in a String.
     *
     * @param   data    The string.
     * @param   toQuote The characters to quote. Don't forget to add the quote
     *                  character itself.
     * @param   quote   The quote character.
     *
     * @return  The unquoted String.
     */
    public static String unquote(String data, String toQuote, char quote)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "unquote(String,String,char)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "data: " + data + ", " +
                       "toQuote: " + toQuote + ", " +
                       "quote: " + quote);
        }

        if (data == null) {
            return null;
        }

        if (toQuote == null) {
            return data;
        }

        StringBuffer sb = new StringBuffer ();
        for (int i = 0; i < data.length(); i++) {
            char c = data.charAt(i);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unquote(String,String,char)",
                           "i: " + i);
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unquote(String,String,char)",
                           "c: " + c);
            }
            if (c == quote && i < data.length() - 1) {
                char c2 = data.charAt(i + 1);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "unquote(String,String,char)",
                               "c2: " + c2);
                }
                if (data.indexOf(c2) >= 0) {
                    sb.append(c2);
                    i++;
                } else {
                    sb.append(c);
                }
            } else {
                sb.append(c);
            }
        }
        return sb.toString();
    }

    // }}}
    // {{{ unquotedIndexOf(String,String,String)

    /**
     * Find the index of an unquoted version of a character in a String.
     *
     * @param   data    The String.
     * @param   chars   The characters to look for.
     * @param   start   The position to start the search at.
     *
     * @return  The index of the first unquoted char or -1.
     */
    public static int unquotedIndexOf(String data, String chars, int start)
    {
        return unquotedIndexOf(data, chars, start, '\\');
    }

    // }}}
    // {{{ unquotedIndexOf(String,String,String,char)

    /**
     * Find the index of an unquoted version of a character in a String.
     *
     * @param   data    The String.
     * @param   chars   The characters to look for.
     * @param   start   The position to start the search at.
     * @param   quote   The quote character.
     *
     * @return  The index of the first unquoted char or -1.
     */
    public static int unquotedIndexOf(String data, String chars, int start,
                                      char quote)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "unquotedIndexOf(String,String,int,char)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "data: " + data + ", " +
                       "chars: " + chars + ", " +
                       "start: " + start + ", " +
                       "quote: " + quote);
        }

        if (start < 0 || (data != null && start > data.length())) {
            throw new IllegalArgumentException
                ("start " + start + " out of range");
        }
        if (data == null || chars == null) {
            return -1;
        }

        boolean quoted = false;
        for (int i = start; i < data.length(); i++) {
            char c = data.charAt(i);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unquotedIndexOf(String,String,int,char)",
                           "i: " + i);
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unquotedIndexOf(String,String,int,char)",
                           "c: " + c);
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unquotedIndexOf(String,String,int,char)",
                           "quoted: " + quoted);
            }

            if (c == quote) {
                quoted = ! quoted;
            } else if (quoted) {
                quoted = false;
            } else if (chars.indexOf(c) >= 0) {
                return i;
            }
        }
        return -1;
    }

    // }}}
    // {{{ toHex

    /**
     * Convert a byte array into a hex string.
     *
     * @param   data    The data to convert.
     *
     * @return  The hex string.
     */
    public static String toHex(byte[] data)
    {
        return toHex(data, 0, data.length);
    }

    // }}}
    // {{{ toHex

    /**
     * Convert a byte array into a hex string.
     *
     * @param   data    The data to convert.
     * @param   start   Start conversion at this position.
     * @param   count   Convert that many bytes.
     *
     * @return  The hex string.
     */
    public static String toHex(byte[] data, int start, int count)
    {
        StringBuffer sb = new StringBuffer ();
        int b;
        for (int i = start; i < start + count; i++) {
            b = data[i];
            sb.append(nibbles[(b >>> 4) & 0x0F]);
            sb.append(nibbles[b & 0x0F]);
        }
        return sb.toString();
    }

    // }}}
    // {{{ fromHex

    /**
     * Convert a hex String into a byte array.
     *
     * @param   hex     The input String.
     *
     * @return  The converted byte array.
     *
     * @throws  IllegalArgumentException        If the string contains non-hex
     *                                          characters.
     */
    public static byte[] fromHex(String hex)
        throws IllegalArgumentException
    {
        int len = hex.length();
        byte[] ret = new byte[(len + 1)/2];
        int i = 0;
        int k = 0;
        if (len % 2 != 0) {
            // implicit leading 0
            ret[k++] = (byte) hexVal(hex.charAt(i++));
        }
        for (; i < len; ) {
            ret[k++] = (byte) ((hexVal(hex.charAt(i++)) << 4)
                               | hexVal(hex.charAt(i++)));
        }
        return ret;
    }

    // }}}

    //----------------------------------------------------------------------
    // Java version methods
    //----------------------------------------------------------------------
    // {{{ getMajorJavaVersion

    /**
     * Get the major version of the Java VM.
     *
     * @return  The major version of the Java VM.
     */
    public static int getMajorJavaVersion()
    {
        parseVersion();
        return majorVersion;
    }

    // }}}
    // {{{ getMediumJavaVersion

    /**
     * Get the medium version of the Java VM.
     *
     * @return  The medium version of the Java VM.
     */
    public static int getMediumJavaVersion()
    {
        parseVersion();
        return mediumVersion;
    }

    // }}}
    // {{{ getMinorJavaVersion

    /**
     * Get the minor version of the Java VM.
     *
     * @return  The minor version of the Java VM.
     */
    public static int getMinorJavaVersion()
    {
        parseVersion();
        return minorVersion;
    }

    // }}}
    // {{{ isJavaVersion

    /**
     * Test whether the version of the Java VM matches a given version.
     *
     * @param       major   The major version to test, -1 to ignore.
     * @param       medium  The medium version to test, -1 to ignore.
     * @param       minor   The minor version to test, -1 to ignore.
     *
     * @return      True if the version matches, false otherwise.
     */
    public static boolean isJavaVersion(int major, int medium, int minor)
    {
        parseVersion();
        return (major < 0 || major == majorVersion)
            && (medium < 0 || medium == mediumVersion)
            && (minor < 0 || minor == minorVersion);
    }

    // }}}
    // {{{ isJavaVersionAtLeast

    /**
     * Test whether the version of the Java VM is at least a given version.
     *
     * @param       major   The major version to test, -1 to ignore.
     * @param       medium  The medium version to test, -1 to ignore.
     * @param       minor   The minor version to test, -1 to ignore.
     *
     * @return      True if the version matches, false otherwise.
     */
    public static boolean isJavaVersionAtLeast
        (int major, int medium, int minor)
    {
        parseVersion();
        if (major >= 0 && major < majorVersion) {
            return true;
        }
        if (major > majorVersion) {
            return false;
        }
        if (medium >= 0 && medium < mediumVersion) {
            return true;
        }
        if (medium > mediumVersion) {
            return false;
        }
        if (minor >= 0 && minor > minorVersion) {
            return false;
        }
        return true;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ hexVal

    /**
     * Get the hex value of one character.
     *
     * @param   c       The input character.
     *
     * @return  The hex value.
     *
     * @throws  IllegalArgumentException        If the character is not a hex
     *                                          digit.
     */
    private static int hexVal(char c)
        throws IllegalArgumentException
    {
        switch (c) {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            return (int) c - '0';
        case 'A':
        case 'B':
        case 'C':
        case 'D':
        case 'E':
        case 'F':
            return (int) c - ('A' - 10);
        case 'a':
        case 'b':
        case 'c':
        case 'd':
        case 'e':
        case 'f':
            return (int) c - ('a' - 10);
        }
        throw new IllegalArgumentException ("Not a hex digit: " + c);
    }

    // }}}
    // {{{ parseVersion

    /**
     * Parse the java version string.
     */
    private static void parseVersion()
    {
        synchronized (classlock) {
            if (majorVersion >= 0) {
                return;
            }
            majorVersion = 0;
            mediumVersion = 0;
            minorVersion = 0;

            String version = System.getProperty("java.version");

            int pos = version.indexOf('.');
            if (pos <= 0) {
                return;
            }
            try {
                majorVersion = Integer.parseInt(version.substring(0, pos));
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "parseVersion()", ex);
                }
                return;
            }
            version = version.substring(pos + 1);

            pos = version.indexOf('.');
            if (pos <= 0) {
                return;
            }
            try {
                mediumVersion = Integer.parseInt(version.substring(0, pos));
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "parseVersion()", ex);
                }
                return;
            }
            version = version.substring(pos + 1);

            pos = version.length();
            if (pos == 0) {
                return;
            }
            int p = 0;
            for (; p < version.length(); p++) {
                if (! Character.isDigit(version.charAt(p))) {
                    break;
                }
            }
            if (p > 0) {
                try {
                    minorVersion = Integer.parseInt(version.substring(0, p));
                } catch (NumberFormatException ex) {
                    if (logger.level >= Log.WRN) {
                        logger.log(Log.WRN, "parseVersion()", ex);
                    }
                    return;
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------
}
