// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports


import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A simple thread-safe FIFO buffer that stores Objects in an array. If the
 * buffer overflows old Objects are dropped.
 *
 * To pop all Objects from the RingBuffer you can either synchronize on the
 * RingBuffer object - which is deadlock prone - or call popAll.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public class RingBuffer
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.RingBuffer");

    /**
     * The actual buffer.
     */
    private Object[] buf;

    /**
     * The size of the internal buffer.
     */
    private int size;

    /**
     * The place where the next Object is going to be stored.
     */
    private int head;

    /**
     * The place where the first Object is stored.
     */
    private int tail;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ RingBuffer

    /**
     * Create a new RingBuffer.
     *
     * @param   size    The size of the buffer.
     */
    public RingBuffer (int size)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "RingBuffer(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "size: " + size);
        }
        // Internal size is one larger to discern between empty and full.
        this.size = size + 1;
        buf = new Object[this.size];
    }

    // }}}

    //----------------------------------------------------------------------
    // Accessors
    //----------------------------------------------------------------------
    // {{{ push

    /**
     * Push an Object to the front of the buffer.
     *
     * @param   object  The object to push.
     */
    public synchronized void push(Object object)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "push(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "object: " + object);
        }
        buf[head++] = object;
        if (head == size) {
            head = 0;
        }
        if (head == tail) {
            tail++;
            if (tail == size) {
                tail = 0;
            }
        }
    }

    // }}}
    // {{{ pop

    /**
     * Pop an object from the back of the buffer.
     *
     * @return  The popped object.
     *
     * @throws  IllegalStateException   If the buffer is empty.
     */
    public synchronized Object pop()
        throws IllegalStateException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "pop()", "");
        }

        if (head == tail) {
            throw new IllegalStateException ("The RingBuffer is empty");
        }

        Object ret = buf[tail++];
        if (tail == size) {
            tail = 0;
        }
        return ret;
    }

    // }}}
    // {{{ popAll

    /**
     * Pop all objects from the RingBuffer.
     *
     * @return An array with all the objects on the buffer or null if the
     * buffer is empty.
     */
    public synchronized Object[] popAll()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "popAll()", "");
        }
        if (head == tail) {
            return null;
        }

        Object[] ret = new Object[size()];
        if (head > tail) {
            System.arraycopy(buf, tail, ret, 0, head - tail);
        } else {
            System.arraycopy(buf, tail, ret, 0, size - tail);
            System.arraycopy(buf, 0, ret, size - tail, head);
        }
        head = 0;
        tail = 0;
        return ret;
    }

    // }}}
    // {{{ size

    /**
     * Get the number of Objects stored in the RingBuffer.
     *
     * @return  The number of Objects stored in the RingBuffer.
     */
    public synchronized int size()
    {
        int count =  head - tail;
        return count < 0 ? count + size : count;
    }

    // }}}
    // {{{ getMaxSize

    /**
     * Get the total size of the RingBuffer.
     *
     * @return  The total size of the RingBuffer.
     */
    public synchronized int getMaxSize()
    {
        return size - 1;
    }

    // }}}
    // {{{ setMaxSize

    /**
     * Set the total size of the RingBuffer.
     *
     * @param   newsize    The total size to set.
     */
    public synchronized void setMaxSize(int newsize)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setMaxSize(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "newsize: " + newsize);
        }

        if (newsize < 1) {
            throw new IllegalArgumentException
                ("Illegal size for RingBuffer: " +  newsize);
        }

        // Add room for the internal empty cell.
        newsize = newsize + 1;

        if (size == newsize) {
            return;
        }

        Object[] old = buf;
        buf = new Object[newsize + 1];
        if (head != tail) {
            int count = size();

            if (count > (newsize - 1)) {
                tail += count - (newsize - 1);
                if (tail >= size) {
                    tail -= size;
                }
                count = (newsize - 1);
            }

            if (head > tail) {
                System.arraycopy(old, tail, buf, 0, count);
            } else {
                System.arraycopy(old, tail, buf, 0, size - tail);
                System.arraycopy(old, 0, buf, size - tail, head);
            }
            head = count;
        } else {
            head = 0;
        }
        tail = 0;
        size = newsize;
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------
}
