// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.Reader;

import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Multiplexer that listens to input from an InputStream or a Reader and fires
 * Events when new data arrives.
 *
 * @author      Gregor Schmid
 * @since       0.99.0
 */
public class StreamMultiplexer
    extends Thread
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.StreamMultiplexer");

    /**
     * Counter for Thread names.
     */
    private static int count;

    /**
     * Whether to continue processing.
     */
    private boolean cont;

    /**
     * The Reader to receive data from.
     */
    private Reader reader;

    /**
     * Optional object to use as source for the generated StreamEvents.
     */
    private Object source;

    /**
     * The registered StreamListeners.
     */
    private Vector listeners = new Vector ();

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ StreamMultiplexer(InputStream)

    /**
     * Create a new StreamMultiplexer.
     *
     * @param   input   The Stream to receive data from.
     */
    public StreamMultiplexer (InputStream input)
    {
        this (new InputStreamReader(input), null);
    }

    // }}}
    // {{{ StreamMultiplexer(InputStream,Object)

    /**
     * Create a new StreamMultiplexer.
     *
     * @param   input   The Stream to receive data from.
     * @param   source  Optional object to use as source for the generated
     *                  StreamEvents. If null, the StreamMultiplexer itself
     *                  will be the source.
     */
    public StreamMultiplexer (InputStream input, Object source)
    {
        this (new InputStreamReader(input), source);
    }

    // }}}
    // {{{ StreamMultiplexer(Reader)

    /**
     * Create a new StreamMultiplexer.
     *
     * @param   reader  The Reader to receive data from.
     */
    public StreamMultiplexer (Reader reader)
    {
        this (reader, null);
    }

    // }}}
    // {{{ StreamMultiplexer(Reader,Object)

    /**
     * Create a new StreamMultiplexer.
     *
     * @param   reader  The Reader to receive data from.
     * @param   source  Optional object to use as source for the generated
     *                  StreamEvents. If null, the StreamMultiplexer itself
     *                  will be the source.
     */
    public StreamMultiplexer (Reader reader, Object source)
    {
        super ("StreamMultiplexer-" + count++);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "StreamMultiplexer(Reader,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "reader: " + reader + ", " +
                       "source: " + source);
        }
        setDaemon(true);
        this.reader = reader;
        this.source = source;
    }

    // }}}

    //----------------------------------------------------------------------
    // Getters and setters
    //----------------------------------------------------------------------
    // {{{ getSource

    /**
     * Get the source for the events created by the StreamMultiplexer.
     *
     * @return  The source of the StreamMultiplexer.
     */
    public final Object getSource()
    {
        return source;
    }

    // }}}
    // {{{ setSource

    /**
     * Set the source for the events created by the StreamMultiplexer.
     *
     * @param   source  The source to set.
     */
    public final void setSource(Object source)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSource(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "source: " + source);
        }
        this.source = source;
    }



    // }}}
    // {{{ isClosed

    /**
     * Test whether the source of the Multiplexer has been closed or the
     * multiplexer stopped.
     *
     * @return      The closed state of the StreamMultiplexer.
     */
    public final boolean isClosed()
    {
        return ! cont;
    }

    // }}}

    //----------------------------------------------------------------------
    // Work methods
    //----------------------------------------------------------------------
    // {{{ run

    /**
     * The actual work method of the thread.
     */
    public void run()
    {
        cont = true;
        char[] buf = new char[1000];

        while (cont) {
            try {
                int cnt = reader.read(buf);
                if (cnt == -1) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "run()",
                                   "EOF");
                    }
                    cont = false;
                } else if (cnt > 0) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "run()",
                                   "cnt: " + cnt);
                    }
                    if (logger.level >= Log.DBGDETAIL) {
                        logger.log(Log.DBGDETAIL, "run()",
                                   "data: " + new String (buf, 0, cnt));
                    }
                    char[] tmp = new char[cnt];
                    System.arraycopy(buf, 0, tmp, 0, cnt);
                    fireNewInput(tmp);
                }
            } catch (InterruptedIOException ex) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "run()", ex);
                }
                cont = false;
            } catch (IOException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "run()", ex);
                }
                cont = false;
            }
        }
        fireInputClosed();
    }


    // }}}
    // {{{ halt

    /**
     * Halt the multiplexer.
     */
    public void halt ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "halt()", "");
        }
        cont = false;
        Thread.currentThread().interrupt();
    }


    // }}}

    //----------------------------------------------------------------------
    // Listener stuff
    //----------------------------------------------------------------------
    // {{{ addStreamListener

    /**
     * Add a listener to the StreamMultiplexer.
     *
     * @param   listener        The listener to add.
     */
    public synchronized void addStreamListener(StreamListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addStreamListener(StreamListener)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "listener: " + listener);
        }
        // avoid double calls
        listeners.removeElement(listener);
        listeners.addElement(listener);
    }

    // }}}
    // {{{ removeStreamListener

    /**
     * Remove a listener from the StreamMultiplexer.
     *
     * @param   listener        The listener to remove.
     */
    public synchronized void removeStreamListener(StreamListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addStreamListener(StreamListener)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "listener: " + listener);
        }
        listeners.removeElement(listener);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ fireNewInput

    /**
     * Notify all listeners that new data has arrived.
     *
     * @param   data   The newly arrived data.
     */
    protected void fireNewInput(char[] data)
    {
        Vector tmp;
        synchronized (this) {
            if (listeners.size() == 0) {
                return;
            }
            tmp = (Vector) listeners.clone();
        }
        StreamEvent event = new StreamEvent
            (source == null ? this : source, data);
        for (int i = 0; i < tmp.size(); i++) {
            ((StreamListener) tmp.elementAt(i)).newInput(event);
        }
    }

    // }}}
    // {{{ fireInputCLosed

    /**
     * Notify all listeners that input source was closed.
     */
    protected void fireInputClosed()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "fireInputCLosed()", "");
        }
        Vector tmp;
        synchronized (this) {
            if (listeners.size() == 0) {
                return;
            }
            tmp = (Vector) listeners.clone();
        }
        StreamEvent event = new StreamEvent (source == null ? this : source);
        for (int i = 0; i < tmp.size(); i++) {
            ((StreamListener) tmp.elementAt(i)).inputClosed(event);
        }
    }

    // }}}

}
