// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.LinkedList;


import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A TaskQueue is a Thread that executes runnables from a queue. <p>
 *
 * It is used to handle related tasks, one after the other, from a single
 * background thread. New tasks can be added with {@link #addTask addTask}
 * and the thread can be terminated with the {@link #finish finish} method.
 *
 * @author      Gregor Schmid
 */
public class TaskQueue
    extends Thread
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.TaskQueue");

    /**
     * Counter for thread names.
     */
    private static int id = 0;

    /**
     * The queue of runnables.
     */
    private LinkedList queue = new LinkedList ();

    /**
     * Flag to terminate the TaskQueue.
     */
    private int terminated = 0;

    // }}}

    // {{{ constructors

    /**
     * Create a new TaskQueue. This does not start the TaskQueue.
     */
    public TaskQueue ()
    {
        super("TaskQueue" + id++);
    }

    /**
     * Create a new TaskQueue. This does not start the TaskQueue.
     *
     * @param   name    The thread name for the TaskQueue.
     */
    public TaskQueue (String name)
    {
        super(name);
    }

    // }}}

    // {{{ run

    /**
     * Remove tasks from the queue one by one and execute them.
     */
    public void run()
    {
        while(terminated < 2) {
            Runnable task = null;
            synchronized (this) {
                while (terminated < 2 && queue.isEmpty()) {
                    try {
                        wait();
                    } catch (InterruptedException ex) {
                    }
                }
                if (terminated < 2 && ! queue.isEmpty()) {
                    task = (Runnable) queue.removeFirst();
                    if (queue.isEmpty() && terminated == 1) {
                        terminated = 2;
                    }
                }
            }
            if (task != null) {
                try {
                    task.run();
                } catch (Throwable ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log("run()", ex);
                    }
                }
                task = null;
            }
        }
    }

    // }}}
    // {{{ finish

    /**
     * Terminate the operation of the TaskQueue.
     *
     * @param immediately       If true, terminate the TaskQueue as soon as
     *                          possible. Otherwise no more new tasks will be
     *                          accepted and the thread will be terminated
     *                          when all pending tasks have been executed.
     */
    public synchronized void finish(boolean immediately)
    {
        terminated = (immediately || queue.isEmpty()) ? 2 : 1;
        notifyAll();
    }

    // }}}
    // {{{ addTask

    /**
     * Add a task to the TaskQueue. It will be executed as soon as possible.
     *
     * @param   task    The Runnable to add.
     */
    public synchronized void addTask(Runnable task)
    {
        if (terminated == 0) {
            queue.addLast(task);
        }
        notifyAll();
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------
}
