// {{{ copyright

/********************************************************************
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A Thread pool manages a number of Threads to execute asynchronous calls.
 * It helps to avoid Thread creation overhead by keeping the Threads alive
 * and waiting for something to do. <p>
 *
 * The number of live Threads can be customized with the parameters to its
 * constructor. Threads are assigned first come first serve. The priority of
 * of the calling Thread does not make any difference. <p>
 *
 * Calls are executed by passing a Runnable to {@link #execute execute}. Its
 * run Method will be called from the next available Thread.
 *
 * Additionally, the ThreadPool offers the ability to run "watched" tasks with
 * the {@link #executeWatched executeWatched} method. These tasks are run in a
 * background pool thread while the caller is blocked. If the task is
 * cancelled or a timeout is exceeded before the task is done, the call is
 * interupted and control returns to the caller. This is very handy for RMI
 * calls and other non-interruptable tasks.
 *
 * @author      Gregor Schmid
 */
public class ThreadPool
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger ("de.qfs.lib.util.ThreadPool");

    /**
     * The Logger used for logging PoolThread methods.
     */
    private final static Logger ptLogger =
        new Logger ("de.qfs.lib.util.ThreadPool.PoolThread");

    /**
     * The Logger used for logging WatchDog methods.
     */
    private final static Logger wdLogger =
        new Logger ("de.qfs.lib.util.ThreadPool.WatchDog");

    /**
     * Available Threads.
     */
    private List available = new ArrayList ();

    /**
     * Waiters for a Thread.
     */
    private LinkedList waiters = new LinkedList ();

    /**
     * Minimum number of threads.
     */
    private int minThreads;

    /**
     * Average number of threads.
     */
    private int avgThreads;

    /**
     * Maximum number of threads.
     */
    private int maxThreads;

    /**
     * Number of threads currently available.
     */
    private int currentThreads;

    /**
     * ThreadGroup used to create the threads in.
     */
    private ThreadGroup threadGroup;

    /**
     * Id counter for PoolThreads.
     */
    private static int count;

    // }}}

    // {{{ constructors

    /**
     * Create a new ThreadPool.
     *
     * @param   min     Initial number of threads to create.
     * @param   avg     Don't release threads unless there are more than avg
     *                  Threads around.
     * @param   max     Create at most that many Threads.
     */
    public ThreadPool (int min, int avg, int max)
    {
        this (min, avg, max, null);
    }

    /**
     * Create a new ThreadPool.
     *
     * @param   min     Initial number of threads to create.
     * @param   avg     Don't release threads unless there are more than avg
     *                  Threads around.
     * @param   max     Create at most that many Threads.
     * @param   group   Create all threads in this ThreadGroup
     */
    public ThreadPool (int min, int avg, int max, ThreadGroup group)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ThreadPool(int,int,int,ThreadGroup)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "min: " + min + ", " +
                       "avg: " + avg + ", " +
                       "max: " + max + ", " +
                       "group: " + group);
        }

        minThreads = min;
        avgThreads = avg;
        maxThreads = max;
        threadGroup = group;

        if (minThreads > avgThreads
            || avgThreads > maxThreads
            || maxThreads <= 0) {
            throw new IllegalArgumentException
                ("Min, avg and max values don't make sense.");
        }

        // Create initial numbers of threads
        for (currentThreads = 0;
             currentThreads < minThreads;
             currentThreads++) {
            PoolThread pt = new PoolThread (threadGroup);
            pt.start();
            available.add(pt);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ execute

    /**
     * Execute a runnable in the next available Thread. If there is no Thread
     * currently available and the maximum number of Threads is reached, the
     * calling Thread will be blocked, until a Thread is available.
     *
     * @param   run     The runnable whose run method will be called.
     */
    public void execute(Runnable run)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "execute(Runnable)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "run: " + run);
        }
        allocate().execute(run);
    }

    // }}}
    // {{{ executeWatched

    /**
     * Execute a task in the next available thread and block until the task is
     * finished. If the task is cancelled by {@link #cancelCall cancelCall} or
     * the timeout is exceeded, interrupt the thread and return execution to
     * the caller.
     *
     * @param   run     The runnable whose run method will be called.
     * @param   timeout Timeout after which the task is interrupted.
     *                  0 for no timeout.
     * @param   callid  An id that uniquely identifies the task. It is the
     *                  callers responsibility to ensure that ids are unique.
     *
     * @return  The result of the UnsafeRunnable's run method.
     *
     * @throws  CancelledException      If the task is cancelled.
     * @throws  TimedOutException       If the task is interrupted after the
                                        timeout.
     * @throws  Throwable               If the runnable throws anything.
     *
     * @since   0.99.0
     */
    public Object executeWatched(UnsafeRunnable run, long timeout, long callid)
        throws Throwable
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "executeWatched(UnsafeRunnable,long,long)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "run: " + run + ", " +
                       "timeout: " + timeout + ", " +
                       "callid: " + callid);
        }
        PoolThread pt = allocate();
        try {
            pt.execute(run);
            synchronized(pt) {
                try {
                    boolean[] timedOut = new boolean[] {false};
                    WatchDog.add(Thread.currentThread(), timeout, callid,
                                 timedOut);
                    while (pt.runner != null) {
                        try {
                            if (logger.level >= Log.DBG) {
                                logger.log(Log.DBG,
                                           "executeWatched(UnsafeRunnable," +
                                           "long,long)",
                                           "Blocking...");
                            }
                            pt.wait();
                        } catch (InterruptedException ex) {
                            if (logger.level >= Log.DBG) {
                                logger.log(Log.DBG,
                                           "executeWatched(UnsafeRunnable," +
                                           "long,long)",
                                           "Interrupted");
                            }
                            pt.interrupt();
                            if (timedOut[0]) {
                                throw new TimedOutException
                                    ("Call " + callid + " timed out after " +
                                     timeout + " milliseconds");
                            } else {
                                throw new CancelledException
                                    ("Call " + callid + " was cancelled");
                            }
                        }
                    }
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG,
                                   "executeWatched(UnsafeRunnable,long,long)",
                                   "Done");
                    }
                } finally {
                    WatchDog.remove(callid);
                }
            }
            if (pt.exception != null) {
                throw pt.exception;
            }
            return pt.retval;
        } finally {
            release(pt);
        }
    }

    // }}}
    // {{{ cancelCall

    /**
     * Cancel a call executed with {@link #executeWatched executeWatched}.
     *
     * @param   callid  The id of the task.
     *
     * @since   0.99.0
     */
    public void cancelCall(long callid)
    {
        WatchDog.cancel(callid);
    }

    // }}}

    //----------------------------------------------------------------------
    // Public interfaces
    //----------------------------------------------------------------------
    // {{{ interface UnsafeRunnable

    /**
     * This interface is similar to Runnable, except that the run method
     * may return a value and throw any kind of Exception.
     */
    public static interface UnsafeRunnable
    {
        /**
         * Execute the unsafe runnable.
         *
         * @return      Whatever the UnsafeRunnable decides to return.
         *
         * @throws      Exception       If anything goes wrong.
         */
        public Object run() throws Exception;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ allocate

    /**
     * Allocate a pool thread. If no thread is available, block until one is.
     *
     * @return  The nexet available thread.
     */
    private synchronized PoolThread allocate()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "allocate()", "");
        }
        if (available.size() > 0) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "allocate()",
                           "Thread available: " + available.get(0));
            }
            return (PoolThread) available.remove(0);
        }
        if (currentThreads < maxThreads) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "allocate()",
                           "Must create new Thread");
            }
            currentThreads++;
            PoolThread pt = new PoolThread (threadGroup);
            pt.start();
            return pt;
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "allocate()",
                       "Waiting for Thread");
        }
        ResourceWaiter waiter = new ResourceWaiter (this);
        waiters.add(waiter);
        PoolThread pt = (PoolThread) waiter.waitFor(-1);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "allocate()",
                       "got it: " + pt);
        }
        return pt;
    }

    // }}}
    // {{{ release

    /**
     * Release a pool thread and make it available again.
     *
     * @param   thread  The thread to release.
     */
    private synchronized void release(PoolThread thread)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "release(PoolThread)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "thread: " + thread);
        }
        if (waiters.size() > 0) {
            ((ResourceWaiter) waiters.remove(0)).handResource(thread);
        } else if (currentThreads > avgThreads
                   && available.size() >= minThreads) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "release(PoolThread)",
                           "Terminating Thread " + thread);
            }
            thread.terminate();
            currentThreads--;
        } else {
            available.add(thread);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class PoolThread

    /**
     * A worker Thread used by the ThreadPool.
     */
    private class PoolThread extends Thread
    {
        // {{{ variables

        /**
         * The task to execute.
         */
        private Object runner;

        /**
         * Signal to quit the PoolThread.
         */
        private boolean quit;

        /**
         * Return value from last call.
         */
        private Object retval;

        /**
         * Exception thrown in last call.
         */
        private Throwable exception;

        // }}}

        // {{{ constructor

        /**
         * Create a new PoolThread.
         *
         * @param       group   The ThreadGroup of which the PoolThread should
         *                      be a member.
         */
        public PoolThread (ThreadGroup group)
        {
            super(group, "PoolThread-" + count++);

            if (ptLogger.level >= Log.MTD) {
                ptLogger.log(Log.MTD, "PoolThread()", "count: " + count++);
            }
            setDaemon(true);
        }

        // }}}
        // {{{ run

        /**
         * Wait for a tasks to become available and execute them.
         */
        public void run()
        {
            if (ptLogger.level >= Log.MTD) {
                ptLogger.log(Log.MTD, "run()", "");
            }
            Object runnable;
            while (true) {
                synchronized (this) {
                    while (runner == null && !quit) {
                        try {
                            this.wait();
                        } catch (InterruptedException ex) {
                        }
                    }
                    runnable = runner;
                    if (quit) {
                        runner = null;
                        this.notifyAll();
                        break;
                    }
                }
                try {
                    if (ptLogger.level >= Log.DBG) {
                        ptLogger.log(Log.DBG, "run()",
                                     "Executing " + runnable);
                    }
                    retval = null;
                    exception = null;
                    if (runnable instanceof Runnable) {
                        ((Runnable) runnable).run();
                    } else {
                        retval = ((UnsafeRunnable) runnable).run();
                    }
                } catch (Throwable th) {
                    if (runnable instanceof Runnable) {
                        if (ptLogger.level >= Log.ERR) {
                            ptLogger.log("run()", th);
                        }
                    } else {
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "run()", th);
                        }
                    }
                    exception = th;
                } finally {
                    synchronized (this) {
                        runner = null;
                        this.notifyAll();
                    }
                }
                if (runnable instanceof Runnable) {
                    // unsafe must fetch values first
                    release(this);
                }
            }
        }



        // }}}
        // {{{ execute

        /**
         * Execute one task.
         *
         * @param       runnable        The task to execute.
         */
        public synchronized void execute(Runnable runnable)
        {
            _execute(runnable);
        }

        // }}}
        // {{{ execute

        /**
         * Execute one task.
         *
         * @param       runnable        The task to execute.
         */
        public synchronized void execute(UnsafeRunnable runnable)
        {
            _execute(runnable);
        }

        // }}}
        // {{{ terminate

        /**
         * Quit waiting for tasks and terminate the PoolThread.
         */
        public synchronized void terminate()
        {
            if (ptLogger.level >= Log.MTD) {
                ptLogger.log(Log.MTD, "terminate()",
                             "this: " + this);
            }
            quit = true;
            this.notifyAll();
        }

        // }}}

        //------------------------------------------------------------------
        // Helper methods
        //------------------------------------------------------------------
        // {{{ _execute

        /**
         * Execute one task.
         *
         * @param       runnable        The task to execute.
         */
        private synchronized void _execute(Object runnable)
        {
            if (ptLogger.level >= Log.MTD) {
                ptLogger.log(Log.MTD, "_execute(Object)",
                             ptLogger.level < Log.MTDDETAIL ? "" :
                             "this: " + this +
                             ", runnable: " + runnable);
            }
            while (runner != null) {
                try {
                    this.wait();
                } catch (InterruptedException ex) {
                }
            }
            runner = runnable;
            this.notifyAll();
        }



        // }}}
    }

    // }}}
    // {{{ class WatchDog

    /**
     * This helper class is used to watch over tasks executed by {@link
     * #executeWatched executeWatched}.
     */
    private static class WatchDog
        extends Thread
    {
        // {{{ variables

        /**
         * Id counter for thread name.
         */
        private static int id = 0;

        /**
         * The threads being watched by callid.
         */
        private Hashtable threads = new Hashtable ();

        /**
         * The timedOut return holders by callid.
         */
        private Hashtable timedOuts = new Hashtable ();

        /**
         * A list of timeouts.
         */
        private List timeouts = new ArrayList ();

        /**
         * A paraalel list of callids.
         */
        private List callids = new ArrayList ();

        /**
         * Static call lock.
         */
        private static Object lock = new Object ();

        /**
         * Static WatchDog singleton.
         */
        private static WatchDog instance;

        /**
         * Flag used to shut thread down.
         */
        private boolean stop = false;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ WatchDog

        /**
         * Create a new WatchDog.
         */
        private WatchDog ()
        {
            super ("WatchDog-" + id++);
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "WatchDog()", "");
            }
            setDaemon(true);
            start();
        }

        // }}}

        //------------------------------------------------------------------
        // Static methods
        //------------------------------------------------------------------
        // {{{ add

        /**
         * Add a new thread for the WatchDog to watch over.
         *
         * @param       thread  The thread to add.
         * @param       timeout The timeout for the task.
         * @param       callid  The task's unique id.
         * @param       timedOut A retrun value holder for whether the task
         *                       timed out.
         *
         * @since       0.99.0
         */
        public static void add(Thread thread, long timeout, long callid,
                               boolean[] timedOut)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "add(Thread,long,long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "thread: " + thread + ", " +
                           "timeout: " + timeout + ", " +
                           "callid: " + callid);
            }
            synchronized (lock) {
                if (instance == null) {
                    instance = new WatchDog ();
                }
                instance.addThread(thread, timeout, callid, timedOut);
            }
        }

        // }}}
        // {{{ remove

        /**
         * Remove a thread from the WatchDog after its task has finished.
         *
         * @param       callid  The thread to remove.
         *
         * @since       0.99.0
         */
        public static void remove(long callid)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "remove(long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "callid: " + callid);
            }
            synchronized (lock) {
                if (instance == null) {
                    return;
                }
                instance.removeThread(callid);
            }
        }

        // }}}
        // {{{ cancel

        /**
         * Cancel a task maintained by the WatchDog.
         *
         * @param       callid  The task's id.
         *
         * @since       0.99.0
         */
        public static void cancel(long callid)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "cancel(long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "callid: " + callid);
            }
            synchronized (lock) {
                if (instance == null) {
                    return;
                }
                instance.cancelThread(callid);
            }
        }

        // }}}

        //------------------------------------------------------------------
        // The run method
        //------------------------------------------------------------------
        // {{{ run

        /**
         * The worker method of the WatchDog. Check for scheduled tasks and
         * their timeouts. Block until the next timeout is reached or any
         * tasks are added, removed or cancelled.
         *
         * @since       0.99.0
         */
        public void run()
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "run()", "");
            }
            synchronized (this) {
                while (!stop) {
                    // wait until timeout is available
                    while (timeouts.size() == 0) {
                        try {
                            if (wdLogger.level >= Log.DBG) {
                                wdLogger.log(Log.DBG, "run()",
                                           "Waiting for threads...");
                            }
                            wait();
                        } catch (InterruptedException ex) {
                            if (wdLogger.level >= Log.DBG) {
                                wdLogger.log(Log.DBG, "run()",
                                           "Interrupted");
                            }
                            break;
                        }
                    }
                    if (stop) {
                        break;
                    }
                    if (wdLogger.level >= Log.DBG) {
                        wdLogger.log(Log.DBG, "run()",
                                     "Got threads, checking timeouts");
                    }
                    // interrupt all timed out threads
                    long now = System.currentTimeMillis();
                    while (timeouts.size() > 0) {
                        if (((Long) timeouts.get(0)).longValue() > now) {
                            break;
                        }
                        timeouts.remove(0);
                        long id = ((Long) callids.remove(0)).longValue();
                        timeout(id);
                    }
                    if (stop) {
                        break;
                    }
                    if (timeouts.size() > 0) {
                        // wait until next timeout
                        if (wdLogger.level >= Log.DBG) {
                            wdLogger.log(Log.DBG, "run()",
                                       "Waiting for next timeout");
                        }
                        try {
                            wait(((Long) timeouts.get(0)).longValue()
                                 - System.currentTimeMillis());
                        } catch (InterruptedException ex) {
                            if (wdLogger.level >= Log.DBG) {
                                wdLogger.log(Log.DBG, "run()",
                                           "Interrupted");
                            }
                            break;
                        }
                    }
                    if (stop) {
                        break;
                    }
                }
            }
        }

        // }}}

        //------------------------------------------------------------------
        // Helper methods
        //------------------------------------------------------------------
        // {{{ addThread

        /**
         * Add a new thread for the WatchDog to watch over.
         *
         * @param       thread  The thread to add.
         * @param       timeout The timeout for the task.
         * @param       callid  The task's unique id.
         * @param       timedOut A retrun value holder for whether the task
         *                       timed out.
         *
         * @since       0.99.0
         */
        private void addThread(Thread thread, long timeout, long callid,
                               boolean[] timedOut)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "addThread(Thread,long,long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "thread: " + thread + ", " +
                           "timeout: " + timeout + ", " +
                           "callid: " + callid);
            }
            synchronized(this) {
                threads.put(new Long (callid), thread);
                timedOuts.put(new Long (callid), timedOut);
                if (timeout > 0) {
                    long until = System.currentTimeMillis() + timeout;
                    int i = 0;
                    for (; i < timeouts.size(); i++) {
                        if (((Long) timeouts.get(i)).longValue() > until) {
                            break;
                        }
                    }
                    timeouts.add(i, new Long (until));
                    callids.add(i, new Long (callid));
                }
                notifyAll();
            }
        }

        // }}}
        // {{{ removeThread

        /**
         * Remove a thread from the WatchDog after its task has finished.
         *
         * @param       callid  The thread to remove.
         *
         * @since       0.99.0
         */
        private void removeThread(long callid)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "removeThread(long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "callid: " + callid);
            }
            synchronized(this) {
                threads.remove(new Long (callid));
                timedOuts.remove(new Long (callid));
                for (int i = 0; i < callids.size(); i++) {
                    if (((Long) callids.get(i)).longValue() == callid) {
                        callids.remove(i);
                        timeouts.remove(i);
                        break;
                    }
                }
                notifyAll();
            }
        }

        // }}}
        // {{{ cancelThread

        /**
         * Cancel a task maintained by the WatchDog.
         *
         * @param       callid  The task's id.
         *
         * @since       0.99.0
         */
        private void cancelThread(long callid)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "cancelThread(long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "callid: " + callid);
            }
            synchronized(this) {
                Thread thread = (Thread) threads.get(new Long (callid));
                threads.remove(new Long (callid));
                timedOuts.remove(new Long (callid));
                for (int i = 0; i < callids.size(); i++) {
                    if (((Long) callids.get(i)).longValue() == callid) {
                        callids.remove(i);
                        timeouts.remove(i);
                        break;
                    }
                }
                if (thread != null) {
                    thread.interrupt();
                }
                notifyAll();
            }
        }

        // }}}
        // {{{ timeout

        /**
         * Interrupt a timed out task.
         *
         * @param       callid  The id of the task.
         *
         * @since       0.99.0
         */
        private synchronized void timeout(long callid)
        {
            if (wdLogger.level >= Log.MTD) {
                wdLogger.log(Log.MTD, "timeout(long)",
                           wdLogger.level < Log.MTDDETAIL ? "" :
                           "callid: " + callid);
            }
            Long ci = new Long(callid);
            Thread thread = (Thread) threads.get(ci);
            threads.remove(ci);
            boolean[] timedOut = (boolean[]) timedOuts.get(ci);
            timedOuts.remove(ci);
            if (thread == null) {
                if (wdLogger.level >= Log.ERR) {
                    wdLogger.log(Log.ERR, "timeout(long)",
                               "internal error, lost thread for " + callid);
                }
            } else {
                timedOut[0] = true;
                thread.interrupt();
            }
            notifyAll();
        }

        // }}}
    }

    // }}}
    // {{{ class TimedOutException

    /**
     * This exception is thrown when a call executed with {@link
     * #executeWatched executeWatched} times out and is interrupted by the
     * WatchDog.
     */
    public static class TimedOutException
        extends Exception
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ TimedOutException

        /**
         * Create a new TimedOutException.
         */
        public TimedOutException (String msg)
        {
            super (msg);
        }

        // }}}
    }


    // }}}
    // {{{ class CancelledException

    /**
     * This exception is thrown when a call executed with {@link
     * #executeWatched executeWatched} is explicitely cancelled with {@link
     * #cancelCall cancelCall}.
     */
    public static class CancelledException
        extends Exception
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ CancelledException

        /**
         * Create a new CancelledException.
         */
        public CancelledException (String msg)
        {
            super (msg);
        }

        // }}}
    }


    // }}}
}
